/*
 *      s817drv.c  -- Sensoray 817 driver
 *      Copyright (C) 2008 Sensoray Company Inc.
 *      Authors: (D.A., A.K.)
 *               V4L2 ported from driver by S.S.
 *
 *      This driver was designed to match the firmware and other OS drivers
 *      If making changes to driver, please try to keep the same 
 *      structure. 
 *  
 *      Driver was also designed to be backwards compatible with as many
 *      kernel versions as possible.  The driver should be useable both with and without V4L.
 *
 *      Define S817_V4L2 to enable the V4L2 interface.  
 *      At this time the V4L2 driver portion
 *      is strictly experimental/Beta.
 *      It is not officially supported by Sensoray.
 *      The SDK ships with S817_V4L2 turned off.
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/module.h>
#include <linux/version.h>
#define MUTEX_INTRODUCED               KERNEL_VERSION(2, 6, 16)
#define KZALLOC_INTRODUCED             KERNEL_VERSION(2, 6, 14)
#define VERSION_IRQF_SHARED_INTRODUCED KERNEL_VERSION(2, 6, 18)
#define VERSION_PTREG_REMOVED_IRQ      KERNEL_VERSION(2, 6, 19)
#define VERSION_NOPAGE_REMOVED         KERNEL_VERSION(2, 6, 26)
#define VERSION_SGSETPAGE_INTRODUCED   KERNEL_VERSION(2, 6, 24)
#define VERSION_UNLOCKED_IOCTL         KERNEL_VERSION(2, 6, 38)
#define CREATE_PROC_REMOVED            KERNEL_VERSION(3, 10, 0)
#define VERSION_VMALLOC                KERNEL_VERSION(4, 4, 3)
#include <linux/pagemap.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/firmware.h>
#if defined(MUTEX_LOCK) || defined(MUTEX_UNLOCK) || defined(MUTEX_INIT)
#error "kernel API changed, contact support@sensoray.com"
#endif

#if LINUX_VERSION_CODE >= MUTEX_INTRODUCED
#include <linux/mutex.h>
#define MUTEX_LOCK   mutex_lock
#define MUTEX_LOCK_INTERRUPTIBLE   mutex_lock_interruptible
#define MUTEX_UNLOCK mutex_unlock
#define MUTEX_INIT   mutex_init
#else
#define MUTEX_LOCK   down
#define MUTEX_LOCK_INTERRUPTIBLE   down_interruptible
#define MUTEX_UNLOCK up
#define MUTEX_INIT   init_MUTEX
#endif
#include <linux/delay.h>
#include <linux/proc_fs.h>
#ifdef S817_V4L2
#include <linux/videodev2.h>
#include <media/v4l2-common.h>
#include <media/v4l2-ioctl.h>
#include <asm/rtc.h>
#include <linux/ctype.h>
#endif
#if LINUX_VERSION_CODE < KZALLOC_INTRODUCED
static inline void *kzalloc(size_t size, int flags)
{
	void *res;
	res = kmalloc(size, flags);
	if (res != NULL)
		memset(res, 0, size);
	return res;	
}	
#endif
#if LINUX_VERSION_CODE >= CREATE_PROC_REMOVED
#include <linux/seq_file.h>
static int s817_show_proc_output(struct seq_file *s, void *unused);

static int s817_procseq_open(struct inode *inode, struct file *file)
{
	return single_open(file, s817_show_proc_output, PDE_DATA(inode));
}

struct file_operations proc_file= {
	.owner = THIS_MODULE,
	.open = s817_procseq_open,
	.read  = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};

#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
/* __devinit and __devexit removed in these versions */
#define __devinit
#define __devexit
#define __devinit_p(x) x
#define __devexit_p(x) x
#endif

#ifndef DMA_32BIT_MASK
#define DMA_32BIT_MASK DMA_BIT_MASK(32)
#endif

#ifndef VM_RESERVED
#define VM_RESERVED (VM_DONTEXPAND | VM_DONTDUMP)
#endif

#if LINUX_VERSION_CODE >= VERSION_VMALLOC
#include <linux/vmalloc.h>
#endif


#include "s817ioctl.h"

#define MAX_BOARDS 6
#ifdef S817_V4L2
#define S817_VERSION_CODE KERNEL_VERSION(2, 0, 4)
#define S817_VERSION_TEXT "Ver. 2.0.4: July 23, 2010 (Beta V4L2 capable)"
#else
#define S817_VERSION_TEXT "Ver. 2.0.4: July 23, 2010"
#endif
#define DFLT_MAJOR         127

/* The 817 has 5 pci devices, but is a composite device */
/* The board sits behind a PLX PEX8111 */
#define FIRMWARE_FILE_NAME "fw817.slf"
#define DSP_REG_BASE	0x01800000 /* base for register access
				      (subtract from reg addr) */

#define DSP_HSR		0x01C1FFF0 /* HSR */
#define DSP_HSR_INTSRC	1  /* int source active bitmask in HSR */
#define DSP_HSR_INTAVAL 2  /* INTA value bitmask in HSR
			      (1 when INTA asserted) */
#define	DSP_HSR_INTAM	4  /*int enable(0)/disable(1) bitmask in HSR*/
#define DSP_HDCR	0x01C1FFF4	/* HDCR */
#define DSP_HDCR_DSPINT    2 /* host-to-DSP interrupt bit */
#define DSP_HDCR_WARMRESET 1 /* DSP warm reset bit */
#define DSP_DSPP	   0x01C1FFF8 /* DSPP */
#define DSP_PAGE_ISRAMSTART 0x0   /* page <-> 0x00000000 (start of ISRAM) */
#define DSP_PAGE_SDRAMSTART 0x200  /*page <-> 0x80000000 (start of SDRAM) */
#define DSP_JUMPCODE_SIZE   0x20  /* size of jumpEntryPoint code */
/* DSP registers */
#define	DSP_PCIIS 0x01c00008	/* PCI interrupt source reg */
#define DSP_MA	 0x01C00010	/* DSP master address register */
#define DSP_PCIMA 0x01C00014	/* PCI master address register */
#define DSP_PCIMC 0x01C00018	/* PCI master control register */
#define DSP_CDSPA 0x01C0001C    /* Current DSP address register */
#define DSP_CPCIA 0x01C00020	/* Current PCI address register */
#define DSP_CCNT  0x01C00024 /* Current byte count register */
#define DSP_PCI_TRCTL  0x01C30000 /*PCI transfer request control register */

/*DM642 registers used to for communication between host and DSP:*/
/*HDCR_xxx : Host to DSP Communication Register*/

#define HDCR_FLAG		0x01c40184

/*2 (VP0 TSI_STCMPL0) (former DSP_COMM_REG0)
 *HOST->DSP: address of command/status packet in bits 31:12
 *                                    (packet buffer is on a 4K page boundary)
 *           DSP destination a command is addressed to (chanId) in bits 3:1
 *           If address (bits 32:12) is 0, then it's a short command:
 *			 bits 11:4 contain command code,
 *			 add'l argument is passed in HDR_SHORT_CMD_ARG register.
 *Host writes PCI address of the packet
 *DSP resets it to 0 after having read the data*/

#define HDCR_COMMAND		0x01c40194

/*3 (VP0 TSI_STCMPL0)*/
/*DSP->HOST: channel/buffer info*/
/*DSP writes (INDEX_MARKER | (chanId << 8) | bufId)
  (INDEX_MARKER makes it always !=0)                 */
/*host resets it to 0 after having read the data */
#define HDCR_CNANNEL_BUFFER	0x01c4019c
/*4 (VD_FRMSZ0)*/
/*DSP->HOST: thread status*/
/*lower 5 bits are used to signal thread state
  (4 capture threads + PCI thread)               */
/*upper bits used to pass address of sysData area*/
#define HDCR_THREAD_STATUS	0x01C40258
/*the following bits signal thread is running*/
#define REG3_CAPTURE0		0x1
#define REG3_CAPTURE1		0x2
#define REG3_CAPTURE2		0x4
#define REG3_CAPTURE3		0x8
#define REG3_PCI		0x10
/*REG3_UPNRUNNING: 4 capture threads + PCI thread up+running bits*/
#define REG3_UPNRUNNING		0x1f
/*REG3_DSP0: set only on DSP0 - the one handling crosspoint switch*/
#define REG3_DSP0		0x20

/*Register for passing "short command" argument*/
/*HOST->DSP: short command argument (VD_DEFVAL0, all bits operational)*/
#define HDCR_SC_ARG		0x01C4025c

/*Four registers are used to communicate buffer info between host and DSP:*/
/*(1 per channel)*/
/*each of lower 8 bits is set to 1 (by host)
  when corresponding buffer is available */
/*(max 8 buffers per channel)*/
/*Host resets a bit to 0 after receiving a buffer and sets to 1
  after releasing it.*/
/*This is done to avoid contentions (in case both host and DSP were
  modifying the register).*/
/*DSP picks the 1st available free buffer __other than the last one used__*/

#define DSP_BUFF_REG0		0x01c40220		/*register for chan 0*/
#define DSP_BUFF_REG1		(DSP_BUFF_REG0 + 4)	/*-"-               1*/
#define DSP_BUFF_REG2		(DSP_BUFF_REG0 + 8)	/*-"-               2*/
#define DSP_BUFF_REG3		(DSP_BUFF_REG0 + 12)	/*-"-               3*/

/*EMIF registers*/
#define EMIFA_GCTL              0x01800000
#define EMIFA_CE1               0x01800004
#define EMIFA_CE0               0x01800008
#define EMIFA_CE2               0x01800010
#define EMIFA_CE3               0x01800014
#define EMIFA_SDRAMCTL          0x01800018
#define EMIFA_SDRAMTIM          0x0180001c
#define EMIFA_SDRAMEXT          0x01800020
#define EMIFA_CE1SECCTL         0x01800044
#define EMIFA_CE0SECCTL         0x01800048
#define EMIFA_CE2SECCTL         0x01800050
#define EMIFA_CE3SECCTL         0x01800054

#define BAR0_SIZE               0x400000
#define BAR1_SIZE               0x800000


#define BAR2_SIZE               4
/*SYS_DSP_SDRAM_BASE: base address of SDRAM in DSPs address space*/
#define SYS_DSP_SDRAM_BASE	0x80000000

/*index marker (marks status data in DSP_COMM_REG2)*/
#define DSP_INDEX_MARKER 	0x88000000
/*max # of sections*/
#define MAX_SECTIONS		32
/*error reporting channels per DSP*/
#define SYS_ERROR_CHANS		5
/*maximum semaphore count (signals "buffer ready" condition to application)*/
#define SYS_MAX_SEM_COUNT	SYS_BUF_PER_CHAN
/*FIFO size (doesn't have to be exactly that)*/
#define SYS_FIFO_SIZE		(SYS_CHAN_PER_DSP * SYS_BUF_PER_CHAN)
#define SYS_PAGE		4096		/*page size*/
#define SYS_CS_BUF_SIZE		SYS_PAGE	/*status buffer size*/
/*interrupt thread exit time out, ms*/
#define SYS_WAIT_THREAD_TO	2000
#define SYS_NTSC_4CIF_HS	640		/*hor image size*/
#define SYS_NTSC_2CIF_HS	640		/*hor image size*/
#define SYS_NTSC_1CIF_HS	320		/*hor image size*/
#define SYS_PAL_4CIF_HS		704		/*hor image size*/
#define SYS_PAL_2CIF_HS		704		/*hor image size*/
#define SYS_PAL_1CIF_HS		352		/*hor image size*/

#define SYS_NTSC_4CIF_VS	480		/*vert image size*/
#define SYS_NTSC_2CIF_VS	240		/*vert image size*/
#define SYS_NTSC_1CIF_VS	240		/*vert image size*/
#define SYS_PAL_4CIF_VS		576		/*vert image size*/
#define SYS_PAL_2CIF_VS		288		/*vert image size*/
#define SYS_PAL_1CIF_VS		288		/*vert image size*/
//#define SYS_BMP_BUF_SIZE	(SYS_PAL_4CIF_HS * SYS_PAL_4CIF_VS)
#define PCIE_VENDORID_PLX	PCI_VENDOR_ID_PLX /*0x10b5*/
#define PCIE_DEVICEID_PLX	0x8111

#define SYS_PCIE_BAR0_BYTES	0x00010000	/*size of base addr 0 space */

/*For now, may change to Sensoray values */
#define PCIE_VENDORID		PCIE_VENDORID_PLX
#define PCIE_DEVICEID		PCIE_DEVICEID_PLX
#define PCI_VENDORID_DM642	PCI_VENDOR_ID_TI  /*0x104c*/
#define PCI_DEVICEID_DM642	0x9065

#define PCI_SUBSYSTEM_VENDORID	0x6000		/*Sensoray*/
#define PCI_SUBSYSTEM_817	0x0817		/*SubID for 817*/
#define PCI_SUBSYSTEM_817_UP	0x0000		/*SubID for 817 - temporary, unprog'd*/

/*size of DM642 base addr 0 space (DSP memory window)*/
#define SYS_DSP_BAR0_BYTES	0x400000
/*size of DM642 base addr 1 space (DSP mem mapped registers)*/
#define SYS_DSP_BAR1_BYTES	0x800000
/*base address of SDRAM in DSPs address space*/
#define SYS_DSP_SDRAM_BASE	0x80000000

/*DSP firmware start timeout: check every SYS_DSP_START_SLEEP ms */
/*sleep for that much between checks*/
#define SYS_DSP_START_SLEEP	100
/*total DSP start timeout*/
/* 4.5 seconds */
#define SYS_DSP_START_TO	(45 * SYS_DSP_START_SLEEP)

/*HOST->DSP command response timeout*/
/*must be > DSP PCI read timeout*/
/*sleep for that much between checks*/
#define SYS_DSP_CMD_SLEEP	1
/*total DSP interrupt response timeout*/
#define SYS_DSP_CMD_TO		(2000 * SYS_DSP_CMD_SLEEP)
/*offset for error codes passed via REG0*/
#define REG0_ERR_BASE		500
/*command marker*/
#define DSP_CMD_MARKER		0x081700a0
/*index marker (marks status data in DSP_COMM_REG2)*/
#define DSP_INDEX_MARKER 	0x88000000
/*PCI address of buffers on host side is on 4K (page) boundary
  => lower 12 bits are always 0s*/
/*those are used for chanId (bits 3:0) and command code (bits 11:4)*/
/*bits used for command (0xff0) & channel # (0xf)*/
#define DSP_SHORT_CMD_MSK	0x00000fff
/*pseudo channel number for commands addressed to any chan*/
#define SYS_CHAN_BROADCAST	0x8

/*command codes (host to DSP)*/
enum DSP_CMDCODE {
	/*jpg buffer pagetable:arg0=buffer index,channel=channel*/
	DSP_CMD_BUFFER_PT_JPG = 1,
	/*bmp buffer pagetable:arg0=buffer index,channel=channel*/
	DSP_CMD_BUFFER_PT_BMP,
	/*status buffer page table:arg0=buffer index,channel=channel*/
	DSP_CMD_BUFFER_PT_STAT,
	/*set video mode: channel=chanId(0..3),
	  arg0=1/0(video port restart/no restart)*/
	DSP_CMD_VMODE,
	/*set video mode fast: channel=chanId(0..3),
	  arg0=1/0(video port restart/no restart)*/
	DSP_CMD_VMODE_FAST,
	/*set capture mode*/
	DSP_CMD_CMODE,
	/*get channel's tick: channel=chanId(0..3)*/
	DSP_CMD_GETTICK,
	/*set crosspoint switch (short command)*/
	DSP_CMD_SET_VSWITCH,
	/*video out enable/disable (short command)*/
	DSP_SET_VOUT,
	DSP_SET_DATETIME,
	DSP_SET_OSDMODE,
	DSP_DEBUG_CMD
};

/* mode change mask bits (for SET_VMODE_FAST command):
   a  set bit means this mode element needs to be changed */
#define MODE_CHG_BRIGHTNESS     0x00000001
#define MODE_CHG_CONTRAST       0x00000002
#define MODE_CHG_SATURATION     0x00000004
#define MODE_CHG_HUE            0x00000008

/*mode change mask bits (for SET_CMODE_FAST command):
  a set bit means this mode element needs to be changed */
#define MODE_CHG_COMPQUALITY    0x00000010
#define MODE_CHG_FRAMERATE      0x00000020
#define MODE_CHG_CAPMODE        0x00000040
#define MODE_CHG_CAPTYPE        0x00000080
#define MODE_CHG_BMP_FIELD      0x00000100


#ifdef DEBUG_MEM
static int kallocs = 0;
static int vallocs = 0;
static int total_vm =0 ;

static void * KZALLOC(int size, int type, char *tag)
{
	void *val;
	val = kzalloc(size,type);
	printk("[%p] kzalloc %s, %d %d\n", val, tag, size, ++kallocs);
	return val;
}

static void KFREE(void *pmem, char *tag)
{
	printk("[%p] kzalloc free %s, %d\n", pmem, tag, --kallocs);
	kfree(pmem);
	return;
}


static void * VMALLOC32(int size, char *tag)
{
	void *val;
	val = vmalloc_32(size);
	printk("[%p] vmalloc %s, %d %d\n", val, tag, size, ++vallocs);
	return val;
}

static void VFREE(void *pmem, char *tag)
{
	printk("[%p] vmalloc free %s, %d\n", pmem, tag, --vallocs);
	vfree(pmem);
	return;
}
#else

#define VFREE(_x_, _y_) vfree(_x_)
#define VMALLOC32(_x_, _y_) vmalloc_32(_x_)
#define KFREE(_x_, _y_) kfree(_x_)
#define KZALLOC(_x_, _y_, _z_) kzalloc(_x_, _y_)

#endif

struct dsp_cmd_hdr {
	int			marker;		/*DSP_CMD_MARKER*/
	enum DSP_CMDCODE	cmdCode;	/*command code*/
	int			channel;	/*channel (if applicable)*/
	int			arg0;		/*argument*/
	int			arg1;		/*argument*/
	int			dataSize;	/*data size to follow (bytes)*/
};


struct sx_page_table {
	int			*pPT;  /*pointer to page table array*/
	int			PTsize;	      /*page table size*/
};



#define SX_DMA_PAGES 512


struct sx_dma_page {
	int           dir;
	dma_addr_t    dmah;
	unsigned long pPhysicalAddr;
	int           dwBytes;
	struct page   *pg;
};


struct sx_dma {
	void    *pUserAddr;                    /* begining of buffer*/
	s32     dwTransAddr;                   /* for internal use*/
	s32     dwBytes;                       /* size of buffer*/
	s32     dwOptions;                     /* allocation options*/
	s32     dwPages;                       /* number of pages in buffer*/
	struct sx_dma_page Page[SX_DMA_PAGES];
	int       state;
	int       direction;

	int       priv_bChan;
	int       priv_bufId;
	int       priv_type; //0 -command, CAP_JPG, CAP_BMP
	int       priv_bStat; //1 status, 0 image
	struct s817_dev *priv_dev;
};



/*Internal buffer struct*/
struct int_buffer {
	char                    *pData; // user address
	unsigned long            boff;
	struct sx_page_table	pageTbl;	/*page table for BUFFER*/
	struct sx_dma		sxDma;		/*SxDriver specific*/
	int                     map_count;
};


/*The following are SX_Driver dependent*/
struct system_buffer {
	struct int_buffer image;  // jpg or bmp
	struct int_buffer status; // capture status 
#ifdef S817_V4L2
	int               queued;
        int               seqno;  // monotonically increasing sequence num
	struct timeval	  timestamp;	// time that frame was received
#endif
};

/*see DSP code (must match) */
struct dsp_status {
	int			videopresent;
	int			valid;
	int			jpgsize;
	int			tick;
};

#define _SLF_
#ifdef _SLF_
#define FIRMWARE_FILE_NAME	"fw817.slf"
#define MAX_SECTIONS		32		/*max # of sections*/


struct sectiondata {
	/*section's load address*/
	__u32			startAddr;
	/*size, bytes*/
	__u32			size;
	/*padding, bytes (total size must be a mult of 4)*/
	__u32			padding;
};

struct slfheader {
	/*a const marker (start of .slf header)*/
	__u32			slfMapHdr;
	/*number of loaded sections*/
	__u32			nSect;
	/*array of section data records */
	struct sectiondata	sectData[MAX_SECTIONS];
};

#else
#define FIRMWARE_FILE_NAME	"f817.bin"
#endif


/*index for PCI thread error reporting (instead of chanId)*/
#define	EXTERR_INDEX_PCI	4
/*timeout (ms) for host to respond to interrupt */
#define SYS_HOST_TIMEOUT	100
/*DAT_open error in main()*/
#define EXTERR_DAT_OPEN		10
/*video capture channel open error*/
#define EXTERR_CAPCHAN_OPEN	20
/*FVID_alloc error, status passed in reserved[0]*/
#define EXTERR_FVID_ALLOC	30
/*invalid command, status in reserved[0]*/
#define EXTERR_INVALID_CMD	40
/*illegal command code, code in reserved[0]*/
#define EXTERR_ILLEGAL_CMD	50
/*PCI request semaphore pend error, status in res[0], fcount in [res[1]*/
#define EXTERR_PCI_PEND		60
/*Empty queue error, status in res[0], fcount in [res[1]*/
#define EXTERR_EMPTY_PCI_Q	70
/*GIO read error; status in res[0]*/
#define EXTERR_GIO_READ		80
/*GIO write error; status in res[0]*/
#define EXTERR_GIO_WRITE	90
/*illegal PCI request*/
#define EXTERR_PCI_ILLEGAL	100
/*timeout on waiting for host to clear DSP_COMM_REG2*/
#define EXTERR_PCI_HOST_TO	110
/*error writing to TVP; status in reserved[0]*/
#define EXTERR_TVP_CTL		120
/*debug: NB: some bits don't set to 1: 0x8FFF8FFF*/
#define DSP_COMM_REG_DBG0	0x01C44220


#define MASK_CAP_JPG    0x0000ffff
#define MASK_CAP_BMP    0xffff0000
#define MASK_ACHAN      0x0000001f

//MACRO: data ready
//TRUE if dataMask shows data ready for at least one channel
#define DATA_READY(dataMask) \
                (((capType) & CAP_BMP && (dataMask) & MASK_CAP_BMP) ||\
                ((capType) & CAP_JPG && (dataMask) & MASK_CAP_JPG))


//~~dbg
struct buffer_op {
	unsigned op		: 2;
	unsigned dsp	: 4;
	unsigned dChan	: 4;
	unsigned bufId	: 4;
	unsigned type	: 2;
};





static int debug;
static int *s817_debug = &debug;


#define dprintk(level, fmt, arg...)				\
	do {							\
		if (*s817_debug >= (level)) {			\
			printk(KERN_INFO "s817: " fmt, ##arg);	\
		}						\
	} while (0)





static int num_boards;

union fifoelement {
	int		bufId;		/*buffer index*/
};


/*buffer status registers (4 per DSP)*/
static int bufStatReg[4] = {DSP_BUFF_REG0, DSP_BUFF_REG1,
			    DSP_BUFF_REG2, DSP_BUFF_REG3};

/* FIFO structure*/
//Note: actual buffering is size - 1:
//FIFO is full when there's size - 1 entries.
//With 2 pointer approach there's no way to tell between
//empty and full otherwise.

struct fifo {
	int read;  //read pointer
	int write; //write pointer
	int size;  //actual size = requested size + 1
	union fifoelement *elements;  //buffer stores pointers
};

struct s817_dev;

struct s817_v4l2 {
	int                     users;
	int			bChan;		// channel (0-15) on the board
	int		 	numbufs;	// actual # of buffers 
						//     (max SYS_BUF_PER_CHAN)
#ifdef S817_V4L2
	char			firstq;		// set when 1st buffer is queued
	struct s817_dsp		*dsp;		// ptr back to DSP info for chan
	struct s817_dev		*dev;		// ptr back to dev info
#if LINUX_VERSION_CODE >= MUTEX_INTRODUCED
	struct mutex            lock;
#else
	struct semaphore        lock;
#endif
	wait_queue_head_t	frame_wait_queue; // used to wait for a frame
	unsigned int		nextseqno;	  // seqno for next frame
	int 			streaming;		// streaming is active
	struct mutex 		mutex;

	struct video_device	*v4l_device;	// device reg'd for V4L support
	struct proc_dir_entry	*channelroot, *osdroot;
#endif
	int                     reserved;
};

struct s817_output {
	unsigned int input; // input sel (0-15) for the output
	struct s817_dev *dev; //back pointer to device
};

struct s817_dsp {
	/*channel open flag != 0 for opened channels */
	int                     cOpen[SYS_CHAN_PER_DSP];
#ifdef S817_V4L2
	struct s817_v4l2        channel[SYS_CHAN_PER_DSP]; // for V4L2 only
#endif
	unsigned long           phys_bar0;
	void                    *log_bar0;
	unsigned long           phys_bar1;
	void                    *log_bar1;
	int                     intEnabled;
	int                     upAndRunning;
	int                     threadStat;
	/*0-3; resistors on board (GPO1-2)*/
	int			chipID;
	SYS_STATUS              sysData;
	/*buffer to send commands to DSP capture chan*/
	struct int_buffer       *cmd;//[SYS_CHAN_PER_DSP];
	spinlock_t              slock[SYS_CHAN_PER_DSP];
#if LINUX_VERSION_CODE >= MUTEX_INTRODUCED
	struct mutex            dsplock;
#else
	struct semaphore        dsplock;
#endif

	/* DSP device */
	struct pci_dev          *pdev;
	/* back pointer to device */
	struct s817_dev         *dev;
	int dsp_num;
	/* proc mem information(for debug and diagnostics */
	long fifoFull;
	long fifoWriteErr;
	long num_dpcs[SYS_CHAN_PER_DSP];
	long retries;
	long intCounter;
	long intCounterShared;
	long intCounterNotLoaded;
	long intCounterNotOurs;
	long intCounterNull;
	long intCounterLogBar;
	long last_dsp_err; /* from senddspcmd*/
	long dsp_errs;
	long timeouts;
	long nomarker;
};

/* device structure.  1 bridge and 4 PCI bridges */
struct s817_dev {
	struct s817_dsp       dsp[SYS_DSP_PER_BOARD];
	struct s817_output    output[SYS_OUT_PER_BOARD];
	int                   info;
	int                   minor;
	MODE                  mode[SYS_CHAN_PER_BOARD];
	OSD_MODE	      osdmode[SYS_CHAN_PER_BOARD];

	/* list of boards */
	struct list_head      s817_devlist;
	/* bridge PLX */
	struct pci_dev        *pdev;
        /* sysbuf [SYS_CHAN_PER_BOARD][SYS_BUF_PER_CHAN][2]  // */
	// DO NOT move these system buffers to s817_v4l2 or "per channel" structure
	// Code is maintained between windows driver which defines them like this.
	struct system_buffer  **sysBufJ;
	struct system_buffer  **sysBufB;

	void                  *dbgVA_J[SYS_CHAN_PER_BOARD][SYS_BUF_PER_CHAN][2];
	void                  *dbgVA_B[SYS_CHAN_PER_BOARD][SYS_BUF_PER_CHAN][2];
	struct page           *dbgPg_J[SYS_CHAN_PER_BOARD][SYS_BUF_PER_CHAN][2];
	struct page           *dbgPg_B[SYS_CHAN_PER_BOARD][SYS_BUF_PER_CHAN][2];

	/*buffer data FIFOs (1 per channel)*/
	struct fifo	      bufDataFIFO[SYS_CHAN_PER_BOARD][2]; /*buffers (CAP_JPG, CAP_BMP) */
	wait_queue_head_t     wq_done;	/*[SYS_CHAN_PER_BOARD];*/
	int                   dsp0;
	int                   lastChan; //last value returned by S817_BlockDone
	int                   dsp_loaded[SYS_DSP_PER_BOARD];
#if LINUX_VERSION_CODE >= MUTEX_INTRODUCED
	struct mutex          open_lock;
#else
	struct semaphore      open_lock;
#endif
	const struct firmware  *fw;
	int irq_tot;
	int irq_count[SYS_DSP_PER_BOARD];
	int handles;
	long err_physaddr;
#ifdef S817_V4L2
	int capType[SYS_CHAN_PER_BOARD];
#endif
};

static LIST_HEAD(s817_devlist);
static int G_offsets = 1;

static struct pci_device_id s817_pci_ids[] = {
	{
		.vendor = 0x10b5,
		.device = 0x8111,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID
	}, {
		.vendor = 0x10e3,
		.device = 0x8111,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID
	}, {
		.vendor = 0x10b5,
		.device = 0x8112,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID
	}, {
	}
};
MODULE_DEVICE_TABLE(pci, s817_pci_ids);


static int major = DFLT_MAJOR;
static int fwmode = 1;
module_param(major, int, 0444);
module_param(debug, int, 0);
module_param(fwmode, int, 0);
MODULE_DESCRIPTION("s817 driver");
MODULE_PARM_DESC(debug, "Debug level(0-100) default 0");
MODULE_PARM_DESC(fwmode, "Internal use only(default 1)");
MODULE_LICENSE("GPL");


static void s817_release_dsp(struct s817_dsp *dsp);
#define s817_readDSPreg(baseReg, regAddr)  readl((volatile long *)((char *) baseReg + (regAddr - DSP_REG_BASE)))

#define s817_writeDSPreg(baseReg, regAddr, data) writel(data, (volatile long *) ((char *) baseReg + (regAddr - DSP_REG_BASE)))

static void translateCMode(CAPTURE_MODE *inMode, CAPTURE_MODE *outMode);
static int sendDspCmd(struct s817_dev *dev, int dsp_idx, unsigned long physAddr,
		      unsigned long shortArg);
static int s817_openchan(struct s817_dev *dev, int aChan, MODE *pMode);

ECODE MakePageTable(struct int_buffer *pIntBuf, int size);
static int CloseChannel(struct s817_dev *dev,  int bChan, int bTimeout);
static int S817_SetCaptureMode(struct s817_dev *dev, int aChan,
			       CAPTURE_MODE *pCMode);
static int S817_SetVideoMode(struct s817_dev *dev, int aChan,
			     VIDEO_MODE *pVMode);
static int S817_SetOsdMode(struct s817_dev *dev, int aChan, OSD_MODE *osd);
static int S817_SetDateTime(struct s817_dev *dev, struct param_settime *p);
static int S817_SetDateTimeDsp(struct s817_dsp *dsp, struct param_settime *strTm);
static void releaseBufs(struct s817_dev *dev, int dsp, int dChan, int mask, int capType);
static void claimBufs(struct s817_dev *dev, int dsp, int dChan, int mask, int capType);
static int getDataMask(struct s817_dev *dev);
static int roundRobin (struct s817_dev *dev, int dataMask);
static int S817_SetVideoOut(struct s817_dev *dev, unsigned int outEn);
static int S817_SetVideoSwitch(struct s817_dev *dev, int *switchOut);
static int GetBmpWH(MODE *pMode, int *w, int *h);
/*FIFO functions' prototypes */
static int  FIFOcreate(struct fifo *pFIFO, long size);
static void FIFOdestroy(struct fifo *pFIFO);
static union fifoelement FIFOread(struct fifo *pFIFO);
static int FIFOwrite(struct fifo *pFIFO, union fifoelement *pElement);
static int isNotEmpty(struct fifo *pFIFO);
static int isFull(struct fifo *pFIFO);
static int GetSysData(struct s817_dsp *dsp);
//static int CheckBuffers(struct s817_dsp *dsp);

static void s817_dpc(unsigned long dev_id);


static struct tasklet_struct s817_tasklet[SYS_DSP_PER_BOARD][MAX_BOARDS];
static void FreeSysBuff(struct pci_dev *pdev, struct system_buffer *pSysBuf);
static ECODE AllocSysBuff(struct pci_dev *pdev,
			  struct system_buffer *pSysBuf, int capType);
static int free_all_channel_mem(struct s817_dev *dev);
static ECODE AllocCommandBuff(struct pci_dev *pdev, struct int_buffer *pIntBuf);
static void FreeCommandBuff(struct pci_dev *pdev, struct int_buffer *pIntBuf);
#if LINUX_VERSION_CODE >= CREATE_PROC_REMOVED
// handled by seq_read
#else
static int s817_read_procmem(char *buf, char **start, off_t offset, int count,
		      int *eof, void *data);
#endif
/*end FIFO functions*/

/* begin V4L2 functions */
#ifdef S817_V4L2
static int s817_init_v4l2(struct s817_v4l2 *channel, unsigned int number);
static struct proc_dir_entry *s817_proc_root = NULL;
static struct proc_dir_entry *s817_procout_root = NULL;
static int read_fwver(char *page, char **start, off_t off, int count,
		      int *eof, void *data);
void create_chan_proc_entries(struct s817_v4l2 *channel);
void remove_chan_proc_entries(struct s817_v4l2 *channel);
static int read_output(char *page, char **start, off_t off, int count,
		       int *eof, void *data);
static int write_output(struct file *file, const char __user *buf,
			unsigned long count, void *data);
#endif


/* end V4L2 functions */
//returns 1 if malloc failed
int FIFOcreate (struct fifo *pFIFO, long size)
{
	//pFIFO - pointer to FIFO structure;
	//size - FIFO size (elements);
	pFIFO->elements = (union fifoelement *) KZALLOC((size + 1) * sizeof(union fifoelement), GFP_KERNEL, "fifo start");
	
	if (pFIFO->elements == NULL) {
		return 1;
	}
	pFIFO->read = 0;
	pFIFO->write = 0;
	pFIFO->size = size + 1;
	return 0;
}

void FIFOdestroy (struct fifo *pFIFO)
{
	//pFIFO - pointer to FIFO structure;
	if (pFIFO->elements) {
		KFREE(pFIFO->elements, "fifo start");
		pFIFO->elements = NULL;
	}
	pFIFO->read = 0;
	pFIFO->write = 0;
	pFIFO->size = 0;


	return;
}

// do not change this code for V4L2
//The caller must make sure the FIFO is not empty before calling
//There's no other way to tell if the returned value is valid
union fifoelement FIFOread (struct fifo *pFIFO)
{
	union fifoelement		fe;
	if (pFIFO == NULL) {
		printk("null fifo\n");
	}
	if (pFIFO->elements == NULL) {
		printk("null fifo elements\n");
	}
	fe = pFIFO->elements[pFIFO->read];
	pFIFO->read = (pFIFO->read + 1) % pFIFO->size;

	return fe;
}

// do not change this code for V4L2
static int claimFIFOwrite(struct s817_dsp *dsp, int aChan, int id, int capType)
{
	int bChan = aChan - 1;
	int dChan = bChan % SYS_CHAN_PER_DSP;
	int rv;
	unsigned long flags;
	spin_lock_irqsave(&dsp->slock[dChan], flags);
	rv = FIFOwrite (&dsp->dev->bufDataFIFO[bChan][capType - 1], (union fifoelement *) &id);
	if (rv == 0) {
		claimBufs (dsp->dev, dsp->dsp_num, dChan, (1 << id), capType);
	}
	spin_unlock_irqrestore(&dsp->slock[dChan], flags);
	return rv;
}


int FIFOwrite (struct fifo *pFIFO, union fifoelement *pElement)
{
//	If FIFO is full, FIFOwrite does nothing and returns 1
	int			rv = 1;
	int			nextWrite = (pFIFO->write + 1) % pFIFO->size;
	
	if (nextWrite != pFIFO->read) {           //if not full
		pFIFO->elements[pFIFO->write] = *pElement;
		pFIFO->write = nextWrite;
		rv = 0;
	}

	return rv;
}

int isNotEmpty (struct fifo *pFIFO)
{
	return (pFIFO->read != pFIFO->write);
}

int isFull (struct fifo *pFIFO)
{
	int         nextWrite = (pFIFO->write + 1) % pFIFO->size;
	return (pFIFO->read == nextWrite);
}



static void s817_dpc(unsigned long dev_id)
{
	struct s817_dsp *dsp = (struct s817_dsp *) dev_id;
	int dsp_num;
	int board = 0;
	long lBuf;
	/*channel in application context (1..16)*/
	int aChan;
	/*channel in board context (0..15)*/
	int bChan;
	/*channel in DSP context (0..3)*/
	int dChan;
	/*buffer Id (0..SYS_BUF_PER_CHAN - 1)*/
	int bufIdJpg; //jpg buffer id(0..SYS_BUF_PER_CHAN - 1)
	int bufIdBmp; //bmp buffer id(0..SYS_BUF_PER_CHAN - 1)
	int capType; // capture buffer type (bmp or jpg or BOTH)
	/*buffer stored in FIFO (can set event now)*/
	int fifoOKBmp = 0;
	int fifoOKJpg = 0;
#ifdef S817_V4L2
	struct s817_v4l2 *channel;
#endif
	dsp_num = dsp->dsp_num;
	board = dsp->dev->minor;

	lBuf = s817_readDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER);
	if ((lBuf & DSP_INDEX_MARKER) != DSP_INDEX_MARKER) {
		dsp->nomarker++;
		return;
	}

	dsp->dev->irq_tot++;
	dsp->dev->irq_count[dsp_num]++;

	/*channel index (DSP context, 0..3)*/
	dChan = lBuf & 7;

	if ((dChan >= SYS_CHAN_PER_DSP) || (dChan < 0)) {
		//KdPrint(("dChan invalid %x\n", dChan));
		s817_writeDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER, 0);
		(void) s817_readDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER);
		printk(KERN_INFO "DPC: invalid channel index\n");
		return;
	}

	capType = (lBuf >> 12) & 3;	//buffer type: jpg or bmp
	bufIdJpg = (lBuf >> 4) & 7;	//jpg buffer Id
	bufIdBmp = (lBuf >> 8) & 7;	//bmp buffer Id
	bChan = dsp->dsp_num * SYS_CHAN_PER_DSP + dChan; //channel in board context
	aChan = bChan + 1;
	//channel in app context
#ifdef S817_V4L2
	channel = &dsp->channel[dChan];
#endif
	dsp->num_dpcs[dChan]++;
	dprintk(125, "dChan %x\n", dChan);
	/*  if channel not open, return.
	 *  Note: CloseChannel function will disable tasklets which waits
	 *  for all tasklets to exit before tearing down resources
	 */
	if (!dsp->dev->dsp[dsp->dsp_num].cOpen[dChan]) {
		s817_writeDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER, 0);
		(void ) s817_readDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER);
		printk(KERN_INFO "DPC: channel not open\n");
		return;
	}

	if (((capType & CAP_BMP) && bufIdBmp >= SYS_BUF_PER_CHAN)
	||  ((capType & CAP_JPG) && bufIdJpg >= SYS_BUF_PER_CHAN)) {
		printk(KERN_INFO "invalid sys buf %lx %d %d\n", lBuf,
		       bufIdBmp, bufIdJpg);
		dprintk(4, "invalid sys buf \n");
		s817_writeDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER, 0);
		(void ) s817_readDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER);
		return;
	}
	if (bChan < 0 || bChan >= SYS_CHAN_PER_BOARD) {
		dprintk(4, "invalid channel \n");
		printk(KERN_INFO "invalid channel\n");
		s817_writeDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER, 0);
		(void ) s817_readDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER);
		return;
	}
	//store data for required types (bmp or jpg or both)
	if (capType & CAP_BMP) {
		//insert board Id
		*((int *) dsp->dev->sysBufB[bChan][bufIdBmp].status.pData) = board;
#ifdef S817_V4L2
		dsp->dev->sysBufB[bChan][bufIdBmp].seqno = dsp->channel[bChan].nextseqno++;
		do_gettimeofday(&dsp->dev->sysBufB[bChan][bufIdBmp].timestamp);
#endif
		//Normally FIFO size should match the number of buffers, so FIFO should not be full
		//In case of size mismatch check here
		if (!isFull (&dsp->dev->bufDataFIFO[bChan][CAP_BMP - 1])) {					//if bmp FIFO not full
			if (claimFIFOwrite(dsp, aChan, bufIdBmp, CAP_BMP)) {
				++dsp->fifoWriteErr;
			} else {
				fifoOKBmp = 1;
			}
		} else {
			++dsp->fifoFull;
		}
	}
	
	if (capType & CAP_JPG) {
		//insert board Id
		*((int *) dsp->dev->sysBufJ[bChan][bufIdJpg].status.pData) = board;
#ifdef S817_V4L2
		dsp->dev->sysBufJ[bChan][bufIdBmp].seqno = dsp->channel[bChan].nextseqno++;
		do_gettimeofday(&dsp->dev->sysBufJ[bChan][bufIdBmp].timestamp);
#endif
		//Normally FIFO size should match the number of buffers, so FIFO should not be full
		//In case of size mismatch check here
		if (!isFull (&dsp->dev->bufDataFIFO[bChan][CAP_JPG - 1])) {	
			if (claimFIFOwrite(dsp, aChan, bufIdJpg, CAP_JPG)) {
				++dsp->fifoWriteErr;
			} else {
				fifoOKJpg = 1;
			}
		} else {
			++dsp->fifoFull;
		}
	}
	if (fifoOKJpg || fifoOKBmp) {
		/* set event */
		wake_up(&dsp->dev->wq_done);
#ifdef S817_V4L2
		wake_up(&channel->frame_wait_queue); /* set event */
#endif
	}
	/*signal DSP we are ready to accept interrupts
	  by resetting HDCR_CNANNEL_BUFFER*/
	s817_writeDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER, 0);
	(void ) s817_readDSPreg(dsp->log_bar1, HDCR_CNANNEL_BUFFER);
	return;
}

#if LINUX_VERSION_CODE < VERSION_PTREG_REMOVED_IRQ //KERNEL_VERSION(2, 6, 19)
irqreturn_t s817_irq(int irq, void *dev_id, struct pt_regs *ptreg)
#else
irqreturn_t s817_irq(int irq, void *dev_id)
#endif
{
	struct s817_dsp *dsp = (struct s817_dsp *) dev_id;
	int dsp_num = -1;
	/*countd "FIFO full" occurencies
	  counts "illegal data in DSP_COMM_REG2" occurencies*/
	/*counts interrupts*/

	int reg;
	/*process interrupt*/
	++dsp->intCounterShared;

	if (dsp == NULL) {
		++dsp->intCounterNull;
		return IRQ_NONE;
	}


	if (dsp->log_bar1 == NULL) {
		++dsp->intCounterLogBar;
		return IRQ_NONE;
	}

	reg = s817_readDSPreg(dsp->log_bar1, DSP_HSR);

	if ((reg & DSP_HSR_INTSRC ) == 0) {
		++dsp->intCounterNotOurs;
		return IRQ_NONE;	
	}

	++dsp->intCounter;
	if (dsp->dev->dsp_loaded[dsp->dsp_num] == 0) {
		reg &= ~DSP_HSR_INTAM;
		reg |= DSP_HSR_INTSRC;
		s817_writeDSPreg(dsp->log_bar1, DSP_HSR, reg);
		(void) s817_readDSPreg(dsp->log_bar1, DSP_HSR);
		++dsp->intCounterNotLoaded;
		return IRQ_HANDLED;
	}


	/* get status */
	dsp_num = dsp->dsp_num;

	if (dsp_num < SYS_DSP_PER_BOARD)
		tasklet_hi_schedule(&s817_tasklet[dsp_num][dsp->dev->minor]);

	/* our IRQ, mark it as handled */
	reg &= ~DSP_HSR_INTAM;
	reg |= DSP_HSR_INTSRC;
	s817_writeDSPreg(dsp->log_bar1, DSP_HSR, reg);
	/* read back to flush the bridge cache
	   see Documentation/io_ordering.txt */
	s817_readDSPreg(dsp->log_bar1, DSP_HSR);

	return IRQ_HANDLED;
}


/*S817_GetBuffer*/
/*gets CAP_BUFFER (pointers to buffer data)*/
/*the oldest of captured buffers for a given channel*/
ECODE S817_GetBufferId(struct s817_dev *dev, int aChan, CAP_BUFFER *cBuf, int capType)
{
	union fifoelement fe;
	int bChan = aChan - 1;
	int dsp  = bChan / SYS_DSP_PER_BOARD;
	int isNotEmp;

	if (dev == NULL)
		return -EFAULT;

	if (cBuf == NULL)
		return -EFAULT;

	if ((capType != CAP_JPG) && (capType != CAP_BMP)) {
		printk(KERN_ERR "bad buffer %d\n", capType);
		return ERR_ILLEGAL_BUFTYPE;
	}
	if ((bChan < 0) || (bChan >= SYS_CHAN_PER_BOARD) || !dev->dsp[dsp].pdev) {
		printk(KERN_ERR "s817: invalid channel\n");
		return -1;
	}

	isNotEmp = isNotEmpty(&dev->bufDataFIFO[bChan][capType - 1]);
	if (isNotEmp)
		fe = FIFOread(&dev->bufDataFIFO[bChan][capType - 1]);
	else {
		printk(KERN_INFO "s817: fifo empty\n");
		return ERR_FIFO_EMPTY;
	}
	cBuf->bufId = fe.bufId;
	GetBmpWH(&dev->mode[bChan], &cBuf->width, &cBuf->height);
	dprintk(125, "buffer id %d\n", fe.bufId);
	return 0;
}


/* get pointer to the buffer for specific bufId*/
ECODE S817_GetBuffer(struct s817_dev *dev, int aChan,
		     CAP_BUFFER *cBuf, int capType)
{
	int bChan = aChan - 1;
	int dsp  = bChan / SYS_DSP_PER_BOARD;
	int bufId;
	struct dsp_status *stat;
	bufId = cBuf->bufId;		  /*get bufferId of buffer*/
	if ((capType != CAP_JPG) && (capType != CAP_BMP))
		return ERR_ILLEGAL_BUFTYPE;

	if ((bChan < 0) || (bChan >= SYS_CHAN_PER_BOARD) || !dev->dsp[dsp].pdev) {
		printk(KERN_INFO "s817: invalid channel\n");
		return -1;
	}
	if ((bufId < 0) || (bufId >= SYS_BUF_PER_CHAN)) {
		printk(KERN_INFO "s817: invalid buf id\n");
		return -1;
	}
	dprintk(50, "retrieving kernel memory for bufId %d\n", bufId);
	/*get pointers. Note this are just offsets or mappings.  The user does not need the actual kernel pointer value*/
	if (capType == CAP_JPG) {
		cBuf->pData = (char *) dev->sysBufJ[bChan][bufId].image.boff;
		cBuf->pStat = (char *) dev->sysBufJ[bChan][bufId].status.boff;
	} else {
		cBuf->pData = (char *) dev->sysBufB[bChan][bufId].image.boff;
		cBuf->pStat = (char *) dev->sysBufB[bChan][bufId].status.boff;
	}


	GetBmpWH(&dev->mode[bChan], &cBuf->width, &cBuf->height);
	stat = (struct dsp_status *) cBuf->pStat;

	dprintk(25, "GetBuffer sysbuf[%p] %p,  %p %p %d %d\n", 
		dev,
		&dev->sysBufJ[bChan][bufId],
		cBuf->pData, cBuf->pStat,
		bChan, bufId);
	return 0;
}

ECODE S817_ReleaseBuffer(struct s817_dev *dev, int aChan, int bufId, int capType)
{
	int bChan = aChan - 1;
	int dChan = bChan % SYS_CHAN_PER_DSP;
	int dsp  = bChan / SYS_DSP_PER_BOARD;

	if (aChan < 1 || aChan > SYS_CHAN_PER_BOARD || !dev->dsp[dsp].pdev)
		return ERR_ILLEGAL_CHAN;
	if (bufId < 0 || bufId >= SYS_BUF_PER_CHAN)
		return ERR_ILLEGAL_BUF;

	releaseBufs(dev, dsp, dChan, 1 << bufId, capType); /*show buffer as available*/
	return 0;
}

static int getExtErr(struct s817_dev *dev, int dsp, int chanId,
		     EXTERR *extErrBuf)
{
	char *extErrMem;
	/* set page register (data is within first 4MB page of 0x80000000
	   space DSP_PAGE_SDRAMSTART) */
	if (dsp < 0 || dsp >= SYS_DSP_PER_BOARD || !dev->dsp[dsp].pdev)
		return ERR_ILLEGAL_CHAN;

	if (dev->dsp_loaded[dsp] == 0) {
		return -ENODEV;
	}
	if (GetSysData(&dev->dsp[dsp]) != 0)
		return -EINVAL;
	s817_writeDSPreg(dev->dsp[dsp].log_bar1, DSP_DSPP, DSP_PAGE_SDRAMSTART);

	if (dev->dsp[dsp].sysData.extErrAddr) {
		/* sanity check */
		extErrMem = (char *) dev->dsp[dsp].log_bar0 +
			dev->dsp[dsp].sysData.extErrAddr
			- SYS_DSP_SDRAM_BASE +
			sizeof(EXTERR) * SYS_EXTERR_BUF_SIZE * chanId;
		if (extErrBuf) {
			memcpy(extErrBuf, extErrMem,
			       sizeof(EXTERR) * SYS_EXTERR_BUF_SIZE);
		} else return -EINVAL;
	}
	return 0;
}

#if LINUX_VERSION_CODE < VERSION_UNLOCKED_IOCTL
static int s817_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		      unsigned long arg)
#else
static long s817_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#endif
{
	struct s817_dev *dev = (struct s817_dev *) file->private_data;
	void __user *argp = (void __user *) arg;
	int rc = 0;

	switch (cmd) {
	case S817_IOCTL_GETSTATUSINFO:
	{
		struct param_getstatusinfo p;
		struct param_getstatusinfo *pin =
			(struct param_getstatusinfo *) argp;
		int dsp;

		if (copy_from_user(&p, argp,
				   sizeof(struct param_getstatusinfo)))
			return -EINVAL;
		dsp = p.dsp;
		if (dsp < 0 || dsp >= SYS_DSP_PER_BOARD)
			return -EINVAL;

		if (dev->dsp[dsp].sysData.dspAddr == 0)
			return ERR_ADDR_NOT_CONFIG;
		

		if (GetSysData(&dev->dsp[dsp]) != 0) {
			printk(KERN_ERR "invalid dsp addr \n");
			return -EFAULT;
		}

		if (copy_to_user(&pin->sysStat, &dev->dsp[dsp].sysData,
				 sizeof(SYS_STATUS)))
			return -EINVAL;
		memcpy(&pin->irq_count[0],
		       &dev->irq_count[0], sizeof(int) *4);
		pin->irq_tot = dev->irq_tot;
		return 0;
	}
	case S817_IOCTL_GETEXTERRINFO:
	{
		static struct param_getexterr p;
		struct param_getexterr *pin = (struct param_getexterr *) argp;
		if (copy_from_user(&p, argp,
				   sizeof(struct param_getexterr)))
			return -EINVAL;

		rc = getExtErr(dev, p.dsp, p.errChan, p.exterr);

		if (rc != 0)
			return rc;

		if (copy_to_user(pin->exterr, p.exterr,
				 sizeof(EXTERR) * SYS_EXTERR_BUF_SIZE))
			return -EINVAL;

		return 0;
	}
	case S817_IOCTL_OPENCHANNEL:
	{
		struct param_openchan p;
		if (copy_from_user(&p, argp, sizeof(struct param_openchan)))
			return -EINVAL;
		dprintk(25, "dev %p dsp0 %p dsp1 %p dsp2 %p dsp3 %p\n", dev,
			&dev->dsp[0], &dev->dsp[1], &dev->dsp[2], &dev->dsp[3]);
		rc = s817_openchan(dev, p.aChan, &p.mode);
		dprintk(25, "open chan %d\n", rc);
		dprintk(25, "open channel %d res %d\n", p.aChan, rc);
		return rc;
	}
	case S817_IOCTL_CLOSECHANNEL:
	{
		struct param_closechan p;
		if (copy_from_user(&p, argp, sizeof(struct param_closechan)))
			return -EINVAL;
		dprintk(25, "close channel %d\n", p.aChan);
		rc = CloseChannel(dev, (p.aChan-1), 0);
		return rc;
	}
	case S817_IOCTL_SETVIDMODE:
	{
		struct param_setvidmode p;
		if (copy_from_user(&p, argp, sizeof(struct param_setvidmode)))
			return -EINVAL;
		rc = S817_SetVideoMode(dev, p.aChan, &p.mode);
		return rc;
	}
	case S817_IOCTL_SETCAPMODE:
	{
		struct param_setcapmode p;
		if (copy_from_user(&p, argp,
				   sizeof(struct param_setcapmode))) {
			return -EINVAL;
		}
		rc = S817_SetCaptureMode(dev, p.aChan, &p.mode);
		return rc;
	}
	case S817_IOCTL_GETBUFFERID:
	{
		struct param_getbuf p;
		if (copy_from_user(&p, argp, sizeof(struct param_getbuf)))
			return -EINVAL;
		rc = S817_GetBufferId(dev, p.aChan, &p.cBuf, p.capType);
		if (rc != 0)
			return -EFAULT;
		if (copy_to_user(argp, &p, sizeof(struct param_getbuf)))
			return -EINVAL;
		return 0;
	}
	case S817_IOCTL_GETBUFFER:
	{
		struct param_getbuf p;
		if (copy_from_user(&p, argp, sizeof(struct param_getbuf))) {
			printk(KERN_INFO "getbuffers failed\n");
			return -EINVAL;
		}
		dprintk(25, "get buffer %p\n", dev);
		rc = S817_GetBuffer(dev, p.aChan, &p.cBuf, p.capType);
		dprintk(25, "pstat %p (sizeof %lu)\n", p.cBuf.pStat, (unsigned long) sizeof(struct param_getbuf));
		if (copy_to_user(argp, &p, sizeof(struct param_getbuf)))
			return -EINVAL;
		return rc;
	}
	case S817_IOCTL_RELEASEBUF:
	{
		struct param_relbuf p;
		if (copy_from_user(&p, argp, sizeof(struct param_relbuf)))
			return -EINVAL;
		rc = S817_ReleaseBuffer(dev, p.aChan, p.bufId, p.capType);
		return rc;
	}
	case S817_IOCTL_SETDATETIME:
	{
		struct param_settime p;
		if (copy_from_user(&p, argp, sizeof(struct param_settime)))
			return -EINVAL;
		rc = S817_SetDateTime(dev, &p);
		return rc;
	}
	case S817_IOCTL_SETOSDMODE:
	{
		struct param_setosdmode p;
		if (copy_from_user(&p, argp, sizeof(struct param_setosdmode)))
			return -EINVAL;
		rc = S817_SetOsdMode(dev, p.aChan, &p.osdMode);
		return rc;
	}
	case S817_IOCTL_SETVIDOUT:
	{
		struct param_setvidout p;
		if (copy_from_user(&p, argp, sizeof(struct param_setvidout)))
			return -EINVAL;
		rc = S817_SetVideoOut(dev, p.outEn);
		return rc;
	}
	case S817_IOCTL_SETVIDSWITCH:
	{
		struct param_setvidswitch p;
		if (copy_from_user(&p, argp, sizeof(struct param_setvidswitch)))
			return -EINVAL;
		rc = S817_SetVideoSwitch(dev, p.switchOut);
		return rc;
	}
	case S817_IOCTL_GETDATAMASK:
	{
		struct param_getdatamask p;
		p.dm = getDataMask(dev);
		if (copy_to_user(argp, &p, sizeof(struct param_getdatamask)))
			return -EINVAL;
		return 0;
	}
	case S817_IOCTL_ROUNDROBIN:
	{
		int rr;
		struct param_roundrobin p;
		if (copy_from_user(&p, argp, sizeof(struct param_roundrobin)))
			return -EINVAL;
		if (p.dm == 0) {
			printk(KERN_ERR "invalid mask parameter\n");
			return -EINVAL;
		}
		rr = roundRobin(dev, p.dm);
		p.aChan = (rr & 0xffff);
		p.capType = ((rr >> 16) & 0xffff);
		dprintk(50, "rr result %d %d\n", p.aChan, p.capType);
		if (copy_to_user(argp, &p, sizeof(struct param_roundrobin)))
			return -EINVAL;
		return 0;
	}
	default:
		printk(KERN_INFO "s817_ioctl: unknown command\n");
		break;
	}
	return rc;
}

static int s817_open(struct inode *inode, struct file *file)
{
	struct s817_dev *dev = NULL;
	struct s817_dev *h;
	struct list_head *list;
	int minor = MINOR(inode->i_rdev);
	dprintk(4, "open %d\n", minor);
	list_for_each(list, &s817_devlist) {
		h = list_entry(list, struct s817_dev, s817_devlist);
		if (h->minor == minor) {
			dev = h;
			dprintk(50, "found board %d %p\n", minor, dev);
			break;
		}
	}

	if (dev == NULL) {
		return -ENODEV;
	}
	MUTEX_LOCK(&dev->open_lock);
	if (dev->handles >= 1) {
		MUTEX_UNLOCK(&dev->open_lock);
		printk(KERN_INFO "s817: one handle at a time\n");
		return -EBUSY;
	}
	dev->handles++;
	MUTEX_UNLOCK(&dev->open_lock);
	file->private_data = dev;
	return 0;
}


static int s817_release(struct inode *inode, struct file *file)
{
	struct s817_dev *dev = NULL;
	struct s817_dev *h;
	struct list_head *list;
	int j;
	int minor = MINOR(inode->i_rdev);
	dprintk(4, "release %d\n", minor);


	list_for_each(list, &s817_devlist) {
		h = list_entry(list, struct s817_dev, s817_devlist);
		if (h->minor == minor) {
			dev = h;
			dprintk(50, "found board %d %p\n", minor, dev);
			break;
		}
	}
	if (dev == NULL) {
		return -ENODEV;
	}
	MUTEX_LOCK(&dev->open_lock);
	if (dev->handles > 0) {
		dev->handles--;
	} else {
		MUTEX_UNLOCK(&dev->open_lock);
		printk(KERN_ERR "s817: invalid release\n");
		return -ENODEV;
	}

	for (j = 0; j < SYS_CHAN_PER_BOARD; j++) {
		CAPTURE_MODE mode = dev->mode[j].cMode;
		if (mode.capType != CAP_OFF) {
			mode.capType = CAP_OFF;
			S817_SetCaptureMode(dev, j+1, &mode);
		}
	}

	for (j = 0; j < SYS_CHAN_PER_BOARD; j++) {
		CloseChannel(dev, j, 0);
	}
	/* wait for dma to stop. plus one frame */
	msleep(50);

	/* Empty the FIFOs */
	for (j = 0; j < SYS_CHAN_PER_BOARD; j++) {
		while (isNotEmpty(&dev->bufDataFIFO[j][CAP_JPG - 1]))
			FIFOread(&dev->bufDataFIFO[j][CAP_JPG - 1]);
		while (isNotEmpty(&dev->bufDataFIFO[j][CAP_BMP - 1]))
			FIFOread(&dev->bufDataFIFO[j][CAP_BMP - 1]);
	}

	for (j = 0; j < SYS_DSP_PER_BOARD; j++) {
		if (!dev->dsp[j].pdev)
			continue;
		if (dev->dsp[j].intEnabled == 1) {
			int lBuf;
			lBuf = s817_readDSPreg(dev->dsp[j].log_bar1, DSP_HSR);
			/* disable the interrupt */
			lBuf |= DSP_HSR_INTAM;
			s817_writeDSPreg(dev->dsp[j].log_bar1, DSP_HSR, lBuf);
			(void) s817_readDSPreg(dev->dsp[j].log_bar1, DSP_HSR);
			dev->dsp[j].intEnabled = 0;
			
		}
		dev->dsp[j].upAndRunning = 0;
		dev->dsp_loaded[j] = 0;
	}

	file->private_data = NULL;
	MUTEX_UNLOCK(&dev->open_lock);
	return 0;
}

static void __devexit s817_remove(struct pci_dev *pdev)
{
	struct s817_dev *dev = pci_get_drvdata(pdev);
	struct s817_dev *h;
	unsigned int i;
	struct list_head *list;
	char proc[64];

	list_for_each(list, &s817_devlist) {
		h = list_entry(list, struct s817_dev, s817_devlist);
		if (h == dev) {
			sprintf(proc, "s817_%d", h->minor);
			remove_proc_entry(proc, NULL);
			list_del(list);
#ifdef S817_V4L2
			{
				struct s817_v4l2 *channel;
				int i, j;
				remove_proc_entry("fwVer", s817_proc_root);
				for (i = 0; i < SYS_DSP_PER_BOARD; i++) {
					for (j = 0; j < SYS_CHAN_PER_DSP; j++) {
						channel = &dev->dsp[i].channel[j];
						remove_chan_proc_entries(channel);
					}
				}
			}
			for (i = 0; i < SYS_OUT_PER_BOARD; i++) {
				char name[10];
				sprintf(name, "%d", 
					i+ SYS_OUT_PER_BOARD * num_boards);
				remove_proc_entry(name, s817_procout_root);
			}
			remove_proc_entry("output", s817_proc_root);
#endif
			break;
		}
	}

	release_firmware(dev->fw);

	for (i = 0; i < SYS_DSP_PER_BOARD; i++) {
#ifdef S817_V4L2
		int j;
		for (j=0; j < SYS_CHAN_PER_DSP; j++) {
			struct s817_v4l2  *channel = &dev->dsp[i].channel[j];
			if (channel->v4l_device != NULL)
			{
				video_unregister_device(channel->v4l_device);
				channel->v4l_device = NULL;
			} else {
				printk("HEY: v4l_device was NULL\n");
			}
			//s817_queue_free(channel);
		}
#endif
		dprintk(50, "release dsp %d\n", i);
		s817_release_dsp(&dev->dsp[i]);
	}

	dev->info = 0;
	free_all_channel_mem(dev);


	if (dev->sysBufB) {
		for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
			char tag[50];
			sprintf(tag, "sysBuf %d", i);
			KFREE(dev->sysBufB[i], tag);
			dev->sysBufB[i] = NULL;
		}
		KFREE(dev->sysBufB, "sysBuf");
		dev->sysBufB = NULL;
	}
	if (dev->sysBufJ) {
		for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
			char tag[50];
			sprintf(tag, "sysBuf %d", i);
			KFREE(dev->sysBufJ[i], tag);
			dev->sysBufJ[i] = NULL;
		}
		KFREE(dev->sysBufJ, "sysBuf");
		dev->sysBufJ = NULL;
	}
	for (i = 0; i < SYS_DSP_PER_BOARD; i++) {
		KFREE(dev->dsp[i].cmd, "dsp cmd");
	}
	KFREE(dev, "device");
	printk(KERN_INFO "%s\n", __func__);
}

static int attach_dsp(struct s817_dsp *dsp, struct pci_dev *pdev, int num, int num_boards)
{
	int rc;
	int i;
	dsp->pdev = pdev;


	for (i = 0; i < SYS_CHAN_PER_DSP; i++) {
		MUTEX_INIT(&dsp->dsplock);
		spin_lock_init(&dsp->slock[i]);
	}

	rc = pci_enable_device(pdev);

	if (rc != 0)
		return rc;

	pci_set_master(pdev);

	if (pci_set_dma_mask(pdev, DMA_32BIT_MASK)) {
		printk(KERN_ERR "could not set 32bit DMA mask\n");
		return -ENODEV;
	}

	dsp->phys_bar0 = pci_resource_start(pdev, 0);
	dprintk(4, "phys BAR0: %lx\n", dsp->phys_bar0);
	dsp->phys_bar1 = pci_resource_start(pdev, 1);
	dprintk(4, "phys BAR1: %lx\n", dsp->phys_bar1);
	dsp->log_bar0 = ioremap(dsp->phys_bar0, BAR0_SIZE);
	if (dsp->log_bar0 == NULL) {
		printk(KERN_ERR "couldn't get logical base address 0\n");
		return -ENODEV;
	}

	dsp->log_bar1 = ioremap(dsp->phys_bar1, BAR1_SIZE);
	if (dsp->log_bar1 == NULL) {
		printk(KERN_ERR "couldn't get logical base address 1\n");
		return -ENODEV;
	}
	dprintk(4, "log bar 0:%p  1:%p\n", dsp->log_bar0, dsp->log_bar1);
	dsp->intEnabled = 0;
#ifdef S817_V4L2
	/* initialize each of the channel's settings */
	for (i = 0; i < SYS_CHAN_PER_DSP; i++) {
		if (s817_init_v4l2(&dsp->channel[i], 
				   dsp->dsp_num * SYS_DSP_PER_BOARD + i) != 0) {
			return -EINVAL;
		}
	}
#endif

#if LINUX_VERSION_CODE >= VERSION_IRQF_SHARED_INTRODUCED
	rc = request_irq(dsp->pdev->irq, s817_irq, IRQF_SHARED, "s817", (void *)dsp);
#else
	rc = request_irq(dsp->pdev->irq, s817_irq, SA_SHIRQ | SA_INTERRUPT, "s817", (void *)dsp);
#endif

	printk(KERN_INFO "s817: request irq %d\n", dsp->pdev->irq);
	if (rc != 0) {
		printk(KERN_ERR "couldn't get IRQ\n");
		return -ENODEV;
	}
	/* start the tasklet */
	tasklet_init(&s817_tasklet[num][num_boards], s817_dpc, (unsigned long) dsp);

	return 0;
}

static void init_emif(struct s817_dsp *dsp)
{

	/* EMIFA */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_GCTL, 0x00052078);
	/* CE0 SDRAM                     */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_CE0, 0xffffffd3);
	/* CE1 Flash + CPLD              */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_CE1, 0x73a28e01);
	/* CE2 Daughtercard 32-bit async */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_CE2, 0x22a28a22);
	/* CE3 Daughtercard 32-bit sync  */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_CE3, 0x22a28a42);
	/* SDRAM control                 */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_SDRAMCTL, 0x57115000);
	/* SDRAM timing (refresh)        */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_SDRAMTIM, 0x0000081b);
	/* SDRAM extended control        */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_SDRAMEXT, 0x001faf4d);
	/* CE0 Secondary Control Reg.    */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_CE0SECCTL, 0x00000002);
	/* CE1 Secondary Control Reg.    */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_CE1SECCTL, 0x00000002);
	/* CE2 Secondary Control Reg.    */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_CE2SECCTL, 0x00000002);
	/* CE3 Secondary Control Reg.    */
	s817_writeDSPreg(dsp->log_bar1, EMIFA_CE3SECCTL, 0x00000073);
}

static int load_dsp_firmware(struct s817_dsp *dsp)
{
	int		      loaded = 0;	/*loaded bytes counter*/
	u32                   *fwSrc;
	volatile u32	      *fwDst;
	struct slfheader      *slfHead;
	int		      i;
	int                   ts;
	const struct firmware *fw = dsp->dev->fw;
	dprintk(2, "got firmware %s size %lu\n", FIRMWARE_FILE_NAME, (unsigned long) fw->size);
	/*reset DSP*/
	s817_writeDSPreg(dsp->log_bar1, DSP_HDCR, DSP_HDCR_WARMRESET);
	/*wait >= 16 PCI cycles*/
	msleep(20);

	init_emif(dsp);

	slfHead = (struct slfheader *) fw->data;
	fwSrc = (unsigned int *) (fw->data + sizeof(struct slfheader));
	for (i = 0; i < slfHead->nSect; i++) {
		if (slfHead->sectData[i].startAddr == SYS_DSP_SDRAM_BASE) {
			/*.boot section goes to ISRAM, addr 0*/
			/*set page register to ISRAM start*/
			s817_writeDSPreg(dsp->log_bar1, DSP_DSPP,
					 DSP_PAGE_ISRAMSTART);
			fwDst = (u32 *) dsp->log_bar0; /*address 0*/
		} else {
			/*set page register to SDRAM start*/
			s817_writeDSPreg(dsp->log_bar1, DSP_DSPP,
					 DSP_PAGE_SDRAMSTART);
			fwDst = (u32 *) (dsp->log_bar0 +
					 (slfHead->sectData[i].startAddr
					  - SYS_DSP_SDRAM_BASE));
		}
		loaded = 0;
		while (loaded < (slfHead->sectData[i].size +
				 slfHead->sectData[i].padding)) {

			writel(*fwSrc, fwDst); //*fwDst = *fwSrc;
			if (*fwDst != *fwSrc) {
				printk(KERN_ERR "FW verify error: "
				       "in code section %d load address %p\n",
				       i, fwDst);
				return -ENODEV;
			} else {
				++fwDst;
				++fwSrc;
				loaded += sizeof(u32);
			}
		}
	}
	ts = s817_readDSPreg(dsp->log_bar1, HDCR_THREAD_STATUS);
	ts &= ~(REG3_UPNRUNNING);
	s817_writeDSPreg(dsp->log_bar1, HDCR_THREAD_STATUS, ts);
	s817_writeDSPreg(dsp->log_bar1, HDCR_COMMAND, 0);
	/* flush load */
	(void) s817_readDSPreg(dsp->log_bar1, HDCR_COMMAND);

	/*start firmware: write the DSPINT bit in HDCR with a 1*/
	s817_writeDSPreg(dsp->log_bar1, DSP_HDCR, DSP_HDCR_DSPINT);
	(void) s817_readDSPreg(dsp->log_bar1, DSP_HDCR);

	return 0;
}

/*marks buffer(s) selected with 1's in mask as available*/
void releaseBufs(struct s817_dev *dev, int dsp, int dChan, int mask, int capType)
{
	long	      lBuf;
	unsigned long flags;
	spin_lock_irqsave(&dev->dsp[dsp].slock[dChan], flags);
	lBuf = s817_readDSPreg(dev->dsp[dsp].log_bar1, bufStatReg[dChan]);
	if (capType == CAP_JPG)
		lBuf |= mask;
	else
		lBuf |= (mask << 16);
	s817_writeDSPreg(dev->dsp[dsp].log_bar1, bufStatReg[dChan], lBuf);
	/* don't need to read it back.  claim bufs will do this */
	spin_unlock_irqrestore(&dev->dsp[dsp].slock[dChan], flags);
}

/*marks buffer(s) as unavailable*/
static void claimBufs(struct s817_dev *dev, int dsp, int dChan, int mask, int capType)
{
	long	      lBuf;

	lBuf = s817_readDSPreg(dev->dsp[dsp].log_bar1, bufStatReg[dChan]);
	if (capType == CAP_JPG)
		lBuf &= ~mask;
	else 
		lBuf &= ~(mask << 16);
	s817_writeDSPreg(dev->dsp[dsp].log_bar1, bufStatReg[dChan], lBuf);
	/* don't need to read it back.  release bufs will do this */
}


//returns the dataMask: 1's set for channels that have jpg [0:15] and/or bmp [16:31] data
static int getDataMask(struct s817_dev *dev)
{
	int				i;
	int				dataMask = 0;
	
	for (i = SYS_CHAN_PER_BOARD - 1; i >= 0; i--) {
		dataMask <<= 1;
		if (isNotEmpty (&dev->bufDataFIFO[i][CAP_JPG - 1]))
			dataMask |= 1;
		if (isNotEmpty (&dev->bufDataFIFO[i][CAP_BMP - 1]))
			dataMask |= 0x10000;
	}
	return dataMask;
}



/*CheckDSPStatus*/
/*checks HDCR_THREAD_STATUS on DSPs which have upAndRunning flag not set*/
/*and sets it if necessary*/
/*returns 1 when all DSPs are up and running, otherwise returns 0*/
static int CheckDSPStatus(struct s817_dsp *dsp, int idx, int *prv)
{
	int	ret = 1;
	int	rv = -1;
	/*if not set check the register*/

	if (dsp->upAndRunning == 0) {
		/*check if all threads running on DSP*/
		rv = s817_readDSPreg(dsp->log_bar1, HDCR_THREAD_STATUS);
		dsp->threadStat = rv;
		if ((rv & REG3_PCI) == REG3_PCI) {
			/*if PCI thread up*/
			/*address od sysData struct in DSP's mem space*/
			dsp->sysData.dspAddr = rv &
				~(REG3_UPNRUNNING | REG3_DSP0);
		}
		if ((rv & REG3_UPNRUNNING) == REG3_UPNRUNNING) {
			dsp->upAndRunning = 1; /*set flag: DSP up and running*/
			/*check if DSP0(may be unnecessary) */
			if (rv & REG3_DSP0)
				dsp->dev->dsp0 = idx + 1;
			/*set to dsp index + 1,
			  so that configured value is always != 0*/
		}
	}
	ret &= dsp->upAndRunning; /*logical & with ret:
				    if any DSP not u&r ret will be 0*/
	*prv = rv;
	return ret;
}

static void s817_release_dsp(struct s817_dsp *dsp)
{
	if (!dsp) {
		printk(KERN_ERR "s817: null dsp\n");
		return;
	}

#if LINUX_VERSION_CODE >= VERSION_VMALLOC
	if (dsp->pdev->irq) {
		free_irq(dsp->pdev->irq, dsp);
	}
#else
	if (dsp->pdev) {
		free_irq(dsp->pdev->irq, dsp);
	}
#endif
	if (dsp->log_bar0) {
		/*release_mem_region(dsp->phys_bar0, BAR0_SIZE);*/
		iounmap(dsp->log_bar0);
		dsp->log_bar0 = 0;
	}
	if (dsp->log_bar1) {
		/*  release_mem_region(dsp->phys_bar1, BAR1_SIZE);*/
		iounmap(dsp->log_bar1);
		dsp->log_bar1 = 0;
	}

	if (dsp->pdev) {
		pci_disable_device(dsp->pdev);
		dsp->pdev = NULL;
	}

	return;
}

static int GetSysData(struct s817_dsp *dsp)
{
	char *sysDataAddr; /*address of SYS_DATA mapped into host's mem space*/

#if 0
	if (dsp->sysData.dspAddr < SYS_DSP_SDRAM_BASE) {
		printk(KERN_INFO "dsp address invalid\n");
		return;
	}
#endif
#if 0
	if (!dsp->dev[dsp->dsp_num].pdev) {
		printk(KERN_INFO "dsp %d not present\n", dsp->dsp_num);
		return -ENODEV;
	}
#endif
	if (dsp->dev[dsp->dsp_num].dsp_loaded == 0) {
		printk(KERN_INFO "dsp not loaded\n");
		return -ENODEV;
	}

	if ((dsp->sysData.dspAddr & 0xf0000000) != 0x80000000) {
		return -EFAULT;
	}

	sysDataAddr = dsp->log_bar0 + dsp->sysData.dspAddr - SYS_DSP_SDRAM_BASE;
	/*set page register(data is within first 4MB page of 0x80000000
	  space DSP_PAGE_SDRAMSTART)*/
	s817_writeDSPreg(dsp->log_bar1, DSP_DSPP, DSP_PAGE_SDRAMSTART);
	memcpy(&dsp->sysData, (char *) sysDataAddr, sizeof(SYS_STATUS));
	return 0;
}

static int free_all_channel_mem(struct s817_dev *dev)
{
	int i;
	int bChan;

	for (bChan = 0; bChan < SYS_CHAN_PER_BOARD; bChan++) {
		int dChan = bChan % SYS_CHAN_PER_DSP;
		int dsp = bChan / SYS_DSP_PER_BOARD;
		
		for (i = 0; i < SYS_BUF_PER_CHAN; i++) {
			FreeSysBuff(dev->dsp[dsp].pdev, &dev->sysBufJ[bChan][i]);
			FreeSysBuff(dev->dsp[dsp].pdev, &dev->sysBufB[bChan][i]);
		}
		FreeCommandBuff(dev->dsp[dsp].pdev, &dev->dsp[dsp].cmd[dChan]);
		FIFOdestroy(&dev->bufDataFIFO[bChan][CAP_JPG - 1]);
		FIFOdestroy(&dev->bufDataFIFO[bChan][CAP_BMP - 1]);
	}
	return 0;
}


static int alloc_channel_mem(struct s817_dev *dev, int bChan)
{
	int i;
	int dChan = bChan % SYS_CHAN_PER_DSP;
	int dsp = bChan / SYS_DSP_PER_BOARD;
	int ecode;
	/*Allocate struct fifo */
	dprintk(4, "creating channel mem: \n");

	if (FIFOcreate(&dev->bufDataFIFO[bChan][CAP_JPG - 1], SYS_FIFO_SIZE)) {
		printk(KERN_ERR "s817: could not allocate fifo\n");
			return -ENOMEM;
	}
	
	if (FIFOcreate(&dev->bufDataFIFO[bChan][CAP_BMP - 1], SYS_FIFO_SIZE)) {
		printk(KERN_ERR "s817: could not allocate fifo\n");
		return -ENOMEM;
	}
	/*Allocate command buffer*/
	dev->dsp[dsp].cmd[dChan].sxDma.priv_type = 0;
	dev->dsp[dsp].cmd[dChan].sxDma.priv_dev = dev;
	ecode = AllocCommandBuff(dev->dsp[dsp].pdev,
				 &dev->dsp[dsp].cmd[dChan]);
	
	if (ecode) {
		printk(KERN_ERR "s817:could not allocate command buffer\n");
		return -ENOMEM;
	}
	
	for (i = 0; i < SYS_BUF_PER_CHAN; i++) {
		dev->sysBufJ[bChan][i].image.sxDma.priv_type = CAP_JPG;
		dev->sysBufJ[bChan][i].image.sxDma.priv_bChan = bChan;
		dev->sysBufJ[bChan][i].image.sxDma.priv_bufId = i;
		dev->sysBufJ[bChan][i].image.sxDma.priv_dev = dev;
		dev->sysBufJ[bChan][i].image.sxDma.priv_bStat = 0;

		dev->sysBufJ[bChan][i].status.sxDma.priv_type = CAP_JPG;
		dev->sysBufJ[bChan][i].status.sxDma.priv_bChan = bChan;
		dev->sysBufJ[bChan][i].status.sxDma.priv_bufId = i;
		dev->sysBufJ[bChan][i].status.sxDma.priv_dev = dev;
		dev->sysBufJ[bChan][i].status.sxDma.priv_bStat = 1;
		ecode = AllocSysBuff(dev->dsp[dsp].pdev,
				     &dev->sysBufJ[bChan][i], CAP_JPG);

		if (ecode) {
			printk(KERN_ERR "s817: could not allocate buffer\n");
			return -ENOMEM;
		}

		dev->sysBufB[bChan][i].image.sxDma.priv_type = CAP_BMP;
		dev->sysBufB[bChan][i].image.sxDma.priv_bChan = bChan;
		dev->sysBufB[bChan][i].image.sxDma.priv_bufId = i;
		dev->sysBufB[bChan][i].image.sxDma.priv_dev = dev;
		dev->sysBufB[bChan][i].image.sxDma.priv_bStat = 0;
		dev->sysBufB[bChan][i].status.sxDma.priv_type = CAP_BMP;
		dev->sysBufB[bChan][i].status.sxDma.priv_bChan = bChan;
		dev->sysBufB[bChan][i].status.sxDma.priv_bufId = i;
		dev->sysBufB[bChan][i].status.sxDma.priv_dev = dev;
		dev->sysBufB[bChan][i].status.sxDma.priv_bStat = 1;
		ecode = AllocSysBuff(dev->dsp[dsp].pdev,
				     &dev->sysBufB[bChan][i], CAP_BMP);
		if (ecode) {
			printk(KERN_ERR "s817: could not allocate buffer\n");
			return -ENOMEM;
		}
	}
	return 0;
}

static int alloc_all_channel_mem(struct s817_dev *dev)
{

	int bChan;
	int rc;

	for (bChan = 0; bChan < SYS_CHAN_PER_BOARD; bChan++) {
		rc = alloc_channel_mem(dev, bChan);
		if (rc != 0) {
			free_all_channel_mem(dev);
			return rc;
		}
	}
	return 0;
}


static int __devinit s817_probe(struct pci_dev *pdev,
				const struct pci_device_id *ent)
{	
#ifdef CONFIG_PCI
	struct s817_dev *dev;
	struct pci_dev *dsp_pdev = NULL;
	struct pci_bus *bus;
	unsigned int i;
	int rc = 0;
	int dsps_found = 0;
	int allocfail = 0;
	MODE defaultMode = DEFAULT_MODE;
	int retval = -EINVAL;
	OSD_MODE defaultOsd = DEFAULT_OSD_MODE;
	char proc[64];

	printk(KERN_INFO "s817_probe on pci bus %d\n", pdev->bus->number);

	if (num_boards >= MAX_BOARDS) {
		printk(KERN_ERR "Too many boards installed.\n");
		dev = NULL;
		rc = -EFAULT;
		goto err;
	}

	dev = KZALLOC(sizeof(struct s817_dev), GFP_KERNEL, "device");

	/* default dsp0 to the first DSP.  will get changed later if
	   different in CheckDSPStatus */

	if (dev == NULL) {
		printk(KERN_ERR "s817: out of memory");
		rc = -ENOMEM;
		goto err;
	}

	for (i = 0; i < SYS_DSP_PER_BOARD; i++) {
		dev->dsp[i].cmd = KZALLOC(sizeof(struct int_buffer)
					  *SYS_CHAN_PER_DSP,
					  GFP_KERNEL,
					  "dsp cmd");
		if (dev->dsp[i].cmd == NULL) {
			printk(KERN_ERR "s817: out of memory");
			rc = -ENOMEM;
			goto err;
		}
	}

	dev->dsp0 = 1;
	dprintk(2, "probe dev %p\n", dev);

        /* sysBuf [SYS_CHAN_PER_BOARD][SYS_BUF_PER_CHAN][2]  // */

	dev->sysBufJ = KZALLOC(SYS_CHAN_PER_BOARD * sizeof(struct system_buffer *),
			      GFP_KERNEL, "sysBuf");
	if (dev->sysBufJ == NULL) {
		printk(KERN_ERR "s817: out of memory");
		rc = -ENOMEM;
		goto err;
	}
	dev->sysBufB = KZALLOC(SYS_CHAN_PER_BOARD * sizeof(struct system_buffer *),
			      GFP_KERNEL, "sysBuf");

	if (dev->sysBufB == NULL) {
		printk(KERN_ERR "s817: out of memory");
		rc = -ENOMEM;
		goto err;
	}
	for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
		char tag[100];
		sprintf(tag, "sysBuf %d", i);
		dev->sysBufB[i] = KZALLOC(SYS_BUF_PER_CHAN * sizeof(struct system_buffer),
					  GFP_KERNEL, tag);
		if (dev->sysBufB[i] == NULL) {
				allocfail = 1;
				break;
		}
		dev->sysBufJ[i] = KZALLOC(SYS_BUF_PER_CHAN * sizeof(struct system_buffer),
					  GFP_KERNEL, tag);
		if (dev->sysBufJ[i] == NULL) {
			allocfail = 1;
			break;
		}
	}

	if (allocfail) {
		rc = -ENOMEM;
		goto err;
	}

	init_waitqueue_head(&dev->wq_done);
	MUTEX_INIT(&dev->open_lock);
	/* lock open.  Open MUST not be entered until after
	 * probe finishes
	 */
	MUTEX_LOCK(&dev->open_lock);
	for (i = 0; i < SYS_DSP_PER_BOARD; i++) {
#ifdef S817_V4L2
		int j;
#endif
		dev->dsp[i].dsp_num = i;
		dev->dsp_loaded[i] = 0;
#ifdef S817_V4L2
		for (j = 0; j < SYS_CHAN_PER_DSP; j++) {
			dev->dsp[i].channel[j].dsp = &dev->dsp[i];
			dev->dsp[i].channel[j].dev = dev;
		}
#endif
	}
#ifdef S817_V4L2
	for (i = 0; i < SYS_OUT_PER_BOARD; i++) {
		dev->output[i].dev = dev;
		/* create /proc/s817/output/... entries */
		if (s817_procout_root) {
			int mode = S_IFREG | S_IRUGO | S_IWUGO;
			struct proc_dir_entry *proc;
			char name[10];
			sprintf(name, "%d", 
				i+ SYS_OUT_PER_BOARD * num_boards);
			proc = create_proc_entry(name, mode, s817_procout_root);
			if (ent) {
				proc->data = &dev->output[i];
				proc->read_proc = read_output;
				proc->write_proc = write_output;
			}
		}
	}
#endif

	retval = pci_enable_device(pdev);
	if (retval) {
		printk(KERN_ERR " PCI enable fail !\n");
		rc = -ENODEV;
		goto err;
	}

	retval = request_firmware(&dev->fw, FIRMWARE_FILE_NAME, &pdev->dev);

	if (retval < 0) {
		printk(KERN_ERR "s817: failed to retrieve firmware\n");
		rc = -ENOMEM;
		goto err;
	}

        /* initialize the dsp pointers in case they are absent */
	for (i = 0; i < SYS_DSP_PER_BOARD; i++) {
		dev->dsp[i].pdev = NULL;
		dev->dsp[i].dev = dev;
	}


	/* attach all dsp devices on the child buses */
	list_for_each_entry(bus, &pdev->bus->children, node)
	list_for_each_entry(dsp_pdev, &bus->devices, bus_list) {
		if (dsp_pdev->vendor == PCI_VENDORID_DM642 && dsp_pdev->device == PCI_DEVICEID_DM642 &&
		    ((dsp_pdev->subsystem_vendor == PCI_SUBSYSTEM_VENDORID && dsp_pdev->subsystem_device == PCI_SUBSYSTEM_817) ||
		     (dsp_pdev->subsystem_vendor == 0x0000 && dsp_pdev->subsystem_device == PCI_SUBSYSTEM_817_UP))) {
			unsigned int pci_slot;
			pci_slot = PCI_SLOT(dsp_pdev->devfn);
			if (pci_slot >= SYS_DSP_PER_BOARD) {
				printk(KERN_ERR "s817: invalid DSP PCI slot %d\n", pci_slot);
				continue;
			}	

#if LINUX_VERSION_CODE >= VERSION_VMALLOC
			dsp_pdev->irq = pdev->irq++;
#endif
			retval = attach_dsp(&dev->dsp[pci_slot], dsp_pdev, pci_slot, num_boards);

			if (retval != 0) {
				printk(KERN_ERR "failed to attach DSP\n");
				rc = -ENODEV;
				goto err;
			}
			dsps_found++;
			dev->dsp[pci_slot].dev = dev;
		} else {
			printk(KERN_ERR "s817: unexpected device %04x:%04x %04x:%04x on pci bus %d\n",
				dsp_pdev->vendor, dsp_pdev->device, 
				dsp_pdev->subsystem_vendor, dsp_pdev->subsystem_device,
				dsp_pdev->bus->number);
		}
	}
	
	printk(KERN_INFO "s817: found %d DSP%c\n", dsps_found, dsps_found == 1 ? ' ':'s');

	/* allocate mem */
	if (alloc_all_channel_mem(dev) != 0) {
		printk(KERN_ERR "not enough memory. please read README_vmalloc and/or check your memory usage\n");
		rc = -ENOMEM;
		goto err;
	}
	
	for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
		dev->mode[i] = defaultMode;
		dev->osdmode[i] = defaultOsd;
	}
	dev->minor = num_boards++;
	list_add_tail(&dev->s817_devlist, &s817_devlist);
	pci_set_drvdata(pdev, dev);
	sprintf(proc, "s817_%d", dev->minor);
#if LINUX_VERSION_CODE >= CREATE_PROC_REMOVED
	proc_create_data(proc, S_IRUGO, 0, &proc_file, dev);
#else
	create_proc_read_entry(proc, 0, 0, s817_read_procmem, dev);
#endif
	printk(KERN_INFO "s817 board successfully loaded\n");
#ifdef S817_V4L2
	if (fwmode == 0) {
		/* open all channels for faster app start-up */
		int i;
		for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
			(void) s817_openchan(dev, i+1, &dev->mode[i]);
		}

	}
#endif
	MUTEX_UNLOCK(&dev->open_lock);
	return 0;
err:
	printk(KERN_ERR "s817 probe failed\n");
	if (dev) {
		for (i = 0; i < SYS_DSP_PER_BOARD; i++)
			s817_release_dsp(&dev->dsp[i]);

		if (dev->fw)
			release_firmware(dev->fw);

	}
	if (pdev)
		pci_disable_device(pdev);

	if (dev) {
		if (dev->sysBufJ) {
			char tag[100];
			for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
				sprintf(tag, "sysBuf %d", i);
				KFREE(dev->sysBufJ[i], tag);
			}
			KFREE(dev->sysBufJ, "sysbuf");
			dev->sysBufJ = 0;
		}
		if (dev->sysBufB) {
			char tag[100];
			for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
				sprintf(tag, "sysBuf %d", i);
				KFREE(dev->sysBufB[i], tag);
			}
			KFREE(dev->sysBufB, "sysbuf");
			dev->sysBufB = 0;
		}

		for (i = 0; i < SYS_DSP_PER_BOARD; i++) {
			KFREE(dev->dsp[i].cmd, "dsp cmd");
		}
		KFREE(dev, "device");
		dev = NULL;
	}
	MUTEX_UNLOCK(&dev->open_lock);
	return rc;

#else
#error "s817 only works with PCI.  enable PCI in your kernel!\n"
#endif
}

void SX_DMALock(struct pci_dev *pdev, struct sx_dma *dma, int dir)
{
	int i;
	struct page *pg;
	int rem;
	unsigned char *virt;
	struct scatterlist sg;
	dma_addr_t dmah;
	int dbglvl = 4;
	rem = (dma->dwBytes % PAGE_SIZE);
	dma->dwPages = dma->dwBytes / PAGE_SIZE;
	if (dma->dwPages > SX_DMA_PAGES) {
		/* should never get here.  driver controlled*/
		printk(KERN_ERR "too many pages!!\n");
		dma->dwPages = SX_DMA_PAGES;
	}
	
	/*    dprintk(2,"bpt %d, pages %d\n",PAGE_SIZE,dma->dwPages);*/
	virt = (unsigned char *) dma->pUserAddr;

	/*  printk("num_pages %d\n", dma->dwPages);*/
	for (i = 0; i < dma->dwPages; i++) {

		pg = vmalloc_to_page(virt);
#if LINUX_VERSION_CODE < VERSION_SGSETPAGE_INTRODUCED
		sg.page = pg;
		sg.length = PAGE_SIZE;
#else
		sg_set_page(&sg, pg, PAGE_SIZE, 0);
#endif
		/* do not swap out page */
		SetPageReserved(pg);
		/* save the page */
		dma->Page[i].pg = pg;
		dma->Page[i].dwBytes = sg.length;
		dmah = pci_map_page(pdev, pg, 0, sg.length, dir);
		dma->Page[i].dir = dir;
		dma->Page[i].pPhysicalAddr = cpu_to_le32(dmah);
		dma->Page[i].dmah = dmah;
		dbglvl = (i < 4) ? 5 : 125;
		dprintk(dbglvl, "page[%03d] virt/cpu %p dma/phys %lx, test %lx\n",
			i, virt, dma->Page[i].pPhysicalAddr,
			(unsigned long)dmah);
		if ((debug > 5) && (i == 1)) {
			if (dma->priv_type == CAP_JPG) {
				dma->priv_dev->dbgVA_J[dma->priv_bChan][dma->priv_bufId][dma->priv_bStat] = virt;
				dma->priv_dev->dbgPg_J[dma->priv_bChan][dma->priv_bufId][dma->priv_bStat] = pg;
			} else if (dma->priv_type == CAP_BMP) {
				dma->priv_dev->dbgVA_B[dma->priv_bChan][dma->priv_bufId][dma->priv_bStat] = virt;
				dma->priv_dev->dbgPg_B[dma->priv_bChan][dma->priv_bufId][dma->priv_bStat] = pg;
			}
		}
		virt += PAGE_SIZE;

	}
	/* last page*/
	if (rem) {
		/* another page for the runt*/
		pg = vmalloc_to_page(virt);
		sg.dma_address = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
		sg.page = pg;
		sg.length = rem;
#else
		sg_set_page(&sg, pg, rem, 0);
#endif
		/* do not swap out page */
		SetPageReserved(pg);

		dmah = pci_map_page(pdev, pg, 0, sg.length, dir);
		dma->Page[i].pg = pg;
		dma->Page[i].dir = dir;
		dma->Page[i].dwBytes = sg.length;
		dma->Page[i].pPhysicalAddr = cpu_to_le32(dmah);
		dma->Page[i].dmah = dmah;
		dma->dwPages++;
		dprintk(5, "last page[%03d] virt/cpu %p dma/phys %lx %d\n",
			i, virt, dma->Page[i].pPhysicalAddr, sg.length);
	}
	dprintk(3, "pages %p %d %d\n", dma, dma->dwPages, i);
	return;
}

static void SX_DMAUnlock(struct pci_dev *pdev, struct sx_dma *dma, int dir)
{
	int i;
	if (!dma)
		return;

	for (i = 0; i < dma->dwPages; i++) {
		if (dma->Page[i].pg) {
			ClearPageReserved(dma->Page[i].pg);
		}
		pci_unmap_page(pdev,
			       dma->Page[i].dmah,
			       dma->Page[i].dwBytes,
			       dir);
	}

	dma->dwPages = 0;
	return;
}

/*SxAllocDMABuf*/
/*Allocate a DMA buffer using SXDriver*/
/*Create a buffer using malloc, lock for DMA to get phys. page addresses*/
/*pUserAddr keeps the start address to use with free() when deleting*/
ECODE SxAllocDMABuf(struct pci_dev *pdev, int size, struct sx_dma *pdma, int dir)
{
	/*Allocates a DMA buffer*/
	/*uses the SX_DMA structure pointed to by pdma*/
	ECODE ecode = 0;
	/*Page[0] could be less than SYS_PAGE,
	  so discard it and allocate 2 pages more*/
	pdma->dwBytes = (size / SYS_PAGE + 2) * SYS_PAGE;
	/*pdma->dwOptions = DMA_LARGE_BUFFER;*/
	/* ???*/
	pdma->pUserAddr = VMALLOC32(pdma->dwBytes, "pUserAddr");
	if (pdma->pUserAddr == NULL) {
		printk(KERN_ERR "s817: out of mem\n");
		ecode = ERR_MEM_ALLOC;
	} else {
		SX_DMALock(pdev, pdma, dir);
	}
	return ecode;
}

/*SxFreeDMABuf*/
/*free buffer allocated by SxAllocBuff*/
void SxFreeDMABuf(struct pci_dev *pdev, struct sx_dma *pdma, int dir)
{
	SX_DMAUnlock(pdev, pdma, dir);

	if (pdma->pUserAddr)
		VFREE(pdma->pUserAddr, "pUserAddr");

	pdma->pUserAddr = NULL;
}


/*AllocCommandBuff*/
/*Allocates a command buffer (1 per DSP capture channel)*/
static ECODE AllocCommandBuff(struct pci_dev *pdev, struct int_buffer *pIntBuf)
{
	ECODE		ecode;
	

	ecode = SxAllocDMABuf(pdev, SYS_CS_BUF_SIZE, &pIntBuf->sxDma, PCI_DMA_TODEVICE);
	if (ecode)
		return ecode;
	/*adjust user addr past Page[0], which is usually < 4K*/
	/*when page table is created later, Page[0] will be skipped*/
	pIntBuf->pData = (char *) pIntBuf->sxDma.pUserAddr +
		pIntBuf->sxDma.Page[0].dwBytes;
	pIntBuf->boff = (G_offsets++ * PAGE_SIZE);
	dprintk(25, "allocCommandbuf %p %p %d\n",
		pIntBuf->sxDma.pUserAddr,
		pIntBuf->pData,
		pIntBuf->sxDma.Page[0].dwBytes);

	ecode = MakePageTable(pIntBuf, SYS_CS_BUF_SIZE);
	return ecode;
}

/*FreeCommandBuff*/
/*Frees command buffers, unlocks DMA*/
/*Safe to call on unallocated buffers */
static void FreeCommandBuff(struct pci_dev *pdev, struct int_buffer *pIntBuf)
{
	SxFreeDMABuf(pdev, &pIntBuf->sxDma, PCI_DMA_TODEVICE);
	if (pIntBuf->pageTbl.pPT != NULL) {
		KFREE(pIntBuf->pageTbl.pPT, "pPT");
		pIntBuf->pageTbl.pPT = NULL;
		pIntBuf->pageTbl.PTsize = 0;
	}
}



/*MakePageTable*/
/*Creates page table of a buffer:*/
/*array of physical addresses of consecutive 4K pages of the buffer*/
ECODE MakePageTable(struct int_buffer *pIntBuf, int size)
{
	int *ppt;	/*pointer to page table*/
	int phys;       /*current phys address*/
	int numTblPages;/*number of processed pages of page table (4K pages)*/
	int needPages; /*number of pages required for the buffer */
	int p;	/*current page index in DMA structure
			  (pages may be > 4K)*/
	s32 curPageSize;/*current page size*/

	needPages = size/SYS_PAGE + (size % SYS_PAGE ? 1 : 0);
	/*allocate memory for page table
	  (extra page is OK, DMA buffer is at least that size)*/
	pIntBuf->pageTbl.pPT = KZALLOC((needPages+1) * sizeof(int), GFP_KERNEL,
				       "pPT");

	if (pIntBuf->pageTbl.pPT == NULL) {
		printk(KERN_ERR "out of memory!"
		       "could not allocate page table\n");
		return ERR_MEM_ALLOC;
	}

	ppt = pIntBuf->pageTbl.pPT;
	/*Page table contains phys addresses of consecutive SIZE/SYS_PAGE+1*/
	/*4K blocks of a DMA buffer starting with Page[1]*/
	/*Note: SxDriver's .Page[] can be > 4K (contiguous phys addr block)*/
	numTblPages = 0;
	p = 1;
	phys = (int) pIntBuf->sxDma.Page[1].pPhysicalAddr;
	dprintk(8, "curphys addr %d\n", phys);
	curPageSize = 0;
	while (numTblPages++ < needPages) {
		*ppt++ = phys;
		curPageSize += SYS_PAGE;
		if (curPageSize < pIntBuf->sxDma.Page[p].dwBytes) {
			/*current page > SYS_PAGE,
			  contiguous phys address block*/
			phys += SYS_PAGE;
		} else {
			/*current page is over*/
			phys = (int) pIntBuf->sxDma.Page[++p].pPhysicalAddr;
			curPageSize = 0;
		}
	}
	/*store PT size rounded up*/
	pIntBuf->pageTbl.PTsize = needPages;
	return 0;
}



/*GetBmpSize*/
static int GetBmpWH(MODE *pMode, int *w, int *h)
{
	int size = 0;
	if (pMode == NULL) {
		printk(KERN_INFO "s817: invalid mode !\n");
		return 0;
	}

	switch (pMode->vMode.tvFormat) {
	case TV_NTSC:
	default:
		switch (pMode->vMode.jpgSize) {
		case SIZE_4CIF:
		case SIZE_4CIFI:
			*h = SYS_NTSC_4CIF_VS;
			*w = SYS_NTSC_4CIF_HS;
			break;
		case SIZE_2CIF:
			*h = SYS_NTSC_2CIF_VS;
			*w = SYS_NTSC_2CIF_HS;
			break;
		case SIZE_1CIF:
			*h = SYS_NTSC_1CIF_VS;
			*w = SYS_NTSC_1CIF_HS;
			break;
		default:
			break;
		}
		break;

	case TV_PAL:
		switch (pMode->vMode.jpgSize) {
		case SIZE_4CIF:
		case SIZE_4CIFI:
			*h = SYS_PAL_4CIF_VS;
			*w = SYS_PAL_4CIF_HS;
			break;
		case SIZE_2CIF:
			*h = SYS_PAL_4CIF_VS;
			*w = SYS_PAL_4CIF_HS;
			break;
		case SIZE_1CIF:
			*h = SYS_PAL_4CIF_VS;
			*w = SYS_PAL_4CIF_HS;
			break;
		default:
			break;

		}
		break;
	}
	return size;
}

/*AllocSysBuff*/
/*Allocates a struct system_buffer (jpg & bmp)*/
static ECODE AllocSysBuff(struct pci_dev *pdev,
		   struct system_buffer *pSysBuf, int capType)
{
	ECODE	   ecode;
	int	   size;

	if (capType == CAP_JPG)
		size = SYS_JPG_BUF_SIZE;
	else if (capType == CAP_BMP)
		size = SYS_BMP_BUF_SIZE;
	else 
		return ERR_ILLEGAL_BUFTYPE;

	// image
	ecode = SxAllocDMABuf(pdev, size, &pSysBuf->image.sxDma, PCI_DMA_FROMDEVICE);
	if (ecode)
		return ecode;

	/*adjust user addr past Page[0],
	  which is usually < 4K (with SxDriver)*/
	/*when page table is created later, Page[0] will be skipped*/
	pSysBuf->image.pData = (char *) pSysBuf->image.sxDma.pUserAddr +
		pSysBuf->image.sxDma.Page[0].dwBytes;
	pSysBuf->image.boff = (G_offsets++ * PAGE_SIZE);
	dprintk(25, "allocsysbuf %p %p %d\n",
		pSysBuf->image.sxDma.pUserAddr,
		pSysBuf->image.pData,
		pSysBuf->image.sxDma.Page[0].dwBytes);
	ecode = MakePageTable(&pSysBuf->image, size);

	if (ecode)
		return ecode;

	/*status (1 per channel)*/
	ecode = SxAllocDMABuf(pdev, SYS_CS_BUF_SIZE, &pSysBuf->status.sxDma, PCI_DMA_FROMDEVICE);
	if (ecode)
		return ecode;


	/*adjust user addr past Page[0], which is usually < 4K*/
	/*when page table is created later, Page[0] will be skipped*/
	pSysBuf->status.pData =
		(char *) pSysBuf->status.sxDma.pUserAddr +
		pSysBuf->status.sxDma.Page[0].dwBytes;
	pSysBuf->status.boff = (G_offsets++ * PAGE_SIZE);

	dprintk(25, "allocsysbuf status %p %p %d\n",
		pSysBuf->status.sxDma.pUserAddr,
		pSysBuf->status.pData,
		pSysBuf->status.sxDma.Page[0].dwBytes);
	ecode = MakePageTable(&pSysBuf->status, SYS_CS_BUF_SIZE);
	return ecode;
}

/*FreeSysBuff*/
/*Frees system buffers, unlocks DMA*/
/*Safe to call on unallocated buffers */
static void FreeSysBuff(struct pci_dev *pdev, struct system_buffer *pSysBuf)
{
	SxFreeDMABuf(pdev, &pSysBuf->image.sxDma, PCI_DMA_FROMDEVICE);
	if (pSysBuf->image.pageTbl.pPT != NULL) {
		KFREE(pSysBuf->image.pageTbl.pPT, "image PT");
		pSysBuf->image.pageTbl.pPT = NULL;
		pSysBuf->image.pageTbl.PTsize = 0;
	}
	SxFreeDMABuf(pdev, &pSysBuf->status.sxDma, PCI_DMA_FROMDEVICE);
	if (pSysBuf->status.pageTbl.pPT != NULL) {
		KFREE(pSysBuf->status.pageTbl.pPT, "status PT");
		pSysBuf->status.pageTbl.pPT = NULL;
		pSysBuf->status.pageTbl.PTsize = 0;
	}
}


/*CloseChannel*/
/*Frees buffers*/
/*Doesn't affect interrupts (since interrupt serves 4 channels)*/
static int CloseChannel(struct s817_dev *dev, int bChan, int bTimeout)
{
	int		dChan = bChan % SYS_CHAN_PER_DSP;
	int		dsp = bChan / SYS_DSP_PER_BOARD;
	long            lBuf0, lBuf1;
	if ((bChan < 0) || (bChan >= SYS_CHAN_PER_BOARD) || !dev->dsp[dsp].pdev) {
		printk(KERN_INFO "closing invalid channel %d\n", bChan);
		return ERR_ILLEGAL_CHAN;
	}
	/* check if channel already closed */
	MUTEX_LOCK(&dev->dsp[dsp].dsplock);
	if (!dev->dsp[dsp].cOpen[dChan]) {
		dprintk(4, "channel %d not open\n", bChan + 1);
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return 0;
	}
	//stop capture
	//SetCaptureMode is sent as a short command
	//the extra argument arg holds CAPTURE_MODE members:
	//arg[31:28] = capMode & 0xf
	lBuf0 = CAP_OFF << 28;					//CAP_MODE=CAP_OFF
	//Short command: physAddr argument has 0's in bits 0xfffff000, command code in bits 0x00000ff0 and dChan in bits 0x0000000f
	lBuf1 = (DSP_CMD_CMODE << 4) | dChan;
	//send command
	sendDspCmd(dev, dsp, lBuf1, lBuf0);

	tasklet_disable(&s817_tasklet[dsp][dev->minor]);
	dev->dsp[dsp].cOpen[dChan] = 0;
	tasklet_enable(&s817_tasklet[dsp][dev->minor]);


	MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
	return 0;
}



/*
 *S817_OpenChannel
 * Opens a capture channel:
 * - allocates buffers
 * - sets up interrupt thread and semaphore
 * in which case must use mdelay other msleep because
 * we can't sleep from the kernel idle thread
*/
static int s817_openchan(struct s817_dev *dev, int aChan, MODE *pMode)
{
	int	i, j, ecode;
	int	bChan = aChan - 1;
	/*DSP index for given channel*/
	int	dsp = bChan / SYS_DSP_PER_BOARD;
	/*channel in DSP context (0..SYS_CHAN_PER_DSP)*/
	int	dChan = bChan % SYS_CHAN_PER_DSP;
	long	lBuf;
	int	iBuf;
	struct dsp_cmd_hdr dspCmdHdr;
	char    *buf;
	int	*PTdata; /*working pointer to page table data*/
  	VIDEO_MODE *pVMode = &pMode->vMode;
	int      modeChgMask = 0;
	int      dsp_loaded = 0;
	dprintk(4, "open channel function aChan=%d\n", aChan);
	if (aChan < 1 || aChan > SYS_CHAN_PER_BOARD || !dev->dsp[dsp].pdev) {
		printk(KERN_INFO "s817: illegal channel\n");
		return ERR_ILLEGAL_CHAN;
	}
	MUTEX_LOCK(&dev->dsp[dsp].dsplock);
	/* prevent separate channels from loading the DSP.
	 * user may try to open channels 1 and 2(same
	 * DSP) from different threads...
	 */
	if (!dev->dsp_loaded[dsp]) {
		int rc;
		int success = 0;
		int status;
		int retries = 0;
		MODE	 defaultMode = DEFAULT_MODE;
		/* DSP reloaded, reset mode structures */
		dev->mode[bChan] = defaultMode;
		dev->mode[bChan+1] = defaultMode;
		dev->mode[bChan+2] = defaultMode;
		dev->mode[bChan+3] = defaultMode;
		while (retries < 3) {
			printk(KERN_INFO "s817: loading firmware for DSP %d, [%d]\n", dsp + 1, retries);
			rc = load_dsp_firmware(&dev->dsp[dsp]);
			
			if (rc != 0) {
				retries++;
				dev->dsp[dsp].retries++;
				continue;
			}
			/*wait until the DSP is  up and running or timeout*/
			rc = 0;
			while (rc < SYS_DSP_START_TO) {
				if (CheckDSPStatus(&dev->dsp[dsp], dsp, &status)) {
					/* DSP loaded */
					success = 1;
					break;
				} else {
					msleep(SYS_DSP_START_SLEEP);
					rc += SYS_DSP_START_SLEEP;
				}
			}
			if (success)
				break;
			retries++;
		}
		printk(KERN_INFO "s817: dsp %d started\n", dsp + 1);

 		if (!success) {
			printk(KERN_INFO "reload failed!\n");
			MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
			return ERR_DSP_START_TO;
		}
		dsp_loaded = 1;
	}
	dev->dsp_loaded[dsp] = 1;

	if (dev->dsp[dsp].cOpen[dChan]) {
		dprintk(4, "s817: channel already open %d %d\n", dsp,
			dChan);
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return ERR_ALREADYOPEN;
	}


	dprintk(4, "int enabled flag %d\n", dev->dsp[dsp].intEnabled);

	/*set up an interrupt if not already enabled for corresponding DSP*/
	/*interrupt is per DSP (4 channels)*/
	/*Allocate command/status buffer*/
	/*CS buffer is per DSP (4 channnels)*/
	/*Allocate FIFO's (1 per DSP)*/

	if (dev->dsp[dsp].intEnabled == 0) {
		lBuf = s817_readDSPreg(dev->dsp[dsp].log_bar1, DSP_HSR);
		/*clear INTAM bit -> allow DSP to generate PCI interrupts;
		  reset INTSRC bit*/
		lBuf &= ~DSP_HSR_INTAM | DSP_HSR_INTSRC;
		//lBuf = DSP_HSR_INTSRC;
		dprintk(4, "enabling PCI interrupts\n");
		s817_writeDSPreg(dev->dsp[dsp].log_bar1, DSP_HSR, lBuf);
		(void ) s817_readDSPreg(dev->dsp[dsp].log_bar1, DSP_HSR);
		dev->dsp[dsp].intEnabled = 1; /*set the flag*/
		
	}

	//Send buffer page tables info to DSP
	buf = dev->dsp[dsp].cmd[dChan].pData;			//use channel's command buffer
	dspCmdHdr.marker = DSP_CMD_MARKER;					//command marker
	dspCmdHdr.cmdCode = DSP_CMD_BUFFER_PT_JPG;			//command: JPG pagetable follows (reuse this command code)
	dspCmdHdr.arg0 = dev->sysBufJ[bChan][0].image.pageTbl.PTsize;	//size of JPG pagetable in sizeof(int)
	dspCmdHdr.arg1 = dev->sysBufB[bChan][0].image.pageTbl.PTsize;	//size of BMP pagetable in sizeof(int)
	dspCmdHdr.channel = dChan;							//channel in the DSP context: 0 to 3
	dspCmdHdr.dataSize = SYS_BUF_PER_CHAN;				//number of buffers per channel
	*(struct dsp_cmd_hdr *) buf = dspCmdHdr;					//copy header to buffer
	PTdata = (int *) (buf + sizeof(dspCmdHdr));			//advance the pointer past header

	//write page table info into the buffer in the following order:
	//JPG_PT
	//STAT_PT
	//BMP_PT
	//STAT_PT
	// ...SYS_BUF_PER_CHAN times
	for (i = 0; i < SYS_BUF_PER_CHAN; i++) {
		//copy JPG_PT
		for (j = 0; j < dev->sysBufJ[bChan][i].image.pageTbl.PTsize; j++) {
			*PTdata++ = dev->sysBufJ[bChan][i].image.pageTbl.pPT[j];
		}
		//copy STAT_PT
		*PTdata++ = dev->sysBufJ[bChan][i].status.pageTbl.pPT[0];
		//copy BMP_PT
		for (j = 0; j < dev->sysBufB[bChan][i].image.pageTbl.PTsize; j++) {
			*PTdata++ = dev->sysBufB[bChan][i].image.pageTbl.pPT[j];
		}
		//copy STAT_PT
		*PTdata++ = dev->sysBufB[bChan][i].status.pageTbl.pPT[0];
	}

	iBuf = sendDspCmd(dev, dsp, dev->dsp[dsp].cmd[dChan].pageTbl.pPT[0] | dChan, 0);

	if (iBuf) {
		printk(KERN_ERR "failed command\n");
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return iBuf;
	}
	//Set video mode
	//check if valid parameters
	if (pVMode->tvFormat < TV_NTSC || pVMode->tvFormat > TV_SECAM)
		return ERR_ILLEGAL_TVFORMAT;
	if (pVMode->jpgSize < SIZE_4CIF || pVMode->jpgSize > SIZE_4CIFI)
		return ERR_ILLEGAL_JPGSIZE;

	//set mask bits for all parameters
	modeChgMask |= MODE_CHG_BRIGHTNESS;
	modeChgMask |= MODE_CHG_CONTRAST;
	modeChgMask |= MODE_CHG_SATURATION;
	modeChgMask |= MODE_CHG_HUE;
	buf = dev->dsp[dsp].cmd[dChan].pData;	//use channel's command buffer
	dspCmdHdr.marker = DSP_CMD_MARKER;
	dspCmdHdr.cmdCode = DSP_CMD_VMODE;				//send full command
	dspCmdHdr.arg0 = modeChgMask;
	dspCmdHdr.arg1 = 0;
	dspCmdHdr.channel = dChan;	//channel in the DSP context: 0 to 3
	dspCmdHdr.dataSize = sizeof(VIDEO_MODE);
	*(struct dsp_cmd_hdr *) buf = dspCmdHdr;
	*(VIDEO_MODE *) (buf + sizeof(struct dsp_cmd_hdr)) = *pVMode;		//copy MODE into buffer
	ecode = (ECODE) sendDspCmd(dev, dsp, dev->dsp[dsp].cmd[dChan].pageTbl.pPT[0] | dChan, 0);
	if (!ecode) {
	    dev->mode[bChan] = *pMode;
	    dev->dsp[dsp].cOpen[dChan] = 1;
	}
	MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
#ifdef S817_V4L2
	if (dsp_loaded) {
		// set default time and osd mode
		struct param_settime p;
		struct rtc_time	rtctime;
		get_rtc_time(&rtctime);
		memset(&p, 0, sizeof(struct param_settime));
		p.tm_sec = rtctime.tm_sec;
		p.tm_min = rtctime.tm_min;
		p.tm_hour = rtctime.tm_hour;
		p.tm_mday = rtctime.tm_mday;
		p.tm_mon = rtctime.tm_mon;
		p.tm_year = rtctime.tm_year;
		S817_SetDateTimeDsp(&dev->dsp[dsp], &p);
	}
	// load a default OSD channel message
	snprintf(dev->osdmode[bChan].line, 
		 sizeof(dev->osdmode[bChan].line), 
		 "Channel %d", 
		 bChan+1 + dev->minor*SYS_CHAN_PER_BOARD);
	
	S817_SetOsdMode(dev, aChan, &dev->osdmode[bChan]);
#endif	
	return ecode;
}

//channel numbers range in aChan context
#define MAX_CHAN        16
#define MIN_CHAN        1
#define NEXT_CHAN(x)    (++(x) > MAX_CHAN) ? MIN_CHAN : x

//starts checking with the channel next after the last one that returned data.
//dataMask has bits [0..15] set/reset according to JPG data present/absent for channels [1..16]
//dataMask has bits [16..31] set/reset according to BMP data present/absent for channels [1..16]
//it's assumed that dataMask != 0 (guaranteed by the caller).
//Returns first channel number having data (lower 16 bits), data type (higher 16 bits)

//#define _NO_RR_		//no round robin, debug
static int roundRobin (struct s817_dev *dev, int dataMask)
{
	int     nextChan = 0;
	int	dataType = 0;
	int loopcount;

#ifndef _NO_RR_
	nextChan = dev->lastChan;			//lastChan is in aChan context [1..16]
#else
	nextChan = 1;
#endif
	loopcount = 0;
	if (dataMask == 0) {
		printk(KERN_INFO "invalid data mask\n");
		return -1;
	}
	while (!dataType && loopcount < 30) {
		loopcount++;
		nextChan = NEXT_CHAN (nextChan);
		if (dataMask & (1 << (nextChan - 1)))
			dataType |= CAP_JPG;
		if (dataMask & (1 << (nextChan - 1 + 16)))       
			dataType |= CAP_BMP;
	}
	if (loopcount >= 30)
		printk(KERN_ERR "problem with round robin\n");
	dev->lastChan = nextChan;
	return dataType << 16 | nextChan;
}



struct s817_mapping {
	void *pvirt; /* virtual start*/
	unsigned long size; /* size of it*/
	struct s817_dev *dev;
	struct s817_v4l2 *channel;
	int bufId;
	int bStat;
	int bChan;
	int type;
        int count;	/* how many times is struct this used */
};

static void s817_vm_open(struct vm_area_struct *vma)
{
	struct s817_mapping	*pmap;
	struct s817_v4l2	*channel;
	struct s817_dev 	*dev;
	int			count;
	int                     bChan;
        if (vma == NULL)
                return;

	pmap = vma->vm_private_data;
        if (pmap == NULL)
                return;

	if (pmap->channel == NULL) {
		return;
	}
	/* code below is for V4L2 only.  standard SDK sets channel to NULL */
	dev = pmap->dev;
	channel = pmap->channel;
	bChan = channel->bChan;

	/* 
	 * Keep track of how many times the pmap has been used.  That is, how
	 * many times the same vma structure was passed to _mmap() or 
	 * _vm_open().  This allows us to free the pmap after the last 
	 * _vm_close().
	 */
	pmap->count++;
	bChan = channel->bChan;
	if (pmap->bufId >= channel->numbufs)
		return;

	/* 
	 * Keep track of how many times the buffer was mapped.  Don't confuse
	 * this with the pmap->count.  A may be mapped using the same or 
	 * different vma strucutures (same or different _mmap()).
	 */
	if (pmap->type == CAP_JPG) {
		if (pmap->bStat) {
			count = 
			    ++(dev->sysBufJ[bChan][pmap->bufId].status.map_count);
		} else {
			count = 
			    ++(dev->sysBufJ[bChan][pmap->bufId].image.map_count);
		}
	} else {
		if (pmap->bStat) {
			count = 
			    ++(dev->sysBufB[bChan][pmap->bufId].status.map_count);
		} else {
			count = 
			    ++(dev->sysBufB[bChan][pmap->bufId].image.map_count);
		}
	}
        dprintk(4, "VMopen: channel %d %s %s buf %d; count now %d\n", 
			channel->bChan, 
			pmap->type == CAP_JPG ? "JPEG" : "BMP",
			pmap->bStat ? "status" : "image",
			pmap->bufId, count);
	dprintk(4, "VMopen\n");
}

static void s817_vm_close(struct vm_area_struct *vma)
{
	struct s817_mapping	*pmap;
	struct s817_v4l2	*channel;
	struct s817_dev	*dev;
	int count = 0;
	int bChan;

        if (vma == NULL) {
		printk("HEY: s817_vm_close NULL vma\n");
                return;
	}

	pmap = vma->vm_private_data;
        if (pmap == NULL)
                return;

	if (pmap->channel == NULL) {
        	KFREE(pmap, "mmap");
        	vma->vm_private_data = NULL;
		return;
	}
	/* code below is for V4L2 only.  standard SDK sets channel to NULL */
	channel = pmap->channel;
	dev = pmap->dev;
	bChan = channel->bChan;

	if (pmap->bufId < channel->numbufs) {
		if (pmap->type == CAP_JPG) {
			if (pmap->bStat) {
				count = --(dev->sysBufJ[bChan][pmap->bufId].status.map_count);
			} else {
				count = --(dev->sysBufJ[bChan][pmap->bufId].image.map_count);
			}
		} else {
			if (pmap->bStat) {
				count = --(dev->sysBufB[bChan][pmap->bufId].status.map_count);
			} else {
				count = --(dev->sysBufB[bChan][pmap->bufId].image.map_count);
			}
		}
	}
        dprintk(4, "VMclose: channel %d %s %s buf %d; count now %d\n", 
			channel->bChan, 
			pmap->type == CAP_JPG ? "JPEG" : "BMP",
			pmap->bStat ? "status" : "image",
			pmap->bufId, pmap->count);

	if (--pmap->count == 0) {
        	/* free the mapping*/
        	KFREE(pmap, "mmap");
        	vma->vm_private_data = NULL;
	}


	return;
}

#if LINUX_VERSION_CODE < VERSION_NOPAGE_REMOVED
static struct page *s817_vm_nopage(struct vm_area_struct *vma,
				   unsigned long address, int *type)
{
	struct page *page = NULL;//NOPAGE_SIGBUS;
	struct s817_mapping *pmap;
	unsigned long offset;

	pmap = (struct s817_mapping *) vma->vm_private_data;
	if (!pmap) {
		printk(KERN_ERR "no map!\n");
		return NOPAGE_OOM;
	}
	if (address > vma->vm_end) {
		printk(KERN_ERR "no remap!\n");
		return NOPAGE_SIGBUS;
	}
	/* calculate the offset*/
	offset = (unsigned long) ((address - vma->vm_start) + pmap->pvirt);

	dprintk(50, "mmap offset %lx addr %lx pvirt %p \n",
		offset,
		address,
		pmap->pvirt);
	page = vmalloc_to_page((void *)  offset);
	if (!page) {
		printk(KERN_ERR "no page!\n");
		return NOPAGE_OOM;
	}
	/* debug only */
	if ((debug > 5) && (address == vma->vm_start)) {
		int bufId, bChan, bStat;
		bChan = pmap->bChan;
		bufId = pmap->bufId;
		bStat = pmap->bStat;
		printk(KERN_INFO "checking %d mmap %lx %lx\n",
		       debug, address, vma->vm_start);
		if (pmap->type == CAP_JPG) {
			/* verify the page */
			if (pmap->dev->dbgPg_J[bChan][bufId][bStat] != page) {
				printk(KERN_ERR "[%d:%d]invalid page J %p %p %lx(offset)\n",
				       bChan,
				       bufId,
				       pmap->dev->dbgVA_J[bChan][bufId][bStat],
				       pmap->pvirt,
				       offset);
				printk(KERN_ERR "dev is %p\n", pmap->dev);
			}
		}
		if (pmap->type == CAP_BMP) {
			/* verify the page */
			if (pmap->dev->dbgPg_B[bChan][bufId][bStat] != page) {
				printk(KERN_ERR "[%d:%d]invalid page B %p %p %p %lx(offset)\n",
				       bChan,
				       bufId,
				       pmap->dev->dbgVA_J[bChan][bufId][bStat],
				       pmap->dev->dbgVA_B[bChan][bufId][bStat],
				       pmap->pvirt,
				       offset);
			}
		}
	}

	dprintk(50, "before get page\n");

	get_page(page);
	dprintk(50, "after get page\n");
	if (type)
		*type = VM_FAULT_MINOR;

	dprintk(4, "got offset %lx %lx\n", vma->vm_start, vma->vm_pgoff);
	return page;
}

/* virtual memory operations*/
static struct vm_operations_struct s817_vm_ops = {
	.open = s817_vm_open,
	.close = s817_vm_close,
	.nopage = s817_vm_nopage,
};
#else
/* 2.6.27 changes .nopage to .fault.  */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
static int s817_vm_fault(struct vm_fault *vmf)
#else
static int s817_vm_fault(struct vm_area_struct *vma,
				  struct vm_fault *vmf)
#endif
{
	struct page *page = NULL;//NOPAGE_SIGBUS;
	struct s817_mapping *pmap;
	unsigned long offset;
	unsigned long virtual_address;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
	struct vm_area_struct *vma = vmf->vma;
#endif
	

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)
	virtual_address = vmf->address;
#else
	virtual_address = (unsigned long) vmf->virtual_address;
#endif
	

	pmap = (struct s817_mapping *) vma->vm_private_data;
	if (!pmap) {
		printk(KERN_ERR "no map!\n");
		return VM_FAULT_OOM;
	}
	/* calculate the offset*/
	offset = (unsigned long) (((char *)virtual_address -
				   (char *)vma->vm_start)
				  + pmap->pvirt);

	page = vmalloc_to_page((void *)  offset);
	if (!page) {
		printk(KERN_ERR "no page!\n");
		return VM_FAULT_OOM;
	}

	/* debug only */
	if ((debug > 5) && (virtual_address == (unsigned long) vma->vm_start)) {
		int bufId, bChan, bStat;
		bChan = pmap->bChan;
		bufId = pmap->bufId;
		bStat = pmap->bStat;
		printk(KERN_INFO "checking %d mmap %lx %lx\n",
		       debug, virtual_address, vma->vm_start);
		if (pmap->type == CAP_JPG) {
			/* verify the page */
			if (pmap->dev->dbgPg_J[bChan][bufId][bStat] != page) {
				printk(KERN_ERR "[%d:%d]invalid page J %p %p %lx(offset)\n",
				       bChan,
				       bufId,
				       pmap->dev->dbgVA_J[bChan][bufId][bStat],
				       pmap->pvirt,
				       offset);
				printk(KERN_ERR "dev is %p\n", pmap->dev);
			}
		}
		if (pmap->type == CAP_BMP) {
			/* verify the page */
			if (pmap->dev->dbgPg_B[bChan][bufId][bStat] != page) {
				printk(KERN_ERR "[%d:%d]invalid page B %p %p %p %lx(offset)\n",
				       bChan,
				       bufId,
				       pmap->dev->dbgVA_J[bChan][bufId][bStat],
				       pmap->dev->dbgVA_B[bChan][bufId][bStat],
				       pmap->pvirt,
				       offset);
			}
		}
	}
	get_page(page);
	vmf->page = page;
	dprintk(4, "got offset %lx %lx\n", vma->vm_start, vma->vm_pgoff);
	return 0;
}

/* virtual memory operations*/
static struct vm_operations_struct s817_vm_ops = {
	.open = s817_vm_open,
	.close = s817_vm_close,
	.fault = s817_vm_fault,
};
#endif

static int s817_mmap_common(struct s817_dev *dev, struct s817_v4l2 *channel, struct vm_area_struct *vma)
{
	size_t size;
	void *addr;
	unsigned long boff;
	int j,i;
	int bFound = 0;
	int bStat = 0;
	struct s817_mapping *pmap;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
	lock_kernel();
#endif
	addr = (void *) (vma->vm_pgoff << PAGE_SHIFT);
	boff = (unsigned long)  (vma->vm_pgoff << PAGE_SHIFT);
	size = vma->vm_end - vma->vm_start;
	dprintk(5, "mmap pvirt %p %lu, 0x%lx vm_pgoff 0x%lx\n", addr, 
		(unsigned long) size,
		vma->vm_pgoff, vma->vm_pgoff << PAGE_SHIFT);

	pmap = KZALLOC(sizeof(struct s817_mapping), GFP_KERNEL, "mmap");
	if (pmap == NULL) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
		unlock_kernel();
#endif
		return -ENOMEM;
	}

	vma->vm_ops = &s817_vm_ops;

	for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
		for (j = 0; j < SYS_BUF_PER_CHAN; j++) {
			if (dev->sysBufJ[i][j].image.boff == boff) {
				bFound = CAP_JPG;
				bStat = 0;
				pmap->pvirt = dev->sysBufJ[i][j].image.pData;
				break;
			}
			if (dev->sysBufJ[i][j].status.boff == boff) {
				bFound = CAP_JPG;
				bStat = 1;
				pmap->pvirt = dev->sysBufJ[i][j].status.pData;
				break;
			}
			if (dev->sysBufB[i][j].image.boff == boff) {
				bFound = CAP_BMP;
				bStat = 0;
				pmap->pvirt = dev->sysBufB[i][j].image.pData;
				break;
			}
			if (dev->sysBufB[i][j].status.boff == boff) {
				bFound = CAP_BMP;
				bStat = 1;
				pmap->pvirt = dev->sysBufB[i][j].status.pData;
				break;
			}
		}
		if (bFound)
			break;
	}
	if (bFound == 0) {
		kfree(pmap);
		printk(KERN_ERR "s817_mmap: invalid address offset %p, debug %p %p %p %p\n", 
		       addr,
		       dev->sysBufB[SYS_CHAN_PER_BOARD-1][SYS_BUF_PER_CHAN-1].status.pData,
		       dev->sysBufJ[SYS_CHAN_PER_BOARD-1][SYS_BUF_PER_CHAN-1].status.pData,
		       dev->sysBufB[SYS_CHAN_PER_BOARD-1][SYS_BUF_PER_CHAN-1].image.pData,
		       dev->sysBufJ[SYS_CHAN_PER_BOARD-1][SYS_BUF_PER_CHAN-1].image.pData);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
		unlock_kernel();
#endif
		return -EINVAL;
	}
	dprintk(5, "mmap i %d j %d bfound %d\n", i, j, bFound);
	pmap->size = size;
	vma->vm_flags |= VM_RESERVED | VM_DONTEXPAND;
	vma->vm_private_data = pmap;

	pmap->bChan = i;
	pmap->bufId = j;
	pmap->bStat = bStat;
	pmap->type = bFound;
	pmap->dev = dev;
	pmap->channel = channel;
	dprintk(5, "mmap pvirt %p %lu \n", addr, (unsigned long) size);
	s817_vm_open(vma);
	dprintk(5, "mmap done\n");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
	unlock_kernel();
#endif
	return 0;
}


/* map kernel buffers to user space*/
static int s817_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct s817_dev *dev = (struct s817_dev *) file->private_data;
	int rc;
	rc = s817_mmap_common(dev, NULL, vma);
	return rc;
}





static unsigned int s817_poll(struct file *file, poll_table *pt)
{
	struct s817_dev *dev = (struct s817_dev *) file->private_data;
	dprintk(125, "%s %p\n", __func__, dev);
	if (getDataMask(dev)) {
		return POLLIN;
	}
	poll_wait(file, &dev->wq_done, pt);
	if (getDataMask(dev))
		return POLLIN;
	return 0;
}


static struct pci_driver s817_driver = {
	.name = "s817_0",
	.id_table = s817_pci_ids,
	.probe = s817_probe,
	.remove = __devexit_p(s817_remove)
};





struct file_operations s817_fops = {
	.owner = THIS_MODULE,
#if LINUX_VERSION_CODE < VERSION_UNLOCKED_IOCTL
	.ioctl = s817_ioctl,
#else
	.unlocked_ioctl = s817_ioctl,
#endif
	.open  = s817_open,
	.poll  = s817_poll,
	.mmap  = s817_mmap,
	.release = s817_release,
};



static int __init s817_module_init(void)
{
	int retval;
	int j,k;
	dprintk(1, "s817 module init ...\n");
	printk(KERN_INFO "Sensoray 817 driver version %s\n",
	       S817_VERSION_TEXT);

	num_boards = 0;

	for (k = 0;k < SYS_DSP_PER_BOARD; k++) {
		for (j = 0;j < MAX_BOARDS; j++) {
			atomic_set(&s817_tasklet[k][j].count, 0);
			s817_tasklet[k][j].state = 0;
			s817_tasklet[k][j].data = 0;
			s817_tasklet[k][j].func = s817_dpc;
			s817_tasklet[k][j].next = NULL;
		}
	}

#ifdef S817_V4L2
#ifdef CONFIG_PROC_FS
	//s817_proc_root = proc_mkdir("s817", NULL);
	printk("create proc root\n");
	s817_proc_root = create_proc_entry("s817", 
					   S_IFDIR | S_IRWXU | S_IRWXG |S_IRWXO,
					   NULL);
	
	if (s817_proc_root)
	{
		struct proc_dir_entry	*ent;
// TODO find out exactly when this was removed.
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 29)
		s817_proc_root->owner = THIS_MODULE;
#endif
		
		ent = create_proc_entry("fwVer", S_IFREG | S_IRUGO, 
					s817_proc_root);
		if (ent) {
			int 	dirmode = S_IFDIR | S_IRWXU | S_IRWXG | S_IRWXO;

			ent->data = NULL;
			ent->read_proc = read_fwver;
			ent->write_proc = NULL;
	
			s817_procout_root = create_proc_entry("output", 
						dirmode, s817_proc_root);
		}
	}
	else
	{
		printk(KERN_ERR "s817: unable to create proc root\n");
	}
#endif
#endif
	retval = pci_register_driver(&s817_driver);

	if (retval) {
		printk(KERN_ERR "s817: Can't register pci driver\n");
		return -ENODEV;
	}
	dprintk(2, "done s817_module_init\n");
	if (register_chrdev(major, "s817",  &s817_fops)) {
		printk(KERN_ERR "s817: failed to register char dev\n");
		return -ENODEV;
	}

	return 0;
}

static void __exit s817_module_exit(void)
{
	dprintk(2, "Unloading module s817 ...\n");
	pci_unregister_driver(&s817_driver);
	unregister_chrdev(major, "s817_0");
#ifdef S817_V4L2
#ifdef CONFIG_PROC_FS
	remove_proc_entry("s817", NULL);
#endif
#endif
}

module_init(s817_module_init);
module_exit(s817_module_exit);

/*2 kinds of commands: long and short*/
/*Long command: data is placed into a command buffer*/
/*physAddr: physical (PCI) address of the command buffer*/
/*(lower 2 bits of physAddr are set to DSP channel (0..3)
  that the command is addressed to*/
/*by the function that calls sendDspCmd)*/
/**/
/*Short command: the argument (shortArg) is passed via HDCR_SC_ARG*/
/*physAddr (upper bits) must be 0*/
/**/
/*returns error code */
/*Error codes received from DSP (via HDCR_FLAG) are >= 500*/
/*(those mean DSP has problems with the command: PCI errors,
  command formatting errors)*/
/*in case of a timeout the function returns ERR_DSP_CMD_TO*/
int sendDspCmd(struct s817_dev *dev, int dsp, unsigned long physAddr,
	       unsigned long shortArg)
{
	int	b_timeout = 0;
	int	to = 0;
	int	retval;
	unsigned long rv;
	/*Exclude other threads*/
	/*check if we can send a command*/
	/*HDCR_COMMAND must be 0; */
//	printk("send dsp command %d %d\n", dsp, physAddr & 0x0f);
	if (s817_readDSPreg(dev->dsp[dsp].log_bar1, HDCR_COMMAND) & ~0xf) {
		/*got stuck on previous command*/
		retval = ERR_DSP_CMD_BUSY;
	} else {
		if ((physAddr & ~DSP_SHORT_CMD_MSK) == 0) {
			/*if short command (PCI address not set)*/
			s817_writeDSPreg(dev->dsp[dsp].log_bar1,
					 HDCR_SC_ARG, shortArg);
		}
		/*write the PCI address of the command buffer into
		  HDCR_COMMAND*/
		s817_writeDSPreg(dev->dsp[dsp].log_bar1, HDCR_COMMAND,
				 physAddr);
		/*wait until DSP has finished processing command*/
		while (!b_timeout) {
			rv = s817_readDSPreg(dev->dsp[dsp].log_bar1,
					     HDCR_COMMAND);
			if ((rv & ~0xf) || (rv == 0)) {
				if ((rv & 0xffffffff) != (physAddr & 0xffffffff)) {
					dprintk(4, "s817: senddspcmd read "
						"%lx expected %lx\n       ",
						rv, physAddr);
				}
				/*sleep*/
				msleep(SYS_DSP_CMD_SLEEP);
				/*increment counter*/
				to += SYS_DSP_CMD_SLEEP;
				/*set timeout var*/
				b_timeout = (to >= SYS_DSP_CMD_TO);
			} else {
				break; /*DSP ready*/
			}
		}

		if (!b_timeout) {
			/*read error code from DSP*/
			retval = s817_readDSPreg(dev->dsp[dsp].log_bar1,
						 HDCR_FLAG);
			if (retval != 0) {
				printk(KERN_ERR "s817: dsp error %x\n", retval);
				dev->dsp[dsp].dsp_errs++;
				dev->dsp[dsp].last_dsp_err = retval;
			}
		} else {
			dev->dsp[dsp].timeouts++;
			retval = ERR_DSP_CMD_TO;
		}

	}
	return retval;
}


static void translateCMode(CAPTURE_MODE *inMode, CAPTURE_MODE *outMode)
{
	/*handle parameter translation here*/
	/*copy inMode to outMode*/
	*outMode = *inMode;
	/*translate compQuality: 0..100 translates to 10..80:
	  out = 0.7*in + 10*/
	outMode->compQuality = outMode->compQuality * 7 / 10 + 10;
}


int S817_SetVideoMode(struct s817_dev *dev, int aChan, VIDEO_MODE *pVMode)
{
	struct dsp_cmd_hdr dspCmdHdr;
	ECODE		   ecode;
	char		   *buf;
	int		   bChan = aChan - 1; /*channel in board context*/
	int		   dsp = bChan / SYS_DSP_PER_BOARD;
	/*channel in DSP context*/
	int		   dChan = bChan % SYS_CHAN_PER_DSP;
	int		   cmdCode;
	int		   modeChgMask = 0;

	if ((unsigned) aChan > SYS_CHAN_PER_BOARD || !dev->dsp[dsp].pdev)
		return ERR_ILLEGAL_CHAN;
	/*check if valid parameters*/
	if (pVMode->tvFormat < TV_NTSC || pVMode->tvFormat > TV_SECAM)
		return ERR_ILLEGAL_TVFORMAT;
	if (pVMode->jpgSize < SIZE_4CIF || pVMode->jpgSize > SIZE_4CIFI)
		return ERR_ILLEGAL_JPGSIZE;

	/*check if restart required on the DSP video port*/
	/*(restart required if format or size are changed)*/
	if ((pVMode->tvFormat != dev->mode[bChan].vMode.tvFormat) ||
	    (pVMode->jpgSize != dev->mode[bChan].vMode.jpgSize))
		cmdCode = DSP_CMD_VMODE;
	else
		cmdCode = DSP_CMD_VMODE_FAST;

	/*set mask bits for parameters that have changed*/
	if (pVMode->brightness != dev->mode[bChan].vMode.brightness)
		modeChgMask |= MODE_CHG_BRIGHTNESS;
	if (pVMode->contrast != dev->mode[bChan].vMode.contrast)
		modeChgMask |= MODE_CHG_CONTRAST;
	if (pVMode->saturation != dev->mode[bChan].vMode.saturation)
		modeChgMask |= MODE_CHG_SATURATION;
	if (pVMode->hue != dev->mode[bChan].vMode.hue)
		modeChgMask |= MODE_CHG_HUE;

	MUTEX_LOCK(&dev->dsp[dsp].dsplock);
	if (!dev->dsp[dsp].cOpen[dChan]) {
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return ERR_CHAN_NOTOPEN;
	}

	buf = dev->dsp[dsp].cmd[dChan].pData;	/*use CMD buffer*/
	dspCmdHdr.marker = DSP_CMD_MARKER;
	dspCmdHdr.cmdCode = cmdCode;
	dspCmdHdr.arg0 = modeChgMask;
	dspCmdHdr.arg1 = 0;
	/*channel in the DSP context: 0 to 3*/
	dspCmdHdr.channel = dChan;
	dspCmdHdr.dataSize = sizeof(VIDEO_MODE);
	*(struct dsp_cmd_hdr *) buf = dspCmdHdr;
	/*copy MODE into buffer*/
	*(VIDEO_MODE *) (buf + sizeof(struct dsp_cmd_hdr)) = *pVMode;

	if (dev->dsp[dsp].cmd[dChan].pageTbl.pPT[0] & 0x0f) {
		dev->err_physaddr++;
#ifdef DEBUG_PT
		printk(KERN_ERR "invalid physical address\n");
#endif
	}

	ecode = sendDspCmd(dev, dsp,
			   dev->dsp[dsp].cmd[dChan].pageTbl.pPT[0] | dChan,
			   0);

	if (!ecode)
		dev->mode[bChan].vMode = *pVMode; /*save new mode as current*/
	MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
	return ecode;
}

static int S817_SetCaptureMode(struct s817_dev *dev, int aChan,
			       CAPTURE_MODE *pCMode)
{
	ECODE	      ecode;
	CAPTURE_MODE  bufCMode;
	long	      lBuf0, lBuf1;
	int	      bChan = aChan - 1;	/*channel in board context*/
	int	      dsp = bChan / SYS_DSP_PER_BOARD;
	/*channel in DSP context*/
	int	      dChan = bChan % SYS_CHAN_PER_DSP;
	int	      modeChgMask = 0;
	dprintk(4, "set capture mode DSP %d channel %d"
		" channel(dsp context) %d\n", dsp, bChan, dChan);


	if ((unsigned) aChan > SYS_CHAN_PER_BOARD || !dev->dsp[dsp].pdev)
		return ERR_ILLEGAL_CHAN;

	if ((pCMode->compQuality < COMP_QUAL_MIN) ||
	    (pCMode->compQuality > COMP_QUAL_MAX))
		return ERR_ILLEGAL_COMPQUALITY;
	if ((unsigned) pCMode->frameRateBmp >= (unsigned) FR_MAX)
		return ERR_ILLEGAL_FRAMERATE;
	if ((unsigned) pCMode->frameRateJpg >= (unsigned) FR_MAX)
		return ERR_ILLEGAL_FRAMERATE;
	if ((unsigned) pCMode->capMode >= (unsigned) CAP_MODE_MAX)
		return ERR_ILLEGAL_CAPMODE;
	/*set mask bits for parameters that have changed*/
	if (pCMode->frameRateBmp != dev->mode[bChan].cMode.frameRateBmp)
		modeChgMask |= MODE_CHG_FRAMERATE;
	if (pCMode->frameRateJpg != dev->mode[bChan].cMode.frameRateJpg)
		modeChgMask |= MODE_CHG_FRAMERATE;
	if (pCMode->compQuality != dev->mode[bChan].cMode.compQuality)
		modeChgMask |= MODE_CHG_COMPQUALITY;
	if (pCMode->capMode != dev->mode[bChan].cMode.capMode)
		modeChgMask |= MODE_CHG_CAPMODE;

	MUTEX_LOCK(&dev->dsp[dsp].dsplock);
	if (!dev->dsp[dsp].cOpen[dChan]) {
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return ERR_CHAN_NOTOPEN;
	}

	if (pCMode->capType != dev->mode[bChan].cMode.capType) {
		modeChgMask |= MODE_CHG_CAPTYPE;
		if (dev->mode[bChan].cMode.capType == CAP_OFF) {		//if capture was off - release all buffers
			releaseBufs(dev, dsp, dChan, (1 << SYS_BUF_PER_CHAN) - 1, CAP_BMP);	//show all buffers as available
			releaseBufs(dev, dsp, dChan, (1 << SYS_BUF_PER_CHAN) - 1, CAP_JPG);	//show all buffers as available
			//printk(KERN_INFO "capture mode off\n");
		}
	}

	/*some parameters may need translation before being sent
	  (currently compQuality)*/
	translateCMode(pCMode, &bufCMode);
	/*SetCaptureMode is sent as a short command*/
	/*the extra argument arg holds CAPTURE_MODE members:*/
	/*arg[31:28] = capMode & 0xf*/
	/*arg[27:24] = capType & 0xf*/
	/*arg[23:20] = bmpField & 0xf;*/
	/*arg[19:16] = not used*/
	/*arg[15:08] = compQuality & 0xff*/
	/*arg[07:00] = ((frameRateJpg<<4) | frameRateBmp) & 0xff*/
	lBuf0 = ((bufCMode.capMode & 0xf) << 28) |
		((bufCMode.capType & 0xf) << 24) |
		((bufCMode.compQuality & 0xff) << 8) |
		((bufCMode.frameRateJpg & 0xf) << 4) |
		(bufCMode.frameRateBmp & 0xf);
	/*Short command: physAddr argument has 0's in bits 0xfffff000,
	  command code in bits 0x00000ff0 and dChan in bits 0x0000000f*/
	lBuf1 = (DSP_CMD_CMODE << 4) | dChan;
	/*send command*/
	ecode = (ECODE) sendDspCmd(dev, dsp, lBuf1, lBuf0);
	
	/* save new mode as current*/
	if (!ecode) {
		int j;
		int dsp_tmp;
		/* channel in DSP context */
		int dchan_tmp;
		int dsp_onoff[SYS_DSP_PER_BOARD];
		dev->mode[bChan].cMode = *pCMode;

		memset(dsp_onoff, 0, SYS_DSP_PER_BOARD * sizeof(int));

		for (j = 0; j < SYS_CHAN_PER_BOARD; j++) {
			dsp_tmp = j / SYS_DSP_PER_BOARD;
			dchan_tmp = j % SYS_CHAN_PER_DSP;
			if (dev->mode[j].cMode.capType != CAP_OFF)
				dsp_onoff[dsp_tmp] = 1;

		}
	}
	MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
	return ecode;
}


/*Input arguments:
 *board ID;
 *array of 4 int values: input channel (aChan, 1..16)
 to be connected to outputs 0, 1, 2, 3 respectively. */

static int S817_SetVideoSwitch(struct s817_dev *dev, int *switchOut)
{
	int		iBuf;
	long		lBuf0, lBuf1;
	int		i;
	/* this command has to go to DSP0 */
	int		dsp = dev->dsp0 - 1;
	int		dChan;

	/*convert input args into 1 32 bit */
	/*(inp args are input chans (0-15) selected for each of 4 outs */

	lBuf0 = 0;
	for (i = 0; i < 4; i++) {
		if (switchOut[i] < 1 || switchOut[i] > SYS_CHAN_PER_BOARD)
			return ERR_ILLEGAL_CHAN;
		lBuf0 |= ((char) (switchOut[i] - 1) << (i * 8));
	}

	dChan = 0;
	MUTEX_LOCK(&dev->dsp[dsp].dsplock);
	if (!dev->dsp[dsp].cOpen[dChan]) {
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return ERR_CHAN_NOTOPEN;
	}
	/*Short command: physAddr argument has 0's in bits 0xfffff000,
	  command code in bits 0x00000ff0 and dChan in bits 0x0000000f */
	/*For DSP_CMD_SET_VSWITCH dChan = 0. Extra argument is lBuf0 */
	lBuf1 = (DSP_CMD_SET_VSWITCH << 4) | dChan;
	/* send command */
	iBuf = sendDspCmd(dev, dsp, lBuf1, lBuf0);
	MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
	return (ECODE) iBuf;
}

/* outEn: bits [3:0] enable video out[3:0] if 1, disable if 0 */
static int S817_SetVideoOut(struct s817_dev *dev, unsigned int outEn)
{
	int		iBuf;
	long		lBuf;
	/* this command has to go to DSP0 */
	int		dsp = dev->dsp0 - 1;
	int			dChan;
	dChan = 0;	
	MUTEX_LOCK(&dev->dsp[dsp].dsplock);
	if (!dev->dsp[dsp].cOpen[dChan]) {
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return ERR_CHAN_NOTOPEN;
	}
	/* Short command: physAddr argument has 0's in bits 0xfffff000,
	  command code in bits 0x00000ff0 and dChan in bits 0x0000000f */
	/* For DSP_CMD_SET_VSWITCH dChan = 0. Extra argument is outEn */

	lBuf = (DSP_SET_VOUT << 4) | dChan;
	/* send command */
	iBuf = sendDspCmd(dev, dsp, lBuf, outEn);
	MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
	return (ECODE) iBuf;
}

static int S817_SetDateTime(struct s817_dev *dev, struct param_settime *strTm)
{
	ECODE ecode;
	int i;
	for (i = 0; i < SYS_DSP_PER_BOARD; i++) {
		ecode = S817_SetDateTimeDsp(&dev->dsp[i], strTm);
		if (ecode != 0)
			break;
		
	}
	return ecode;

}

/* param_settime should match POSIX struct tm */
static int S817_SetDateTimeDsp(struct s817_dsp *dsp, struct param_settime *strTm)
{
	struct dsp_cmd_hdr	dspCmdHdr;
	ECODE			ecode;
	char			*buf;
	int			dChan = 0;
	struct sx_page_table *pageTbl;
	int bChan;
	/* send to each DSP */
	if (strTm == NULL)
		return -EINVAL;
	if (!dsp->pdev)
		return 0;
	/* use CMD buffer */
	bChan = dsp->dsp_num * SYS_CHAN_PER_DSP + dChan;
	MUTEX_LOCK(&dsp->dsplock);
	if (!dsp->cOpen[dChan]) {
		MUTEX_UNLOCK(&dsp->dsplock);
		return 0;
	}
	buf = dsp->cmd[dChan].pData; /*use cmd buffer */
	if (buf == NULL) {
		MUTEX_UNLOCK(&dsp->dsplock);
		return 0;
	}
	dspCmdHdr.marker = DSP_CMD_MARKER;
	dspCmdHdr.cmdCode = DSP_SET_DATETIME;
	dspCmdHdr.arg0 = 0;
	dspCmdHdr.arg1 = 0;
	/* channel in the DSP context: 0 to 3 */
	dspCmdHdr.channel = dChan;
	dspCmdHdr.dataSize = sizeof(struct param_settime);
	*(struct dsp_cmd_hdr *) buf = dspCmdHdr;
	/* copy to payload */
	*(struct param_settime *)(buf + sizeof(struct dsp_cmd_hdr))
		= *strTm;
	pageTbl = &dsp->cmd[dChan].pageTbl;
	
	if (pageTbl == NULL) {
		MUTEX_UNLOCK(&dsp->dsplock);
		return 0;
	}

	if (pageTbl->pPT[0] & 0x0f) {
		dsp->dev->err_physaddr++;
#ifdef DEBUG_PT
		printk(KERN_ERR "invalid physical address\n");
#endif
	}
	
	ecode = sendDspCmd(dsp->dev, dsp->dsp_num,
			   pageTbl->pPT[0] | dChan,
			   0);
	MUTEX_UNLOCK(&dsp->dsplock);
	return ecode;
}

/* sets OSD mode for a specific board/channel combination */
static int S817_SetOsdMode(struct s817_dev *dev, int aChan, OSD_MODE *osd)
{
	struct dsp_cmd_hdr	dspCmdHdr;
	ECODE		ecode;
	/* channel in board context */
	int		bChan = aChan - 1;
	int		dsp = bChan / SYS_DSP_PER_BOARD;
	/* channel in DSP context */
	int		dChan = bChan % SYS_CHAN_PER_DSP;
	char				*buf;
	/* check if valid input arguments */
	if ((unsigned) aChan > SYS_CHAN_PER_BOARD || !dev->dsp[dsp].pdev)
		return ERR_ILLEGAL_CHAN;

	if (dev == NULL)
		return -EINVAL;

	/* check if valid parameters and fix */
	if (osd->fraction < 0)
		osd->fraction = 0;
	else if (osd->fraction > 2)
		osd->fraction = 2;

	/* use CMD buffer */
	MUTEX_LOCK(&dev->dsp[dsp].dsplock);
	if (!dev->dsp[dsp].cOpen[dChan]) {
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return -EINVAL;
	}

	buf = dev->dsp[dsp].cmd[dChan].pData;

	if (buf == NULL) {
		printk(KERN_ERR "S817_SetOsdMode: null buffer\n");
		MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
		return -EINVAL;
	}
	dspCmdHdr.marker = DSP_CMD_MARKER;
	dspCmdHdr.cmdCode = DSP_SET_OSDMODE;
	dspCmdHdr.arg0 = 0;
	dspCmdHdr.arg1 = 0;
	/* channel in the DSP context: 0 to 3 */
	dspCmdHdr.channel = dChan;
	dspCmdHdr.dataSize = sizeof(OSD_MODE);
	*(struct dsp_cmd_hdr *) buf = dspCmdHdr;
	/* copy to payload */
	*(OSD_MODE *)(buf + sizeof(struct dsp_cmd_hdr)) = *osd;
	if (dev->dsp[dsp].cmd[dChan].pageTbl.pPT[0] & 0x0f) {
		dev->err_physaddr++;
#ifdef DEBUG_PT
		printk(KERN_ERR "invalid physical address\n");
#endif
	}

	ecode = (ECODE) sendDspCmd(dev, dsp,
				   (dev->dsp[dsp].cmd[dChan].pageTbl.pPT[0] |
				    dChan), 0);
	MUTEX_UNLOCK(&dev->dsp[dsp].dsplock);
	if (ecode)
		return ecode;
	// save the OSD mode
	dev->osdmode[bChan] = *osd;
	return 0;
}

#if LINUX_VERSION_CODE >= CREATE_PROC_REMOVED
// handled by seq_read
static int s817_show_proc_output(struct seq_file *s, void *unused)
{
	static EXTERR extErr[SYS_EXTERR_BUF_SIZE];
	int i, j;
	int rc;
	struct s817_dev *h = s->private;

	if (h == NULL)
		return 0;
	seq_printf(s, "===Sensoray 817===\n");
	seq_printf(s, "brd %d (phys %ld)\n", h->minor, h->err_physaddr);
	for (i = 0;i < SYS_DSP_PER_BOARD; i++) {
		if (!h->dsp[i].pdev) {
			seq_printf(s, "dsp %d not present\n", i);
			continue;
		}
		seq_printf(s, "dsp %d\n", i);
		seq_printf(s, "  fifoFull %ld,", h->dsp[i].fifoFull);
		seq_printf(s, "  fifoWrErr %ld,", h->dsp[i].fifoWriteErr);
		seq_printf(s, "  dpcs %ld, %ld, %ld, %ld", h->dsp[i].num_dpcs[0],
			     h->dsp[i].num_dpcs[1],
			     h->dsp[i].num_dpcs[2],
			     h->dsp[i].num_dpcs[3]);
		seq_printf(s, "  ints %ld,", h->dsp[i].intCounter);
		seq_printf(s, "  (shared) %ld,", h->dsp[i].intCounterShared);
		seq_printf(s, "  (dbg) %ld, %ld, %ld, %ld\n",
			     h->dsp[i].intCounterNotOurs,
			     h->dsp[i].intCounterNull,
			     h->dsp[i].intCounterLogBar,
			     h->dsp[i].intCounterNotLoaded);

		seq_printf(s, "  dsp errs %ld [last %ld], nomark %ld, ",
			     h->dsp[i].dsp_errs,
			     h->dsp[i].last_dsp_err,
			     h->dsp[i].nomarker);
		seq_printf(s, "  timeouts %ld\n", h->dsp[i].timeouts);
		seq_printf(s, "  cOpen[0..3] %d %d %d %d\n", h->dsp[i].cOpen[0],
			     h->dsp[i].cOpen[1],
			     h->dsp[i].cOpen[2],
			     h->dsp[i].cOpen[3]);
		for (j = 0; j < 5; j++) {

			int errors;
			int k;
			rc = getExtErr(h, i, j, extErr);
			if (rc != 0)
				continue;
				
			errors = extErr[0].tick;
			seq_printf(s, "  extErrs[%d]: errs %d code %d, rsvd %x %x",
				     j, errors,
				     extErr[0].ecode,
				     extErr[0].reserved[0],
				     extErr[0].reserved[1]);
			if ((j==1) || (j==3) || (j==4))
				seq_printf(s, "\n");
			else
				seq_printf(s, ",");
			// only display 4 errors max proc_output 
			// should fit on one page
			errors = (errors < 4) ? errors : 4;
			for (k = 0; k < errors; k++) {
				seq_printf(s,
					     "exterr: %08X  %04d  %08X  %08X\n",
					     extErr[k + 1].tick,
					     extErr[k + 1].ecode,
					     extErr[k + 1].reserved[0],
					     extErr[k + 1].reserved[1]);	
			}
		}
		rc = GetSysData(&h->dsp[i]);
		if (rc != 0) {
			continue;
		}

		seq_printf(s, "sysdata: extErr 0x%x, dspA 0x%x, pciRd %d, pciWr %d\n",
			     h->dsp[i].sysData.extErrAddr,
			     h->dsp[i].sysData.dspAddr,
			     h->dsp[i].sysData.inPciRead,
			     h->dsp[i].sysData.inPciWrite);
		seq_printf(s, "hstLate %d, rsrvd %d 0x%x 0x%x 0x%x"
			     " 0x%x 0x%x 0x%x 0x%x\n",
			     h->dsp[i].sysData.hostLate,
			     h->dsp[i].sysData.reserved[0],
			     h->dsp[i].sysData.reserved[1],
			     h->dsp[i].sysData.reserved[2],
			     h->dsp[i].sysData.reserved[3],
			     h->dsp[i].sysData.reserved[4],
			     h->dsp[i].sysData.reserved[5],
			     h->dsp[i].sysData.reserved[6],
			     h->dsp[i].sysData.reserved[7]);
		for (j = 0; j < 4; j++) {
			seq_printf(s, "chan %d lastCmd:%d, tick: %d, sleep: %d "
				     "lockPend: %d, ",
				     j,
				     h->dsp[i].sysData.chanData[j].lastCmd,
				     h->dsp[i].sysData.chanData[j].chanTick,
				     h->dsp[i].sysData.chanData[j].putToSleep,
				     h->dsp[i].sysData.chanData[j].lockPend);
			seq_printf(s, "vid %d, noFreeJ %d, noFreeB %d\n",
				     h->dsp[i].sysData.chanData[j].fvidStat,
				     h->dsp[i].sysData.chanData[j].noFreeBufJpg,
				     h->dsp[i].sysData.chanData[j].noFreeBufBmp);
		}
		
	}
	return 0;
}
#else

#ifndef CONFIG_PROC_FS
static int s817_read_procmem(char *buf, char **start, off_t offset,
			     int count, int *eof, void *data)
{
	printk(KERN_INFO "proc fs not configured\n");
	return 0;
}
#else
int s817_proc_output(char *buf, struct s817_dev *h)
{
	char *p = buf;
	static EXTERR extErr[SYS_EXTERR_BUF_SIZE];
	int i, j;
	int rc;

	if (h == NULL)
		return 0;

	p += sprintf(p, "===Sensoray 817===\n");
	p += sprintf(p, "brd %d (phys %ld)\n", h->minor, h->err_physaddr);
	for (i = 0;i < SYS_DSP_PER_BOARD; i++) {
		if (!h->dsp[i].pdev) {
			p += sprintf(p, "dsp %d not present\n", i);
			continue;
		}
		p += sprintf(p, "dsp %d\n", i);
		p += sprintf(p, "  fifoFull %ld,", h->dsp[i].fifoFull);
		p += sprintf(p, "  fifoWrErr %ld,", h->dsp[i].fifoWriteErr);
		p += sprintf(p, "  dpcs %ld, %ld, %ld, %ld", h->dsp[i].num_dpcs[0],
			     h->dsp[i].num_dpcs[1],
			     h->dsp[i].num_dpcs[2],
			     h->dsp[i].num_dpcs[3]);
		p += sprintf(p, "  ints %ld,", h->dsp[i].intCounter);
		p += sprintf(p, "  (shared) %ld,", h->dsp[i].intCounterShared);
		p += sprintf(p, "  (dbg) %ld, %ld, %ld, %ld\n",
			     h->dsp[i].intCounterNotOurs,
			     h->dsp[i].intCounterNull,
			     h->dsp[i].intCounterLogBar,
			     h->dsp[i].intCounterNotLoaded);

		p += sprintf(p, "  dsp errs %ld [last %ld], nomark %ld, ",
			     h->dsp[i].dsp_errs,
			     h->dsp[i].last_dsp_err,
			     h->dsp[i].nomarker);
		p += sprintf(p, "  timeouts %ld\n", h->dsp[i].timeouts);
		p += sprintf(p, "  cOpen[0..3] %d %d %d %d\n", h->dsp[i].cOpen[0],
			     h->dsp[i].cOpen[1],
			     h->dsp[i].cOpen[2],
			     h->dsp[i].cOpen[3]);
		for (j = 0; j < 5; j++) {

			int errors;
			int k;
			rc = getExtErr(h, i, j, extErr);
			if (rc != 0)
				continue;
				
			errors = extErr[0].tick;
			p += sprintf(p, "  extErrs[%d]: errs %d code %d, rsvd %x %x",
				     j, errors,
				     extErr[0].ecode,
				     extErr[0].reserved[0],
				     extErr[0].reserved[1]);
			if ((j==1) || (j==3) || (j==4))
				p += sprintf(p, "\n");
			else
				p += sprintf(p, ",");
			// only display 4 errors max proc_output 
			// should fit on one page
			errors = (errors < 4) ? errors : 4;
			for (k = 0; k < errors; k++) {
				p += sprintf(p,
					     "exterr: %08X  %04d  %08X  %08X\n",
					     extErr[k + 1].tick,
					     extErr[k + 1].ecode,
					     extErr[k + 1].reserved[0],
					     extErr[k + 1].reserved[1]);	
			}
		}
		rc = GetSysData(&h->dsp[i]);
		if (rc != 0) {
			continue;
		}

		p += sprintf(p, "sysdata: extErr 0x%x, dspA 0x%x, pciRd %d, pciWr %d\n",
			     h->dsp[i].sysData.extErrAddr,
			     h->dsp[i].sysData.dspAddr,
			     h->dsp[i].sysData.inPciRead,
			     h->dsp[i].sysData.inPciWrite);
		p += sprintf(p, "hstLate %d, rsrvd %d 0x%x 0x%x 0x%x"
			     " 0x%x 0x%x 0x%x 0x%x\n",
			     h->dsp[i].sysData.hostLate,
			     h->dsp[i].sysData.reserved[0],
			     h->dsp[i].sysData.reserved[1],
			     h->dsp[i].sysData.reserved[2],
			     h->dsp[i].sysData.reserved[3],
			     h->dsp[i].sysData.reserved[4],
			     h->dsp[i].sysData.reserved[5],
			     h->dsp[i].sysData.reserved[6],
			     h->dsp[i].sysData.reserved[7]);
		for (j = 0; j < 4; j++) {
			p += sprintf(p, "chan %d lastCmd:%d, tick: %d, sleep: %d "
				     "lockPend: %d, ",
				     j,
				     h->dsp[i].sysData.chanData[j].lastCmd,
				     h->dsp[i].sysData.chanData[j].chanTick,
				     h->dsp[i].sysData.chanData[j].putToSleep,
				     h->dsp[i].sysData.chanData[j].lockPend);
			p += sprintf(p, "vid %d, noFreeJ %d, noFreeB %d\n",
				     h->dsp[i].sysData.chanData[j].fvidStat,
				     h->dsp[i].sysData.chanData[j].noFreeBufJpg,
				     h->dsp[i].sysData.chanData[j].noFreeBufBmp);
		}
			
	}
	return p - buf;
}

int s817_read_procmem(char *buf, char **start, off_t off, int count,
		      int *eof, void *data)
{
	int len;
	struct s817_dev *h;
	h = (struct s817_dev *) data;

	len = s817_proc_output(buf, h);
        if (len <= off+count)
		*eof = 1;
        *start = buf + off;

        len -= off;
        if (len>count)
		len = count;
        if (len<0)
		len = 0;
        return len;
}

#endif
#endif

// ---------------------------------------------------------------------------
// ported from S.S. V4L2 implementation
// begin V4L2 code. do not mess with the order of functions or reorganize code.
// ---------------------------------------------------------------------------
/********************************  V4L2 ioctls  ******************************/

#ifdef S817_V4L2

static int s817_v4l2_streamoff(struct s817_v4l2 *channel);
static int s817_v4l2_streamon(struct s817_v4l2 *channel);
static const char *s817_driver_name = "s817";
static const char *s817_driver_description = "s817 V4L2";

/* 
 * Wait until a frame is captured, up to 100ms
 */

static int s817_wait_for_frame(struct s817_v4l2 *channel)
{
	wait_queue_t wait;
	int err = 0;

	init_waitqueue_entry(&wait, current);
	/* add ourselves into wait queue */
	add_wait_queue(&channel->frame_wait_queue, &wait);

	schedule_timeout_interruptible(msecs_to_jiffies(100));

	if (signal_pending(current))
		err = -EINTR;

	remove_wait_queue(&channel->frame_wait_queue, &wait);

	return err;
}


//See if the buffer id is on the FIFO
int FIFOfind(struct s817_v4l2 *channel, int id)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	int		capType = dev->mode[bChan].cMode.capType;
	struct fifo	*pFIFO = &dev->bufDataFIFO[bChan][capType - 1];
	int		i;

	i = pFIFO->read;
	while(i != pFIFO->write)
	{
                if (pFIFO->elements[i].bufId == id)
        		return 1;
        	i = (i + 1) % pFIFO->size;
	}
	return 0;
}

/* 
 * returns the channel's number of buffers that are currently mmap'd by a user
 */
static int s817_v4l2_has_mapped_buffers(struct s817_v4l2 *channel)
{
	unsigned int	i;
	int		ret = 0;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	for (i = 0; i < channel->numbufs; i++) {
		ret += dev->sysBufJ[bChan][i].image.map_count;
		ret += dev->sysBufJ[bChan][i].status.map_count;
		ret += dev->sysBufB[bChan][i].image.map_count;
		ret += dev->sysBufB[bChan][i].status.map_count;
	}
	return ret;
}

/*
 * This function is called after changes are mode to a channel's mode.vMode to 
 * send the changes to the board.
 * returns 0 on success, ECODE on error
 */
int s817_v4l2_video_mode(struct s817_v4l2 *channel, VIDEO_MODE *newmode)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	int aChan = bChan + 1;
	int dChan = bChan % SYS_CHAN_PER_DSP;
	int rc;
	if (!dev->dsp[channel->dsp->dsp_num].cOpen[dChan]) {
		s817_openchan(channel->dev, aChan,
			      &channel->dev->mode[channel->bChan]);
	}
	rc = S817_SetVideoMode(dev, aChan, newmode);
	return rc;
}


ECODE s817_set_videoStd(struct s817_v4l2 *channel, int tvFormat)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;

	if (dev->mode[bChan].vMode.tvFormat == tvFormat)
		return 0;

	if ((tvFormat < 0) || (tvFormat >= TV_MAX))
		return -EINVAL;//ERR_ILLEGAL_MODE;

	if (channel->streaming) {
		dprintk(4, "stream is busy\n");
		return -EBUSY;
	}
	dev->mode[bChan].vMode.tvFormat = tvFormat;
	return 0;
}

ECODE s817_set_jpgSize(struct s817_v4l2 *channel, int jpgSize)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	VIDEO_MODE vMode = dev->mode[bChan].vMode;
	if (dev->mode[bChan].vMode.jpgSize == jpgSize)
		return 0;
	if ((jpgSize < 0) || (jpgSize >= SIZE_JPG_MAX))
		return ERR_ILLEGAL_JPGSIZE;
	dprintk(4, "jpg size %d\n", jpgSize);
	vMode.jpgSize = jpgSize;
	return s817_v4l2_video_mode(channel, &vMode);
}
#if 0
ECODE s817_set_bmpType(struct s817_v4l2 *channel, int bmpType)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	VIDEO_MODE vmode = dev->mode[bChan].vMode;
	if ((bmpType < 0) || (bmpType >= BMP_MAX))
		return ERR_ILLEGAL_MODE;

	vmode.bmpType = bmpType;
	return s817_v4l2_video_mode(channel, &vmode);
}
#endif

ECODE s817_set_brightness(struct s817_v4l2 *channel, int brightness)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	VIDEO_MODE vMode = dev->mode[bChan].vMode;
	if (dev->mode[bChan].vMode.brightness == brightness)
		return 0;
	if ((brightness < 0) || (brightness >= 256))
		return ERR_ILLEGAL_MODE;

	vMode.brightness = brightness;
	return s817_v4l2_video_mode(channel, &vMode);
}

int s817_get_brightness(struct s817_v4l2 *channel)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	return dev->mode[bChan].vMode.brightness;
}

ECODE s817_set_contrast(struct s817_v4l2 *channel, int contrast)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	VIDEO_MODE vMode = dev->mode[bChan].vMode;
	if (dev->mode[bChan].vMode.contrast == contrast)
		return 0;
	if ((contrast < 0) || (contrast >= 256))
		return ERR_ILLEGAL_MODE;
	vMode.contrast = contrast;
	return s817_v4l2_video_mode(channel, &vMode);
}

int s817_get_contrast(struct s817_v4l2 *channel)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	return dev->mode[bChan].vMode.contrast;
}

ECODE s817_set_saturation(struct s817_v4l2 *channel, int saturation)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	VIDEO_MODE vMode = dev->mode[bChan].vMode;
	if (dev->mode[bChan].vMode.saturation == saturation)
		return 0;
	if ((saturation < 0) || (saturation >= 256))
		return ERR_ILLEGAL_MODE;

	vMode.saturation = saturation;
	return s817_v4l2_video_mode(channel, &vMode);
}

int s817_get_saturation(struct s817_v4l2 *channel)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	return dev->mode[bChan].vMode.saturation;
}

ECODE s817_set_hue(struct s817_v4l2 *channel, int hue)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	VIDEO_MODE vMode = dev->mode[bChan].vMode;
	if (dev->mode[bChan].vMode.hue == hue)
		return 0;
	if ((hue < 0) || (hue >= 256))
		return ERR_ILLEGAL_MODE;

	vMode.hue = hue;
	return s817_v4l2_video_mode(channel, &vMode);
}

int s817_get_hue(struct s817_v4l2 *channel)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	return dev->mode[bChan].vMode.hue;
}

int s817_v4l2_capture_mode(struct s817_v4l2 *channel, CAPTURE_MODE *cMode)
{
	int bChan = channel->bChan;
	int aChan = bChan + 1;
	struct s817_dev *dev = channel->dev;
	int dChan = bChan % SYS_CHAN_PER_DSP;
	int rc;
	if (!dev->dsp[channel->dsp->dsp_num].cOpen[dChan]) {
		s817_openchan(channel->dev, aChan,
			      &channel->dev->mode[channel->bChan]);
	}
	rc = S817_SetCaptureMode(dev, aChan, cMode);
	return rc;
}


ECODE s817_set_BmpFPS(struct s817_v4l2 *channel, int frameRate)
{
	struct s817_dev         *dev = channel->dev;
	int                     bChan = channel->bChan;	
	CAPTURE_MODE            cMode = dev->mode[bChan].cMode;
	if (dev->mode[bChan].cMode.frameRateBmp == frameRate)
		return 0;
	if (frameRate > FR_MAX)
		return ERR_ILLEGAL_FRAMERATE;
	cMode.frameRateBmp = frameRate;
	return s817_v4l2_capture_mode(channel, &cMode);
}

ECODE s817_set_JpgFPS(struct s817_v4l2 *channel, int frameRate)
{
	struct s817_dev         *dev = channel->dev;
	int                     bChan = channel->bChan;
	CAPTURE_MODE            cMode = dev->mode[bChan].cMode;
	if (dev->mode[bChan].cMode.frameRateJpg == frameRate)
		return 0;
	if (frameRate > FR_MAX)
		return ERR_ILLEGAL_FRAMERATE;

	cMode.frameRateJpg = frameRate;
	return s817_v4l2_capture_mode(channel, &cMode);
}
#if 0
ECODE s817_set_capMode(struct s817_v4l2 *channel, int capMode)
{
	struct s817_dev         *dev = channel->dev;
	int                     bChan = channel->bChan;
	if (dev->mode[bChan].cMode.capMode == capMode)
		return 0;
	if ((capMode < 0) || (capMode >= CAP_MODE_MAX))
		return ERR_ILLEGAL_CAPMODE;

	dev->mode[bChan].cMode.capMode = capMode;
	return s817_set_capture_mode(channel);
}
#endif

ECODE s817_set_capType(struct s817_v4l2 *channel, int capType)
{
	struct s817_dev         *dev = channel->dev;
	int                     bChan = channel->bChan;
	if (dev->mode[bChan].cMode.capType == capType)
		return 0;
	if (capType & ~(CAP_JPG | CAP_BMP))
		return ERR_ILLEGAL_MODE;
	// save our capture type, but NOT in the mode structure.
	// capType in the mode structure controls stream on/off.
	dev->capType[bChan] = capType;
	return 0;
}

ECODE s817_set_compQuality(struct s817_v4l2 *channel, int compQuality)
{
	struct s817_dev         *dev = channel->dev;
	int                     bChan = channel->bChan;
	CAPTURE_MODE            cMode = dev->mode[bChan].cMode;
	if (dev->mode[bChan].cMode.compQuality == compQuality)
		return 0;
	if ((compQuality < COMP_QUAL_MIN) || (compQuality > COMP_QUAL_MAX))
		return ERR_ILLEGAL_COMPQUALITY;

	cMode.compQuality = compQuality;
	return s817_v4l2_capture_mode(channel, &cMode);
}


/* Internal data structure definitions which define the video */

struct s817_image_format {
	char *description;		/* text description */
	__u32 pixelformat;		/* V4L2 fourcc code */
	unsigned int flags;		/* V4L2 Image Format Desc Flags */
	enum v4l2_colorspace colorspace;
	__u32 sizeimage;
	int capType;
};


static const struct s817_image_format s817_image_formats[] = {
	{
		.description	= "JPEG",
		.pixelformat	= V4L2_PIX_FMT_JPEG,
		.flags		= V4L2_FMT_FLAG_COMPRESSED,
		.colorspace	= V4L2_COLORSPACE_JPEG,
		.sizeimage	= SYS_JPG_BUF_SIZE,
		.capType	= CAP_JPG,
	},{
		.description	= "8 Bit Luminance",
		.pixelformat	= V4L2_PIX_FMT_GREY,
		.flags		= 0,
		.colorspace	= 0,
		.sizeimage	= SYS_BMP_BUF_SIZE,
		.capType	= CAP_BMP,
	}
};
#define S817_NUM_FORMATS	ARRAY_SIZE(s817_image_formats)

struct s817_data_norm {
	char		*description;
	unsigned int	width, height;
	v4l2_std_id	std;
	unsigned int	fps_min, fps_max;
	__u32		framelines;
};

static const struct s817_data_norm s817_data_norms[] = {
	{
		.description	= "NTSC",
		.std		= V4L2_STD_NTSC_M,
		.fps_min	= 4,
		.fps_max	= 30,
		.framelines	= 525,
		.width		= SYS_NTSC_4CIF_VS,
		.height		= SYS_NTSC_4CIF_HS,
	},{
		.description	= "PAL",
		.std		= V4L2_STD_PAL_B,
		.fps_min	= 2,
		.fps_max	= 25,
		.framelines	= 625,
		.width		= SYS_PAL_4CIF_VS,
		.height		= SYS_PAL_4CIF_HS,
	}
};
#define S817_NUM_DATA_NORMS	ARRAY_SIZE(s817_data_norms)

struct v4l2_queryctrl s817_v4l2_controls[] = {
	{
		.id = V4L2_CID_BRIGHTNESS,
		.type = V4L2_CTRL_TYPE_INTEGER,
		.name = "Brightness",
		.minimum = 1,
		.maximum = 255,
		.step = 1,
		.default_value = 128,
	},{
		.id = V4L2_CID_CONTRAST,
		.type = V4L2_CTRL_TYPE_INTEGER,
		.name = "Contrast",
		.minimum = 0,
		.maximum = 255,
		.step = 1,
		.default_value = 128,
	},{
		.id = V4L2_CID_SATURATION,
		.type = V4L2_CTRL_TYPE_INTEGER,
		.name = "Saturation",
		.minimum = 0,
		.maximum = 255,
		.step = 1,
		.default_value = 128,
	},{
		.id = V4L2_CID_HUE,
		.type = V4L2_CTRL_TYPE_INTEGER,
		.name = "Hue",
		.minimum = 0,
		.maximum = 255,
		.step = 1,
		.default_value = 128,
	}
};
#define S817_V4L2_CONTROL_COUNT		ARRAY_SIZE(s817_v4l2_controls)

struct s817_ctrl_funcs {
	int id;
	int (*getfunction)(struct s817_v4l2 *);
	int (*setfunction)(struct s817_v4l2 *, int value);
};

struct s817_ctrl_funcs s817_control_functions[] = {
	{
		.id = V4L2_CID_BRIGHTNESS,
		.setfunction = s817_set_brightness,
		.getfunction = s817_get_brightness
	},{
		.id = V4L2_CID_CONTRAST,
		.setfunction = s817_set_contrast,
		.getfunction = s817_get_contrast
	},{
		.id = V4L2_CID_SATURATION,
		.setfunction = s817_set_saturation,
		.getfunction = s817_get_saturation
	},{
		.id = V4L2_CID_HUE,
		.setfunction = s817_set_hue,
		.getfunction = s817_get_hue
	}
};

static void s817_v4l2_querycap(struct v4l2_capability *cap)
{
	memset(cap, 0, sizeof(struct v4l2_capability));

	strcpy(cap->driver, s817_driver_name);
	strcpy(cap->card, s817_driver_description);
	strcpy(cap->bus_info, "PCI");
	cap->version = S817_VERSION_CODE;
	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING;
}

static int s817_v4l2_enuminput(struct s817_v4l2 *channel, 
                               struct v4l2_input *i)
{
	__u32 index = i->index;

	if (index >= SYS_CHAN_PER_BOARD) {
		return -EINVAL;
	}

	memset(i, 0, sizeof(struct v4l2_input));

	i->index = index;
	i->type = V4L2_INPUT_TYPE_CAMERA;
	i->std = V4L2_STD_NTSC | V4L2_STD_PAL;
	sprintf(i->name, "Channel %d", i->index);

	return 0;
}

static int s817_v4l2_g_input(struct s817_v4l2 *channel, unsigned int *i)
{
	int	boardno = channel->dsp->dev->minor;

	*i = channel->bChan + boardno*SYS_CHAN_PER_BOARD;

	return 0;
}

static int s817_v4l2_s_input(struct s817_v4l2 *channel, unsigned int *i)
{
	int	boardno = channel->dsp->dev->minor;

	/* changing channels not currently implemented */
	if (*i != channel->bChan + boardno*SYS_CHAN_PER_BOARD) {
		return -EINVAL;
	} else {
		return 0;
	}
}

static int s817_v4l2_enumstd(struct s817_v4l2 *channel, struct v4l2_standard *s)
{
	int index = s->index;

	if (index >= S817_NUM_DATA_NORMS) {
		return -EINVAL;
	}

	dprintk(125, "std name = %s\n", s817_data_norms[index].description);

	memset(s, 0, sizeof(struct v4l2_standard));
	s->index = index;

	s->id = s817_data_norms[index].std;
	s->frameperiod.numerator = 1;
	s->frameperiod.denominator = s817_data_norms[index].fps_max;
	s->framelines = s817_data_norms[index].framelines;
	strcpy(s->name, s817_data_norms[index].description);

	return 0;
}

static int s817_v4l2_querystd(struct s817_v4l2 *channel, v4l2_std_id *std)
{
	/*
	 * According to the V4L2 API, if we can't detect the standard, we must 
	 * report all supported standards.
	 */
	dprintk(4, "%s", __func__);
	*std = V4L2_STD_NTSC_M | V4L2_STD_PAL_B;

	return 0;
}

static int s817_v4l2_g_std(struct s817_v4l2 *channel, v4l2_std_id *std)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	dprintk(4, "%s", __func__);
	switch(dev->mode[bChan].vMode.tvFormat)
	{
	case TV_NTSC:
		*std = V4L2_STD_NTSC_M;
		break;
	case TV_PAL:
		*std = V4L2_STD_PAL_B;
		break;
	default:
		*std = 0;
	}
	return 0;
}

static int s817_v4l2_s_std(struct s817_v4l2 *channel, v4l2_std_id *std)
{
	int	DSPstd;
	int	ret = 0;
	dprintk(4, "%s", __func__);
	if (MUTEX_LOCK_INTERRUPTIBLE(&channel->mutex))
		return -EINTR;
	/* translate from V4L2 standard defines to DSP standard defines */
	if (*std & V4L2_STD_PAL) {
		DSPstd = TV_PAL;
	} else if (*std & V4L2_STD_NTSC) {
		DSPstd = TV_NTSC;
	} else {
		ret = -EINVAL;
		goto out;
	}
	
	ret = s817_set_videoStd(channel, DSPstd);
out:
	MUTEX_UNLOCK(&channel->mutex);

	return ret;
}

static int s817_v4l2_enum_fmt(struct s817_v4l2 *channel, 
                              struct v4l2_fmtdesc *fd)
{
	int	index = fd->index;
	dprintk(4, "%s", __func__);
	if (fd->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	if ((index < 0) || (index >= S817_NUM_FORMATS)) {
		return -EINVAL;
	}

	memset(fd, 0, sizeof(struct v4l2_fmtdesc));
	fd->index = index;
	fd->pixelformat = s817_image_formats[index].pixelformat;
	fd->flags = s817_image_formats[index].flags;
	strcpy(fd->description, s817_image_formats[index].description);

	return 0;
}

static int s817_v4l2_g_fmt(struct s817_v4l2 *channel, struct v4l2_format *f)
{
	struct v4l2_pix_format	*pf = &f->fmt.pix;
	int			i;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	dprintk(4, "%s", __func__);
	if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	if (MUTEX_LOCK_INTERRUPTIBLE(&channel->mutex))
		return -EINTR;
	GetBmpWH(&dev->mode[bChan], &pf->width, &pf->height);
	for (i=0; i<S817_NUM_FORMATS; i++) {
		if (dev->capType[bChan] == s817_image_formats[i].capType) {
			pf->pixelformat = s817_image_formats[i].pixelformat;
			pf->colorspace = s817_image_formats[i].colorspace;
			pf->sizeimage = s817_image_formats[i].sizeimage;
		}
	}
	pf->field = V4L2_FIELD_NONE;
	pf->bytesperline = pf->width;
	pf->priv = 0;

	MUTEX_UNLOCK(&channel->mutex);
	return 0;
}

// set the video format, or just try it (no real change)
static int s817_v4l2_s_fmt(struct s817_v4l2 *channel, 
				struct v4l2_format *f, int try)
{
	struct v4l2_pix_format	*pf = &f->fmt.pix;
	int			found = 0, i, ret = 0;
	int			width, height;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	dprintk(4, "%s", __func__);
	if  (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	if (MUTEX_LOCK_INTERRUPTIBLE(&channel->mutex))
		return -EINTR;

	for (i=0; i<S817_NUM_FORMATS; i++) {
		if (pf->pixelformat == s817_image_formats[i].pixelformat) {
			found = 1;
			if (!try) {
				s817_set_capType(channel, 
						s817_image_formats[i].capType);
			}
			pf->colorspace = s817_image_formats[i].colorspace;
			pf->sizeimage = s817_image_formats[i].sizeimage;
		}
	}
	if (!found) {
		ret = EINVAL;
		goto out;
	}
	GetBmpWH(&dev->mode[bChan], &width, &height);
	if (pf->width == -1)
		pf->width = width;
	if (pf->height == -1)
		pf->height = height;

	/* 
	 * We only support two heights and two widths.  Find out which is 
	 * closer to what the user asked for.
	 */
	if (pf->width > 480) {
		// width is closer to 640 than 320
		pf->width = 640;
		if (pf->height > 360) {
			// height is closer to 480 than 240
			pf->height = 480;
			if (!try)
				s817_set_jpgSize(channel, SIZE_4CIF); //640x480
		} else {
			pf->height = 240;
			if (!try)
				s817_set_jpgSize(channel, SIZE_2CIF); //640x240
		}
	} else {
		pf->width = 320;
		pf->height = 240;
		if (!try)
			s817_set_jpgSize(channel, SIZE_1CIF);	// 320x240
	}
	pf->bytesperline = pf->width;
	pf->field = V4L2_FIELD_NONE;
	pf->priv = 0;

out:
	MUTEX_UNLOCK(&channel->mutex);
	dprintk(4, "s_fmt width %d height %d bytes %d\n",
		pf->width, pf->height, pf->bytesperline);
	return ret;
}

static int s817_v4l2_cropcap(struct s817_v4l2 *channel, 
                             struct v4l2_cropcap *ccap)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	dprintk(4, "%s", __func__);
	if (ccap->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	if (MUTEX_LOCK_INTERRUPTIBLE(&channel->mutex))
		return -EINTR;

	ccap->bounds.left = 0;
	ccap->bounds.top = 0;
	GetBmpWH(&dev->mode[bChan], &ccap->bounds.width, &ccap->bounds.height);
	memcpy(&ccap->defrect, &ccap->bounds, sizeof(struct v4l2_rect));

	ccap->pixelaspect.numerator = 1;
	ccap->pixelaspect.denominator = 1;

	MUTEX_UNLOCK(&channel->mutex);

	return 0;
}

static int s817_v4l2_g_crop(struct s817_v4l2 *channel, struct v4l2_crop *c)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	dprintk(4, "%s", __func__);
	if  (c->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	if (MUTEX_LOCK_INTERRUPTIBLE(&channel->mutex))
	 	return -EINTR;

	// always full size
	c->c.left = 0;
	c->c.top = 0;
	GetBmpWH(&dev->mode[bChan], &c->c.width, &c->c.height);


	MUTEX_UNLOCK(&channel->mutex);
	return 0;
}

static int NTSCrates[]  = { 300, 240, 200, 150, 120, 100, 60, 30, 20, 0 };	/* in frames per 10s */
static int PALrates[]   = { 250, 200, 167, 125, 100, 83, 50, 25, 17, 0 };	/* in frames per 10s */

/* 
 * Find the index into the table rates that best provides the frame rate 
 * expressed by demoninator/numerator.  To avoid decimals, the table rate is 
 * expressed as frames per 10s and is terminated with a 0 entry.  Note that 
 * numerator/demoninator gives you the time between frames, but that's not 
 * what we are working with.
 */
static int  best_rate_index(int *rates, int numerator, int denominator)
{
	int		i, desired, best = 0;
	unsigned int	besterror = -1;		/* max unsigned int */
	unsigned int	error;

	/* Special case: asking for the maximum frame rate */
	if ((numerator == 0) || (denominator == 0)) {
		return 0;
	}

	desired = (denominator * 10)/numerator;		// frames per 10s
	for (i = 0; rates[i]; i++)
	{
		error = desired > rates[i] ? 
				desired - rates[i] : rates[i] - desired;
		if (error < besterror) {
			besterror = error;
			best = i;
		}
	}
	return best;
}


static int s817_v4l2_g_parm(struct s817_v4l2 *channel, 
				struct v4l2_streamparm *sp)
{
	struct v4l2_captureparm	*cp = &sp->parm.capture;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	if (sp->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	MUTEX_LOCK(&channel->mutex);
	memset(cp, 0, sizeof(struct v4l2_captureparm));

	cp->capability = V4L2_CAP_TIMEPERFRAME;

	cp->timeperframe.numerator = 10;
	if (dev->mode[bChan].vMode.tvFormat == TV_NTSC) {
		cp->timeperframe.denominator = 
				NTSCrates[dev->mode[bChan].cMode.frameRateJpg];
	} else {
		cp->timeperframe.denominator = 
				PALrates[dev->mode[bChan].cMode.frameRateJpg];
	}

	cp->reserved[0] = 0;
	cp->reserved[0] = 0;
	cp->reserved[0] = 0;
	cp->reserved[0] = 0;
	cp->extendedmode = 0;
	MUTEX_UNLOCK(&channel->mutex);
	return 0;
}

static int s817_v4l2_s_parm(struct s817_v4l2 *channel, 
				struct v4l2_streamparm *sp)
{
	struct v4l2_captureparm	*cp = &sp->parm.capture;
	int			ret = 0;
	int			rateindex;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	dprintk(4, "%s", __func__);
	if (sp->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}
	MUTEX_LOCK(&channel->mutex);

	if (dev->mode[bChan].vMode.tvFormat == TV_NTSC) {
		rateindex = best_rate_index(NTSCrates, 
				cp->timeperframe.numerator, 
				cp->timeperframe.denominator);
	} else {	/* PAL */
		rateindex = best_rate_index(PALrates, 
				cp->timeperframe.numerator, 
				cp->timeperframe.denominator);
	}
	if (s817_set_BmpFPS(channel, rateindex)) {
		ret = -EINVAL;
		goto out;
	}
	if (s817_set_JpgFPS(channel, rateindex)) {
		ret = -EINVAL;
		goto out;
	}

 out:
	cp->reserved[0] = 0;
	cp->reserved[0] = 0;
	cp->reserved[0] = 0;
	cp->reserved[0] = 0;
	cp->extendedmode = 0;
	MUTEX_UNLOCK(&channel->mutex);
	return 0;
}

/* buffers are already pre-allocated.  No need to 
 * send a smaller number of buffers to the firmware
 * as buffers will not be used unless releaseBuf is called
 */
static int s817_v4l2_reqbufs(struct s817_v4l2 *channel, 
				struct v4l2_requestbuffers *rb)
{
	int ret = 0;
	dprintk(4, "%s", __func__);
	if (rb->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	if (rb->memory != V4L2_MEMORY_MMAP) {
		dprintk(4, "reqbufs type not mmap\n");
		return -EINVAL;
	}

	if (rb->count > SYS_BUF_PER_CHAN) {
		printk("too many buffers requested !\n");
		return -EINVAL;
	}

	dprintk(4, "reqbufs count = %d\n", rb->count);

	if (MUTEX_LOCK_INTERRUPTIBLE(&channel->mutex))
		return -EINTR;
	if (rb->count > SYS_BUF_PER_CHAN)
		rb->count = SYS_BUF_PER_CHAN;

	if (rb->count > 0) {
		if (s817_v4l2_has_mapped_buffers(channel)) {
			dprintk(2, "busy, buffers still mapped\n");
			ret = -EBUSY;
			goto out;
		} else {
			s817_v4l2_streamoff(channel);
			CloseChannel(channel->dev, channel->bChan, 0);
			/*
			 * Note. system buffers are preallocated when 
			 * driver loads.  It is not necessary to allocate
			 * them again here and free them later.  This is done
			 * for performance reasons.  If rb->count < SYS_BUF_PER_CHAN,
			 * this is ok because the buffers aren't used by the hardware
			 *  (after streaming starts) until they are released.  When rb->count
			 *  < SYS_BUF_PER_CHAN, the extra buffers won't be used.
			 * An alternative could be to use videobuf and vmalloc the buffers each
			 * time here, but this would break the existing structure and is unnecessary.
			 * videobuf is also not supported on many lower kernel versions.
			 */
			s817_openchan(channel->dev, channel->bChan + 1,
				      &channel->dev->mode[channel->bChan]);
			channel->numbufs = SYS_BUF_PER_CHAN;
		}
	} else {
		s817_v4l2_streamoff(channel);
		CloseChannel(channel->dev, channel->bChan, 0);
	}
out:
	MUTEX_UNLOCK(&channel->mutex);

	return ret;
}

static int s817_v4l2_querybuf(struct s817_v4l2 *channel, 
				struct v4l2_buffer *b)
{
	int	err;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	int capType = dev->capType[bChan];
	dprintk(4, "%s", __func__);
	if (capType == (CAP_BMP | CAP_JPG)) 
		printk(KERN_WARNING "error, 2 cap types not supported for V4L single stream\n");

	if (b->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	b->flags = 0;

	MUTEX_LOCK(&channel->mutex);

	if (b->index >= channel->numbufs) {
		err = -EINVAL;
		goto out;
	}

        b->memory = V4L2_MEMORY_MMAP;
	b->field = V4L2_FIELD_NONE;
	if (dev->capType[bChan] == CAP_JPG) {
        	b->m.offset = (__u32) dev->sysBufJ[bChan][b->index].image.boff;
        	b->sequence = dev->sysBufJ[bChan][b->index].seqno;
        	b->length = SYS_JPG_BUF_SIZE;
		memcpy(&b->timestamp, &dev->sysBufJ[bChan][b->index].timestamp,
				sizeof(struct timeval));
		if (dev->sysBufJ[bChan][b->index].queued)
			b->flags |= V4L2_BUF_FLAG_QUEUED;
		if ((dev->sysBufJ[bChan][b->index].image.map_count) ||
	    			(dev->sysBufJ[bChan][b->index].status.map_count)) {
			b->flags |= V4L2_BUF_FLAG_MAPPED;
		}
	} else {
		b->m.offset = (unsigned long) dev->sysBufB[bChan][b->index].image.boff;
        	b->sequence = dev->sysBufB[bChan][b->index].seqno;
        	b->length = SYS_BMP_BUF_SIZE;
		memcpy(&b->timestamp, &dev->sysBufB[bChan][b->index].timestamp,
				sizeof(struct timeval));
		if (dev->sysBufB[bChan][b->index].queued)
			b->flags |= V4L2_BUF_FLAG_QUEUED;
		if ((dev->sysBufB[bChan][b->index].image.map_count) ||
		    (dev->sysBufB[bChan][b->index].status.map_count)) {
			b->flags |= V4L2_BUF_FLAG_MAPPED;
		}
	}

	if (FIFOfind(channel, b->index)) {
		b->flags |= V4L2_BUF_FLAG_DONE;
        	b->bytesused = b->length;
	} else {
        	b->bytesused = 0;
	}
out:
	MUTEX_UNLOCK(&channel->mutex);
	return 0;
}

static int s817_v4l2_qbuf(struct s817_v4l2 *channel, struct v4l2_buffer *b)
{
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	int dChan = bChan % SYS_CHAN_PER_DSP;
	int capType = dev->capType[bChan];
	dprintk(4, "%s", __func__);
	if (capType == (CAP_BMP | CAP_JPG)) 
		printk(KERN_WARNING "error, 2 cap types not supported for V4L single stream\n");

	if (b->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	if (b->memory != V4L2_MEMORY_MMAP)
		return -EINVAL;

	if (b->index >= channel->numbufs)
		return -EINVAL;

	dprintk(4, "qbuf2 %d %d %d\n", b->index, dev->mode[bChan].cMode.capType,
		dev->capType[bChan]);
	// release the buffer to the board
	releaseBufs(dev, channel->dsp->dsp_num, dChan, 1 << b->index, dev->capType[bChan]);
	if (dev->capType[bChan] == CAP_JPG)
		dev->sysBufJ[bChan][b->index].queued = 1;
	else
		dev->sysBufB[bChan][b->index].queued = 1;

	return 0;
}

static int s817_v4l2_dqbuf(struct s817_v4l2 *channel,
			   struct v4l2_buffer *b,
			   unsigned int nonblocking)
{
	union fifoelement 	fe;
	CAP_STAT 		*stat;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	int capType = dev->capType[bChan];
	dprintk(4, "%s", __func__);
	if (b->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		printk("inval\n");
		return -EINVAL;
	}
	if (capType == (CAP_BMP | CAP_JPG)) 
		printk("error, 2 cap types not supported for V4L single stream\n");
	if (!isNotEmpty(&dev->bufDataFIFO[bChan][capType - 1]))
	{
		if (nonblocking) {
			dprintk(50, "non-blocking I/O was selected and there "
				"are no buffers to dequeue\n");
			return -EAGAIN;
		}
		if (s817_wait_for_frame(channel))
		{
			/* interrupt or error */
			return -EIO;
		}

	}
	fe = FIFOread(&dev->bufDataFIFO[bChan][capType - 1]);

	b->index = fe.bufId;
        b->memory = V4L2_MEMORY_MMAP;
	b->field = V4L2_FIELD_NONE;
	b->flags = V4L2_BUF_FLAG_DONE;
	if (capType == CAP_JPG) {
        	stat = (CAP_STAT *) dev->sysBufJ[bChan][b->index].status.pData;
        	b->m.offset = (__u32) dev->sysBufJ[bChan][b->index].image.boff;
        	b->length = SYS_JPG_BUF_SIZE;
        	b->bytesused = stat->jpgsize;
        	b->sequence = dev->sysBufJ[bChan][b->index].seqno;
		memcpy(&b->timestamp, &dev->sysBufJ[bChan][b->index].timestamp,
				sizeof(struct timeval));
		if ((dev->sysBufJ[bChan][b->index].image.map_count) ||
	    			(dev->sysBufJ[bChan][b->index].status.map_count)) {
			b->flags |= V4L2_BUF_FLAG_MAPPED;
		}

		dev->sysBufJ[bChan][b->index].queued = 0;
	} else {
        	b->m.offset = (__u32) dev->sysBufB[bChan][b->index].image.boff;
        	b->sequence = dev->sysBufB[bChan][b->index].seqno;
        	b->length = SYS_BMP_BUF_SIZE;
        	b->bytesused = SYS_BMP_BUF_SIZE;
		memcpy(&b->timestamp, &dev->sysBufB[bChan][b->index].timestamp,
				sizeof(struct timeval));
		if ((dev->sysBufB[bChan][b->index].image.map_count) ||
	    			(dev->sysBufB[bChan][b->index].status.map_count)) {
			b->flags |= V4L2_BUF_FLAG_MAPPED;
		}
		dev->sysBufB[bChan][b->index].queued = 0;
	}
	return 0;
}

static int s817_v4l2_streamon(struct s817_v4l2 *channel)
{
	int aChan = channel->bChan + 1;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	int rc;
	dprintk(4, "%s", __func__);
	if (channel->streaming)
		return 0;

	if (channel->numbufs < 1) {
		dprintk(4, "no buffers allocated\n");
		return -EINVAL;
	}
	MUTEX_LOCK(&channel->mutex);
	// open channel (in case not opened already)
	s817_openchan(channel->dev, aChan,
			      &channel->dev->mode[channel->bChan]);
	/* always continuous mode for V4L */
	dev->mode[bChan].cMode.capMode = CAP_MODE_CONT;
	dev->mode[bChan].cMode.capType = dev->capType[bChan];
	rc = S817_SetVideoMode(dev, aChan, &dev->mode[bChan].vMode);
	if (rc != 0) {
		rc = -EIO;
		goto exit;
	}
	rc = S817_SetCaptureMode(dev, aChan, &dev->mode[bChan].cMode);

	if (rc != 0) {
		rc = -EIO;
	} else {
		channel->streaming = 1;
		channel->nextseqno = 0;
	}
exit:
	MUTEX_UNLOCK(&channel->mutex);
	return rc;
}

static int s817_v4l2_streamoff(struct s817_v4l2 *channel)
{
	int		retval = 0;
	struct s817_dev *dev = channel->dev;
	int bChan = channel->bChan;
	dprintk(4, "%s", __func__);
	if (!channel->streaming)
		return 0;
	channel->streaming = 0;
	dev->mode[bChan].cMode.capType = CAP_OFF;
	S817_SetCaptureMode(dev, bChan + 1, &dev->mode[bChan].cMode);
	return retval;
}

static int s817_v4l2_queryctrl(struct s817_v4l2 *channel, 
				struct v4l2_queryctrl *queryctrl)
{
	int i;
	for (i = 0; i < S817_V4L2_CONTROL_COUNT; i++) {
		if (s817_v4l2_controls[i].id == queryctrl->id) {
			memcpy(queryctrl, &s817_v4l2_controls[i], 
				sizeof(struct v4l2_queryctrl));
			queryctrl->reserved[0] = 0;
			return 0;
		}
	}

	return -EINVAL;
}

static int s817_v4l2_g_ctrl(struct s817_v4l2 *channel, 
				struct v4l2_control *control)
{
	int	i;
	int	ret = -EINVAL;
	MUTEX_LOCK(&channel->mutex);

	for (i = 0; i < S817_V4L2_CONTROL_COUNT; i++) {
		if (s817_control_functions[i].id == control->id) {
			control->value = 
				s817_control_functions[i].getfunction(channel);
			ret = 0;
			break;
		}
	}
	MUTEX_UNLOCK(&channel->mutex);

	return ret;
}

static int s817_v4l2_s_ctrl(struct s817_v4l2 *channel, 
				struct v4l2_control *control)
{
	int	i;
	int	ret = -EINVAL;
	MUTEX_LOCK(&channel->mutex);

	for (i = 0; i < S817_V4L2_CONTROL_COUNT; i++) {
		if (s817_control_functions[i].id == control->id) {
			if (s817_control_functions[i].setfunction(channel, 
						control->value)) {
				ret = -ERANGE;
			} else {
				ret = 0;
			}
		}
	}

	MUTEX_UNLOCK(&channel->mutex);

	return ret;
}


#ifdef S817_DEBUG_V4L2
static const char *v4l2_ioctls[] = {
	[_IOC_NR(VIDIOC_QUERYCAP)]         = "VIDIOC_QUERYCAP",
	[_IOC_NR(VIDIOC_RESERVED)]         = "VIDIOC_RESERVED",
	[_IOC_NR(VIDIOC_ENUM_FMT)]         = "VIDIOC_ENUM_FMT",
	[_IOC_NR(VIDIOC_G_FMT)]            = "VIDIOC_G_FMT",
	[_IOC_NR(VIDIOC_S_FMT)]            = "VIDIOC_S_FMT",
	[_IOC_NR(VIDIOC_REQBUFS)]          = "VIDIOC_REQBUFS",
	[_IOC_NR(VIDIOC_QUERYBUF)]         = "VIDIOC_QUERYBUF",
	[_IOC_NR(VIDIOC_G_FBUF)]           = "VIDIOC_G_FBUF",
	[_IOC_NR(VIDIOC_S_FBUF)]           = "VIDIOC_S_FBUF",
	[_IOC_NR(VIDIOC_OVERLAY)]          = "VIDIOC_OVERLAY",
	[_IOC_NR(VIDIOC_QBUF)]             = "VIDIOC_QBUF",
	[_IOC_NR(VIDIOC_DQBUF)]            = "VIDIOC_DQBUF",
	[_IOC_NR(VIDIOC_STREAMON)]         = "VIDIOC_STREAMON",
	[_IOC_NR(VIDIOC_STREAMOFF)]        = "VIDIOC_STREAMOFF",
	[_IOC_NR(VIDIOC_G_PARM)]           = "VIDIOC_G_PARM",
	[_IOC_NR(VIDIOC_S_PARM)]           = "VIDIOC_S_PARM",
	[_IOC_NR(VIDIOC_G_STD)]            = "VIDIOC_G_STD",
	[_IOC_NR(VIDIOC_S_STD)]            = "VIDIOC_S_STD",
	[_IOC_NR(VIDIOC_ENUMSTD)]          = "VIDIOC_ENUMSTD",
	[_IOC_NR(VIDIOC_ENUMINPUT)]        = "VIDIOC_ENUMINPUT",
	[_IOC_NR(VIDIOC_G_CTRL)]           = "VIDIOC_G_CTRL",
	[_IOC_NR(VIDIOC_S_CTRL)]           = "VIDIOC_S_CTRL",
	[_IOC_NR(VIDIOC_G_TUNER)]          = "VIDIOC_G_TUNER",
	[_IOC_NR(VIDIOC_S_TUNER)]          = "VIDIOC_S_TUNER",
	[_IOC_NR(VIDIOC_G_AUDIO)]          = "VIDIOC_G_AUDIO",
	[_IOC_NR(VIDIOC_S_AUDIO)]          = "VIDIOC_S_AUDIO",
	[_IOC_NR(VIDIOC_QUERYCTRL)]        = "VIDIOC_QUERYCTRL",
	[_IOC_NR(VIDIOC_QUERYMENU)]        = "VIDIOC_QUERYMENU",
	[_IOC_NR(VIDIOC_G_INPUT)]          = "VIDIOC_G_INPUT",
	[_IOC_NR(VIDIOC_S_INPUT)]          = "VIDIOC_S_INPUT",
	[_IOC_NR(VIDIOC_G_OUTPUT)]         = "VIDIOC_G_OUTPUT",
	[_IOC_NR(VIDIOC_S_OUTPUT)]         = "VIDIOC_S_OUTPUT",
	[_IOC_NR(VIDIOC_ENUMOUTPUT)]       = "VIDIOC_ENUMOUTPUT",
	[_IOC_NR(VIDIOC_G_AUDOUT)]         = "VIDIOC_G_AUDOUT",
	[_IOC_NR(VIDIOC_S_AUDOUT)]         = "VIDIOC_S_AUDOUT",
	[_IOC_NR(VIDIOC_G_MODULATOR)]      = "VIDIOC_G_MODULATOR",
	[_IOC_NR(VIDIOC_S_MODULATOR)]      = "VIDIOC_S_MODULATOR",
	[_IOC_NR(VIDIOC_G_FREQUENCY)]      = "VIDIOC_G_FREQUENCY",
	[_IOC_NR(VIDIOC_S_FREQUENCY)]      = "VIDIOC_S_FREQUENCY",
	[_IOC_NR(VIDIOC_CROPCAP)]          = "VIDIOC_CROPCAP",
	[_IOC_NR(VIDIOC_G_CROP)]           = "VIDIOC_G_CROP",
	[_IOC_NR(VIDIOC_S_CROP)]           = "VIDIOC_S_CROP",
	[_IOC_NR(VIDIOC_G_JPEGCOMP)]       = "VIDIOC_G_JPEGCOMP",
	[_IOC_NR(VIDIOC_S_JPEGCOMP)]       = "VIDIOC_S_JPEGCOMP",
	[_IOC_NR(VIDIOC_QUERYSTD)]         = "VIDIOC_QUERYSTD",
	[_IOC_NR(VIDIOC_TRY_FMT)]          = "VIDIOC_TRY_FMT",
	[_IOC_NR(VIDIOC_ENUMAUDIO)]        = "VIDIOC_ENUMAUDIO",
	[_IOC_NR(VIDIOC_ENUMAUDOUT)]       = "VIDIOC_ENUMAUDOUT",
	[_IOC_NR(VIDIOC_G_PRIORITY)]       = "VIDIOC_G_PRIORITY",
	[_IOC_NR(VIDIOC_S_PRIORITY)]       = "VIDIOC_S_PRIORITY",
	[_IOC_NR(VIDIOC_G_SLICED_VBI_CAP)] = "VIDIOC_G_SLICED_VBI_CAP",
	[_IOC_NR(VIDIOC_LOG_STATUS)]       = "VIDIOC_LOG_STATUS",
	[_IOC_NR(VIDIOC_G_EXT_CTRLS)]      = "VIDIOC_G_EXT_CTRLS",
	[_IOC_NR(VIDIOC_S_EXT_CTRLS)]      = "VIDIOC_S_EXT_CTRLS",
	[_IOC_NR(VIDIOC_TRY_EXT_CTRLS)]    = "VIDIOC_TRY_EXT_CTRLS",
	[_IOC_NR(VIDIOC_ENUM_FRAMESIZES)]  = "VIDIOC_ENUM_FRAMESIZES",
	[_IOC_NR(VIDIOC_ENUM_FRAMEINTERVALS)] = "VIDIOC_ENUM_FRAMEINTERVALS",
	[_IOC_NR(VIDIOC_G_ENC_INDEX)] 	   = "VIDIOC_G_ENC_INDEX",
	[_IOC_NR(VIDIOC_ENCODER_CMD)] 	   = "VIDIOC_ENCODER_CMD",
	[_IOC_NR(VIDIOC_TRY_ENCODER_CMD)]  = "VIDIOC_TRY_ENCODER_CMD",

	[_IOC_NR(VIDIOC_DBG_S_REGISTER)]   = "VIDIOC_DBG_S_REGISTER",
	[_IOC_NR(VIDIOC_DBG_G_REGISTER)]   = "VIDIOC_DBG_G_REGISTER",

//	[_IOC_NR(VIDIOC_G_CHIP_IDENT)]     = "VIDIOC_G_CHIP_IDENT",
	[_IOC_NR(VIDIOC_S_HW_FREQ_SEEK)]   = "VIDIOC_S_HW_FREQ_SEEK",
};
#define V4L2_IOCTLS ARRAY_SIZE(v4l2_ioctls)
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
static int s817_v4l2_open(struct file *file)
#else
static int s817_v4l2_open(struct inode *inode, struct file *file)
#endif
{
	struct s817_v4l2 *channel = video_drvdata(file);
	struct s817_dev *dev = channel->dev;
	if (dev == NULL) {
		return -ENODEV;
	}
	MUTEX_LOCK(&dev->open_lock);
	MUTEX_LOCK(&channel->mutex);
	if (channel->users >= 1) {
		printk(KERN_INFO "s817: one handle at a time\n");
		MUTEX_UNLOCK(&channel->mutex);
		MUTEX_UNLOCK(&dev->open_lock);
		return -EBUSY;
	}
	channel->users++;
	MUTEX_UNLOCK(&channel->mutex);
	MUTEX_UNLOCK(&dev->open_lock);
	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
static int s817_v4l2_close(struct file *file)
#else
static int s817_v4l2_close(struct inode *inode, struct file *file)
#endif
{
	struct s817_v4l2 *channel = video_drvdata(file);
	struct s817_dev *dev = channel->dev;
	int			dChan;
	int j;
	struct s817_dsp		*dsp;
	CAPTURE_MODE mode;
	int any_channel_open = 0;
	dprintk(125, "close():\n");
	if (channel == NULL)
		return 0;

	MUTEX_LOCK(&channel->mutex);

	dChan = channel->bChan % SYS_CHAN_PER_DSP;
	dsp = channel->dsp;

	channel->users--;
	
	j = channel->bChan;
	mode = dev->mode[j].cMode;
	if (mode.capType != CAP_OFF) {
		mode.capType = CAP_OFF;
		S817_SetCaptureMode(dev, j+1, &mode);
	}
	CloseChannel(dev, j, 0);
	/* wait for dma to stop. plus one frame */
	msleep(50);

	/* Empty the FIFOs */
	while (isNotEmpty(&dev->bufDataFIFO[j][CAP_JPG - 1]))
		FIFOread(&dev->bufDataFIFO[j][CAP_JPG - 1]);
	while (isNotEmpty(&dev->bufDataFIFO[j][CAP_BMP - 1]))
		FIFOread(&dev->bufDataFIFO[j][CAP_BMP - 1]);

	for (j = 0; j < SYS_CHAN_PER_DSP; j++) {
		if (dsp->cOpen[j]) {
			any_channel_open = 1;
			break;
		}
	}

	if ((fwmode == 1) && !any_channel_open) {
		if (dsp->intEnabled == 1) {
			int lBuf;
			lBuf = s817_readDSPreg(dsp->log_bar1, DSP_HSR);
			/* disable the interrupt */
			lBuf |= DSP_HSR_INTAM;
			s817_writeDSPreg(dsp->log_bar1, DSP_HSR, lBuf);
			(void) s817_readDSPreg(dsp->log_bar1, DSP_HSR);
			dsp->intEnabled = 0;
		}
		dsp->upAndRunning = 0;
		dev->dsp_loaded[dsp->dsp_num] = 0;
	}
	MUTEX_UNLOCK(&channel->mutex);
	return 0;
}
 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 29)
static int s817_v4l2_ioctl_core(struct inode *inode, struct file *file,
		      unsigned int cmd, void *arg)
#else
static long s817_v4l2_ioctl_core(struct file *file,
			   unsigned int cmd, void *arg)
#endif
{
	struct s817_v4l2 *channel = video_drvdata(file);

	switch (cmd) {
	/* V4L2 interface */
	case VIDIOC_QUERYCAP: {
		s817_v4l2_querycap(arg);
		break;
	}
	case VIDIOC_ENUMINPUT: {
		return s817_v4l2_enuminput(channel, arg);
	}
	case VIDIOC_G_INPUT: {
		return s817_v4l2_g_input(channel, arg);
	}
	case VIDIOC_S_INPUT: {
		return s817_v4l2_s_input(channel, arg);
	}
	case VIDIOC_ENUMSTD: {
		return s817_v4l2_enumstd(channel, arg);
	}
	case VIDIOC_QUERYSTD: {
		return s817_v4l2_querystd(channel, arg);
	}
	case VIDIOC_G_STD: {
		return s817_v4l2_g_std(channel, arg);
	}
	case VIDIOC_S_STD: {
		return s817_v4l2_s_std(channel, arg);
	}
	case VIDIOC_ENUM_FMT: {
		return s817_v4l2_enum_fmt(channel, arg);
	}
	case VIDIOC_G_FMT: {
		return s817_v4l2_g_fmt(channel, arg);
	}
	case VIDIOC_S_FMT: {
		return s817_v4l2_s_fmt(channel, arg, 0);
	}
	case VIDIOC_TRY_FMT: {
		return s817_v4l2_s_fmt(channel, arg, 1);
	}
	case VIDIOC_CROPCAP: {
		return s817_v4l2_cropcap(channel, arg);
	}
	case VIDIOC_G_CROP: {
		return s817_v4l2_g_crop(channel, arg);
	}
	case VIDIOC_G_PARM: {

		return s817_v4l2_g_parm(channel, arg);
	}
	case VIDIOC_S_PARM: {
		return s817_v4l2_s_parm(channel, arg);
	}
	case VIDIOC_REQBUFS: {
		return s817_v4l2_reqbufs(channel, arg);
	}
	case VIDIOC_QUERYBUF: {
		return s817_v4l2_querybuf(channel, arg);
	}
	case VIDIOC_QBUF: {
		return s817_v4l2_qbuf(channel, arg);
	}
	case VIDIOC_DQBUF: {
		return s817_v4l2_dqbuf(channel, arg, 
					file->f_flags & O_NONBLOCK);
	}
	case VIDIOC_STREAMON: {
		return s817_v4l2_streamon(channel);
	}
	case VIDIOC_STREAMOFF: {
		return s817_v4l2_streamoff(channel);
	}
	case VIDIOC_QUERYCTRL: {
		return s817_v4l2_queryctrl(channel, arg);
	}
	case VIDIOC_G_CTRL: {
		return s817_v4l2_g_ctrl(channel, arg);
	}
	case VIDIOC_S_CTRL: {
		return s817_v4l2_s_ctrl(channel, arg);
	}
	case VIDIOC_S_CROP:		// croping not supported
	default:
		return -ENOIOCTLCMD;
	}

	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
static long s817_v4l2_ioctl(struct file *file,
		      unsigned int cmd, unsigned long arg)
#else
static int s817_v4l2_ioctl(struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg)
#endif
{
	int ret;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
	ret = video_usercopy(file, cmd, arg, s817_v4l2_ioctl_core);
#else
	ret = video_usercopy(inode, file, cmd, arg, s817_v4l2_ioctl_core);
#endif

#ifdef S817_DEBUG_V4L2
	if ((_IOC_TYPE(cmd) == 'V') && (ret != 0)) {
		const char *cmdstr;

		if (_IOC_NR(cmd) >= V4L2_IOCTLS)
			cmdstr = "unknown v4l2 ioctl";
		else
			cmdstr = v4l2_ioctls[_IOC_NR(cmd)];
		printk("%s returned %d\n", cmdstr, ret);
	}
#endif
	return ret;
}
 



static unsigned int s817_v4l2_poll(struct file *file, poll_table *pt)
{
	struct s817_v4l2 *channel = video_drvdata(file);
	struct s817_dev *dev = channel->dev;
	if (isNotEmpty(&dev->bufDataFIFO[channel->bChan][CAP_JPG - 1]) ||
	    isNotEmpty(&dev->bufDataFIFO[channel->bChan][CAP_BMP - 1])) {
		return POLLIN;
	}
	poll_wait(file, &channel->frame_wait_queue, pt);
	if (isNotEmpty(&dev->bufDataFIFO[channel->bChan][CAP_JPG - 1]) ||
	    isNotEmpty(&dev->bufDataFIFO[channel->bChan][CAP_BMP - 1])) {
		return POLLIN;
	}
	return 0;

}
/* map kernel buffers to user space*/
static int s817_mmap_v4l2(struct file *file, struct vm_area_struct *vma)
{
	struct s817_v4l2 *channel = video_drvdata(file);
	struct s817_dev *dev = channel->dev;
	int rc;
	rc = s817_mmap_common(dev, channel, vma);
	return rc;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
static const struct v4l2_file_operations s817_v4l2_fops = {
#else
static const struct file_operations s817_v4l2_fops = {
#endif
	.owner		= THIS_MODULE,
	.open		= s817_v4l2_open,
	.release	= s817_v4l2_close,
	.ioctl		= s817_v4l2_ioctl,
	.mmap		= s817_mmap_v4l2,
	.poll		= s817_v4l2_poll,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 29)
	.llseek = no_llseek,
#endif
};

static struct video_device v4l_device_template = {
	.name		= "NOT SET",
	.fops		= &s817_v4l2_fops,
	.minor		= -1,
};



/*
 * Initial the channel.  This is called by at board initialization time.
 */
static int s817_init_v4l2(struct s817_v4l2 *channel, unsigned int number)
{
	int	boardno = channel->dsp->dev->minor;

	channel->bChan = number;
	channel->users = 0;

	MUTEX_INIT(&channel->mutex);
	init_waitqueue_head(&channel->frame_wait_queue);

	channel->v4l_device = video_device_alloc();
	if (!channel->v4l_device) {
		return -ENOMEM;
	}

	memcpy(channel->v4l_device, &v4l_device_template, 
		sizeof(struct video_device));
	snprintf(channel->v4l_device->name, sizeof(channel->v4l_device->name), 
		"Channel %d", number + boardno*SYS_CHAN_PER_BOARD);
	channel->v4l_device->release = video_device_release;
	video_set_drvdata(channel->v4l_device, channel);

	if  (video_register_device(channel->v4l_device, 
					VFL_TYPE_GRABBER, -1) < 0) {
		printk(KERN_ERR "Video4Linux-device registration failed\n");
		video_device_release(channel->v4l_device);
		channel->v4l_device = NULL;
		return -EINVAL;
	}
#ifdef CONFIG_PROC_FS
	create_chan_proc_entries(channel);
#endif
	return 0;
}


#ifdef CONFIG_PROC_FS
/***************  proc filesystem routines *****************************/

static int read_fwver(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_dev		*dev;
	int			len = 0;
	int			ver = 0;

	dev = list_first_entry(&s817_devlist, struct s817_dev, s817_devlist);
	if (dev)
	{
		ver = dev->dsp[0].sysData.fwVer;
	}
	sprintf(page, "%d.%d\n", ver >> 16, ver & 0xffff);
	len = strlen(page);
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}
static int read_output(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_output	*output = (struct s817_output *)data;
	int			len = 0;

	sprintf(page, "%d\n", output->input);
	len = strlen(page);
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_output(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_output	*output = (struct s817_output *)data;
	struct s817_dev		*dev = output->dev;
	struct s817_dsp		*dsp;
	int		        dsp_idx = dev->dsp0 - 1; // send to DSP 0
	int			retval = count;
	char			*page, *buffer;
	unsigned int		switchOut = 0;
	int			i;
	dsp = &dev->dsp[dsp_idx];
	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		unsigned int	sel = (unsigned int)simple_strtol(buffer, 
								&buffer, 10);
		long		lBuf0, lBuf1;
		ECODE		ecode = 0;

		if (sel > SYS_CHAN_PER_BOARD) {
			retval = -EINVAL;
			goto out;
		}

		output->input = sel;

		/* Need to send the board info for all outputs at once */
		for (i=0; i<SYS_OUT_PER_BOARD; i++) {
			switchOut |= dev->output[i].input << (8 * i);
		}

		MUTEX_LOCK(&dsp->dsplock);
		lBuf0 = switchOut;
		lBuf1 = DSP_CMD_SET_VSWITCH << 4;
		ecode = sendDspCmd(dev, dsp_idx, lBuf1, lBuf0);
		if (ecode == 0) {
			/* enable all outputs */
			lBuf0 = 0xf;
			lBuf1 = DSP_SET_VOUT << 4;
			ecode = sendDspCmd(dev, dsp_idx, lBuf1, lBuf0);
		}
		MUTEX_UNLOCK(&dsp->dsplock);
	}

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_osd_on(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;
	sprintf(page, "%d\n", channel->dev->osdmode[channel->bChan].osdOn);
	len = 2;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else {
		len = count;
	}
	return len;
}

static int write_osd_on(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		if (*buffer == '0')
			channel->dev->osdmode[channel->bChan].osdOn = 0;
		else
			channel->dev->osdmode[channel->bChan].osdOn = 1;

		MUTEX_LOCK(&channel->mutex);
		S817_SetOsdMode(channel->dev, channel->bChan+1, &channel->dev->osdmode[channel->bChan]);
		MUTEX_UNLOCK(&channel->mutex);
	}

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_osd_bmp(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;

	sprintf(page, "%d\n", channel->dev->osdmode[channel->bChan].osdBmp);
	len = 2;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_osd_bmp(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		if (*buffer == '0')
			channel->dev->osdmode[channel->bChan].osdBmp = 0;
		else
			channel->dev->osdmode[channel->bChan].osdBmp = 1;

		MUTEX_LOCK(&channel->mutex);
		S817_SetOsdMode(channel->dev, channel->bChan+1, &channel->dev->osdmode[channel->bChan]);
		MUTEX_UNLOCK(&channel->mutex);
	}

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_osd_transparent(char *page, char **start, off_t off, int count,
				int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;

	sprintf(page, "%d\n", channel->dev->osdmode[channel->bChan].transparent);
	len = 2;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_osd_transparent(struct file *file, const char __user *buf,
				unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		if (*buffer == '0')
			channel->dev->osdmode[channel->bChan].transparent = 0;
		else
			channel->dev->osdmode[channel->bChan].transparent = 1;

		MUTEX_LOCK(&channel->mutex);
		S817_SetOsdMode(channel->dev, channel->bChan+1, &channel->dev->osdmode[channel->bChan]);
		MUTEX_UNLOCK(&channel->mutex);
	}

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_osd_top(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;

	sprintf(page, "%d\n", channel->dev->osdmode[channel->bChan].positionTop);
	len = 2;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_osd_top(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		if (*buffer == '0')
			channel->dev->osdmode[channel->bChan].positionTop = 0;
		else
			channel->dev->osdmode[channel->bChan].positionTop = 1;

		MUTEX_LOCK(&channel->mutex);
		S817_SetOsdMode(channel->dev, channel->bChan+1, &channel->dev->osdmode[channel->bChan]);
		MUTEX_UNLOCK(&channel->mutex);
	}

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_osd_ddmm(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;

	sprintf(page, "%d\n", channel->dev->osdmode[channel->bChan].ddmm);
	len = 2;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_osd_ddmm(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		if (*buffer == '0')
			channel->dev->osdmode[channel->bChan].ddmm = 0;
		else
			channel->dev->osdmode[channel->bChan].ddmm = 1;

		MUTEX_LOCK(&channel->mutex);
		S817_SetOsdMode(channel->dev, channel->bChan+1, &channel->dev->osdmode[channel->bChan]);
		MUTEX_UNLOCK(&channel->mutex);
	}

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_osd_year2(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;

	sprintf(page, "%d\n", channel->dev->osdmode[channel->bChan].year2);
	len = 2;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_osd_year2(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		if (*buffer == '0')
			channel->dev->osdmode[channel->bChan].year2 = 0;
		else
			channel->dev->osdmode[channel->bChan].year2 = 1;

		MUTEX_LOCK(&channel->mutex);
		S817_SetOsdMode(channel->dev, channel->bChan+1, &channel->dev->osdmode[channel->bChan]);
		MUTEX_UNLOCK(&channel->mutex);
	}

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_osd_fraction(char *page, char **start, off_t off, int count,
				int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;

	sprintf(page, "%d\n", channel->dev->osdmode[channel->bChan].fraction);
	len = 2;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_osd_fraction(struct file *file, const char __user *buf,
				unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		if ((*buffer >= '0') && (*buffer <= '2'))
			channel->dev->osdmode[channel->bChan].fraction = *buffer - '0';
		else
			retval = -EINVAL;

		MUTEX_LOCK(&channel->mutex);
		S817_SetOsdMode(channel->dev, channel->bChan+1, &channel->dev->osdmode[channel->bChan]);
		MUTEX_UNLOCK(&channel->mutex);
	}

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_osd_text(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;

	sprintf(page, "%s\n", channel->dev->osdmode[channel->bChan].line);
	len = strlen(channel->dev->osdmode[channel->bChan].line) + 1;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0) return 0;
	} else
		len = count;
	return len;
}

static int write_osd_text(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	int			copysize;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	buffer = page;

	/* Skip over leading whitespace */
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	copysize = count + 1;
	if (copysize > sizeof(channel->dev->osdmode[channel->bChan].line))		//silently limit
	{
		copysize = sizeof(channel->dev->osdmode[channel->bChan].line);
		buffer[copysize-1] = '\0';
	}
	strncpy(channel->dev->osdmode[channel->bChan].line, buffer, copysize);

	MUTEX_LOCK(&channel->mutex);
	S817_SetOsdMode(channel->dev, channel->bChan+1, &channel->dev->osdmode[channel->bChan]);
	MUTEX_UNLOCK(&channel->mutex);

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_inuse(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;

	sprintf(page, "%d\n", channel->users ? 1 : 0);
	len = 2;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}
static int read_fps(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;
	MODE *mode = &channel->dev->mode[channel->bChan];
	if (mode->vMode.tvFormat == TV_NTSC)
	{
		sprintf(page, "%d\n", 
			NTSCrates[mode->cMode.frameRateJpg]);
	}
	else
	{
		sprintf(page, "%d\n", 
			PALrates[mode->cMode.frameRateJpg]);
	}
	len = strlen(page);
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_fps(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	int			rate, rateindex;
	char			*page, *buffer;
	MODE *mode = &channel->dev->mode[channel->bChan];
	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	buffer = page;

	/* Skip over leading whitespace */
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}
	rate = (int)simple_strtol(buffer, &buffer, 10);

	MUTEX_LOCK(&channel->mutex);
	if (mode->vMode.tvFormat == TV_NTSC) {
		rateindex = best_rate_index(NTSCrates, 10, rate);
	} else {	/* PAL */
		rateindex = best_rate_index(PALrates, 10, rate);
	}
	s817_set_BmpFPS(channel, rateindex);
	s817_set_JpgFPS(channel, rateindex);
	MUTEX_UNLOCK(&channel->mutex);

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_standard(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;
	MODE *mode = &channel->dev->mode[channel->bChan];
	if (mode->vMode.tvFormat == TV_NTSC)
	{
		sprintf(page, "NTSC\n");
	}
	else
	{
		sprintf(page, "PAL\n");
	}
	len = strlen(page);
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_standard(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	buffer = page;

	/* Skip over leading whitespace */
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}
	if (!strncmp("PAL", buffer, 3))
		s817_set_videoStd(channel, TV_PAL);
	else
		s817_set_videoStd(channel, TV_NTSC);

out:
	free_page((unsigned long)page);
	return retval;
}

static int read_size(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			len = 0;
	MODE *mode = &channel->dev->mode[channel->bChan];
	sprintf(page, "%d\n", mode->vMode.jpgSize);
	len = strlen(page);
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;
	return len;
}

static int write_size(struct file *file, const char __user *buf,
			unsigned long count, void *data)
{
	struct s817_v4l2	*channel = (struct s817_v4l2 *)data;
	int			retval = count;
	char			*page, *buffer;
	int			jpgsize;

	if (count > PAGE_SIZE) {
		printk(KERN_ERR "count is %lu > %d!!!\n", 
					count, (int)PAGE_SIZE);
		return -ENOSPC;
	}

	if (!(page = (char *)__get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	if(copy_from_user(page, buf, count))
	{
		retval = -EFAULT;
		goto out;
	}

	if (page[count-1] == '\n')
		page[count-1] = '\0';
	else if (count < PAGE_SIZE)
		page[count] = '\0';
	else if (page[count]) {
		retval = -EINVAL;
		goto out;
	}

	/* Skip over leading whitespace */
	buffer = page;
	while (count && isspace(*buffer)) {
		--count;
		++buffer;
	}

	if (count > 0) {
		if ((*buffer >= '0') && (*buffer <= '3'))
		{
			jpgsize = *buffer - '0';
			MUTEX_LOCK(&channel->mutex);
			s817_set_jpgSize(channel, jpgsize);
			MUTEX_UNLOCK(&channel->mutex);
		}
		else
		{
			retval = -EINVAL;
		}
	}

out:
	free_page((unsigned long)page);
	return retval;
}

void create_chan_proc_entries(struct s817_v4l2 *channel)
{
	struct proc_dir_entry	*channelroot = NULL;
	struct proc_dir_entry   *osdroot = NULL, *ent;
	mode_t 			mode, dirmode;
	char			name[20];
	int			boardno = channel->dsp->dev->minor;

	mode = S_IFREG | S_IRUGO | S_IWUGO;
	dirmode = S_IFDIR | S_IRWXU | S_IRWXG | S_IRWXO;
	
	if (s817_proc_root == NULL) {
		printk("null root\n");
		return;
	}

	sprintf(name, "channel%d", channel->bChan + boardno*SYS_CHAN_PER_BOARD);
	channelroot = create_proc_entry(name, dirmode, s817_proc_root);

	if (channelroot) {
		ent = create_proc_entry("inuse", mode, channelroot);
		if (ent) {
			ent->data = channel;
			ent->read_proc = read_inuse;
			ent->write_proc = NULL;
		}
		ent = create_proc_entry("fps", mode, channelroot);
		if (ent) {
			ent->data = channel;
			ent->read_proc = read_fps;
			ent->write_proc = write_fps;
		}
		ent = create_proc_entry("standard", mode, channelroot);
		if (ent) {
			ent->data = channel;
			ent->read_proc = read_standard;
			ent->write_proc = write_standard;
		}
		ent = create_proc_entry("size", mode, channelroot);
		if (ent) {
			ent->data = channel;
			ent->read_proc = read_size;
			ent->write_proc = write_size;
		}

		osdroot = create_proc_entry("osd", dirmode, channelroot);
		if (osdroot) {
			ent = create_proc_entry("on", mode, osdroot);
			if (ent) {
				ent->data = channel;
				ent->read_proc = read_osd_on;
				ent->write_proc = write_osd_on;
			}
			ent = create_proc_entry("BMP", mode, osdroot);
			if (ent) {
				ent->data = channel;
				ent->read_proc = read_osd_bmp;
				ent->write_proc = write_osd_bmp;
			}
			ent = create_proc_entry("transparent", mode, osdroot);
			if (ent) {
				ent->data = channel;
				ent->read_proc = read_osd_transparent;
				ent->write_proc = write_osd_transparent;
			}
			ent = create_proc_entry("top", mode, osdroot);
			if (ent) {
				ent->data = channel;
				ent->read_proc = read_osd_top;
				ent->write_proc = write_osd_top;
			}
			ent = create_proc_entry("ddmm", mode, osdroot);
			if (ent) {
				ent->data = channel;
				ent->read_proc = read_osd_ddmm;
				ent->write_proc = write_osd_ddmm;
			}
			ent = create_proc_entry("year2", mode, osdroot);
			if (ent) {
				ent->data = channel;
				ent->read_proc = read_osd_year2;
				ent->write_proc = write_osd_year2;
			}
			ent = create_proc_entry("fraction", mode, osdroot);
			if (ent) {
				ent->data = channel;
				ent->read_proc = read_osd_fraction;
				ent->write_proc = write_osd_fraction;
			}
			ent = create_proc_entry("text", mode, osdroot);
			if (ent) {
				ent->data = channel;
				ent->read_proc = read_osd_text;
				ent->write_proc = write_osd_text;
			}
		}	
	} else {
			printk(KERN_ERR "Failed to make proc directory "
					"for %s\n", name);
	}
	channel->channelroot = channelroot;
	channel->osdroot = osdroot;
}


void remove_chan_proc_entries(struct s817_v4l2 *channel)
{
	struct proc_dir_entry	*channelroot, *osdroot;
	char			name[20];
	int			boardno = channel->dsp->dev->minor;
	channelroot = channel->channelroot;
	osdroot = channel->osdroot;
	remove_proc_entry("BMP", osdroot);
	remove_proc_entry("ddmm", osdroot);
	remove_proc_entry("fraction", osdroot);
	remove_proc_entry("on", osdroot);
	remove_proc_entry("text", osdroot);
	remove_proc_entry("top", osdroot);
	remove_proc_entry("transparent", osdroot);
	remove_proc_entry("year2", osdroot);
	remove_proc_entry("osd", channelroot);
	remove_proc_entry("fps", channelroot);
	remove_proc_entry("inuse", channelroot);		
	remove_proc_entry("standard", channelroot);
	remove_proc_entry("size", channelroot);
	sprintf(name, "channel%d", channel->bChan + boardno*SYS_CHAN_PER_BOARD);
	remove_proc_entry(name, s817_proc_root);
	return;
}

#endif
#endif
